/* exs.c */

/*
 * Copyright 2007, 2008, 2009, 2010, 2011, 2012, 2013 by
 *	University of New Hampshire Interoperability Laboratory
 *	121 Technology Drive, Suite 2
 *	Durham, NH 03824-4716
 *	phone: 1-603-862-0090
 *	fax: 1-603-862-4181
 *	e-mail: info@iol.unh.edu
 * All rights reserved.
 *
 * This file is part of the UNH EXS implementation of the
 * Consortium Specification: Extended Sockets API (ES-API) Issue 1.0,
 * Document Number C050 published by The Open Group, January 2005,
 * and copyright 2005 by The Open Group.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grant No. OCI-1127228.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

/*
 * The UNH IOL implementation of the Extended Sockets API (ES-API) Issue 1.0
 *
 * defined in:	The Interconnect Software Consortium Issue 1.0
 *		ISBN: 1-931624-52-6
 *		Published by The open Group, January 2005.
 *
 * all page references in this source code are to pages in that document.
 */

/*
 * Most of the coding in this file done by:
 *	Bob Russell
 *	rdr@iol.unh.edu
 */


/* This was implemented using OFED 1.5.4.1
 * OFED .h files stored in: /usr/include/rdma and /usr/include/infiniband
 *
 * In ofed library, int rdma_xxx() functions return 0 if ok, else -1 with errno set
 *			rdma_create_id()		checked
 *			rdma_destroy_id()		ok, return value never used
 *			rdma_bind_addr()		checked
 *			rdma_resolve_addr()		checked
 *			rdma_resolve_route()		checked
 *			rdma_create_qp()		checked
 *			rdma_connect()			checked
 *			rdma_listen()			checked
 *			rdma_accept()			checked
 *			rdma_reject()			checked
 *			rdma_migrate_id()		checked
 *			rdma_notify()			never called
 *			rdma_disconnect()		checked
 *			rdma_join_multicast()		never called
 *			rdma_leave_multicast()		never called
 *			rdma_ack_cm_event()		checked
 *			rdma_get_cm_event()		checked
 *	    the following return 0 if ok, else the errno value itself > 0
 *			ibv_dealloc_pd()		ok, return value never used
 *			ibv_destroy_comp_channel()	checked
 *			ibv_destroy_cq()		checked
 *			ibv_req_notify_cq()		checked
 *			ibv_dereg_mr()			checked
 *			ibv_post_recv()			checked
 *	    the following return 0 if ok, else -(errno value) < 0
 *			ibv_post_send()			checked
 *	    the following return non-NULL if ok, else NULL with errno set
 *			rdma_get_devices()		ok, we set errno = EAGAIN
 *			rdma_create_event_channel()	checked
 *	    the following return non-NULL if ok, else NULL on error
 *			(and we assume they set errno before return)
 *			ibv_alloc_pd()			checked
 *			ibv_create_comp_channel()	checked
 *			ibv_create_cq()			checked
 *			ibv_reg_mr()			checked
 *	    the following return nothing (i.e., void)
 *			rdma_free_devices()		ok, return value never used
 *	    the following return nothing (i.e., void)
 *			(but man page says errno set)
 *			rdma_destroy_event_channel()	ok, return value never used
 *			rdma_destroy_qp()		ok, return value never used
 *	    the following (are static inline uint16_t)
 *			and return != 0 ok, == 0 not connected/bound
 *			rdma_get_dst_port()		return value always used
 *			rdma_get_src_port()		return value always used
 *
 */

/*	We use the following ofed structures defined in
 *		/usr/include/infiniband/verbs.h
 *
 *	enumb ibv_wr_opcode {
 *			IBV_WR_RDMA_WRITE,
 *			IBV_WR_RDMA_WRITE_WITH_IMM,
 *			IBV_WR_SEND,
 *			IBV_WR_SEND_WITH_IMM,
 *			IBV_WR_RDMA_READ,
 *			IBV_WR_ATOMIC_CMP_AND_SWP,
 *			IBV_WR_ATOMIC_FETCH_AND_ADD
 *		};
 *
 *	enum ibv_send_flags {
 *			IBV_SEND_FENCE		= 1 << 0,
 *			IBV_SEND_SIGNALED	= 1 << 1,
 *			IBV_SEND_SOLICITED	= 1 << 2,
 *			IBV_SEND_INLINE		= 1 << 3
 *		};
 *
 *	struct ibv_sge {
 *			uint64_t		addr;
 *			uint32_t		length;
 *			uint32_t		lkey;
 *		};
 *
 *	struct ibv_send_wr {
 *			uint64_t		wr_id;
 *			struct ibv_send_wr	*next;
 *			struct ibv_sge		*sg_list;
 *			int			num_sge;
 *			enum ibv_wr_opcode	opcode;
 *			enum ibv_send_flags	send_flags;
 *			uint32_t		imm_data; network byte order
 *			union {
 *			} wr;
 *			uint32_t		xrc_remote_srq_num;
 *		};
 *
 *	struct ibv_recv_wr {
 *			uint64_t		wr_id;
 *			struct ibv_recv_wr	*next;
 *			struct ibv_sge		*sg_list;
 *			int			num_sge;
 *		};
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#define _POSIX_C_SOURCE 200112L
#define _ISOC99_SOURCE
#define _XOPEN_SOURCE 600

#include <limits.h>
#include <inttypes.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <semaphore.h>
#include <signal.h>
#include <sys/stat.h>
#include <time.h>
#include <byteswap.h>
#include <pthread.h>
#include <sys/mman.h>

/* /usr/include/rdma/rdma_cma.h includes:
 *	/usr/include/infiniband/verbs.h
 *	/usr/include/infiniband/sa.h
 */
#include <rdma/rdma_cma.h>

#include <evt_queue.h>
#include <our_calloc_free.h>
#include <affinity.h>
#include <atomic.h>

#if __BYTE_ORDER == __LITTLE_ENDIAN
static inline uint64_t htonll(uint64_t x) { return bswap_64(x); }
static inline uint64_t ntohll(uint64_t x) { return bswap_64(x); }
#else
static inline uint64_t htonll(uint64_t x) { return x; }
static inline uint64_t ntohll(uint64_t x) { return x; }
#endif


/* this is made available via the nonstandard exs_get_version() function */
static const char our_version[] = VERSION;

/* this is made available externally in exs.h,
 * to enable debugging, set to it nonzero
 * for now, default is disable debugging
 */
int exs_debug_enable = 0;

/* table of discovered maximum inline_data sizes
 *
 * 280 seems to be the max accepted by chelsio T4 iwarp cards
 * 460 seems to be the max accepted by Mellanox DDR InfiniBand cards
 * 912 seems to be the max accepted by Mellanox QDR InfiniBand cards
 *	any more than this and the rdma_create_qp() fails with ENOMEM
 * NOTE intel iwarp cards accept 280 or 460 but return 0 when queried
 *
 * NOTE: maximum is 256 for now because anything more decreases performance, at
 * least on the Mellanox QDR InfiniBand card that I was testing with.
 */
#define	KNOWN_INLINE_DATA_SIZE_LIMIT	4
static const uint32_t	inline_data_size[] = {912, 460, 280, 0};

/* table to print enum ibv_mtu values */
static const int	path_size[] = {0, 256, 512, 1024, 2048, 4096};

/* table to print exs_event_t values in exs_evt_type_to_string() */
static const char *exs_evt_type_strings[] = {
			"<illegal exs_evt_type>",	/*  0 */
			"EXS_EVT_POLL",			/*  1 */
			"EXS_EVT_CONNECT",		/*  2 */
			"EXS_EVT_ACCEPT",		/*  3 */
			"EXS_EVT_SEND",			/*  4 */
			"EXS_EVT_RECV",			/*  5 */
			"EXS_EVT_SENDMSG",		/*  6 */
			"EXS_EVT_RECVMSG",		/*  7 */
			"EXS_EVT_SENDFILE",		/*  8 */
			"EXS_EVT_SENDTIMEOUT",		/*  9 */
			"EXS_EVT_RECVTIMEOUT",		/* 10 */
			"EXS_EVT_CLOSE"			/* 11 */
};	/* exs_evt_type_strings */


enum our_message_header_types {
	OUR_ADVERTISEMENT = 1,	/* 1 */
	OUR_WRITE_NOTIFY = 2,	/* 2 */
	OUR_EOF = 3,		/* 3 */
	OUR_SETUP_REQUEST,	/* 4 */
	OUR_SETUP_REPLY,	/* 5 */
	OUR_ACK,		/* 6 */
};	/* enum our_message_header_types */

#define HEADER_TYPE_MASK	(0x07)

#define EXS_FD_ALL_FLAGS_MASK	(EXS_FD_SP_COPY_UNREG_ON_SEND	\
				| EXS_FD_SP_COPY_UNREG_ON_RECV	\
				| EXS_FD_BYPASS_PARAM_CHECKING	\
				| EXS_FD_NODELAY		\
				| EXS_FD_BUSYPOLL               \
				| EXS_FD_NO_WRITE_WITH_IMM      \
				)

enum conn_states {
	EXS_CONN_FINISHED,	/*  0 - after exs_close */
	EXS_CONN_UNOPENED,	/*  1 - at end of good calloc() */
	EXS_CONN_UNBOUND,	/*  2 - at end of good exs_socket() */
	EXS_CONN_BOUND,		/*  3 - at end of good exs_bind() */
	EXS_CONN_ADDRESSING,	/*  4 - at start of rdma_resolve_addr() */
	EXS_CONN_ROUTING,	/*  5 - at start of rdma_resolve_route() */
	EXS_CONN_SETUP,		/*  6 - at end of good our_setup_qp() */
	EXS_CONN_CONNECTING,	/*  7 - at start of rdma_connect() */
	EXS_CONN_RUNNING,	/*  8 - after good rdma_connect() */
	EXS_CONN_LISTENING,	/*  9 - at end of good exs_listen() */
	EXS_CONN_ACCEPTING,	/* 10 - after exs_accept and CONNECT_REQUEST */
	EXS_CONN_CLOSING,	/* 11 - at start of mainline rdma_disconnect() */
	EXS_CONN_TERMINATING,	/* 12 - remotely triggered rdma_disconnect() */
	EXS_CONN_UNLISTENING,	/* 13 - closed but owns still active accepts */
	EXS_CONN_CLOSED,	/* 14 - old state during exs_close() */
	EXS_CONN_CLIENT_SETUP,	/* 15 - during client setup exchange */
	EXS_CONN_AGENT_SETUP,	/* 16 - during agent setup exchange */
	EXS_CONN_DELAY_CLOSING	/* 17 - after exs_close with pending sends */
};	/* enum conn_states */


/* state transitions
 * Listener connection sequence:
 *		|
 *		|	exs_socket()
 *	       \|/
 *	EXS_CONN_UNOPENED
 *		|
 *		|
 *	       \|/
 *	EXS_CONN_UNBOUND
 *		|
 *		|	exs_bind()
 *	       \|/
 *	EXS_CONN_BOUND
 *		|
 *		|	exs_listen()
 *	       \|/
 *	EXS_CONN_LISTENING
 *		|
 *		|	exs_close()
 *	       \|/
 *	EXS_CONN_UNLISTENING
 *		|
 *		|
 *	       \|/
 *	EXS_CONN_FINISHED
 *
 * Agent connection sequence:
 *		|
 *		|	exs_accept()
 *	       \|/
 *	EXS_CONN_UNOPENED
 *		|
 *		|
 *	       \|/
 *	EXS_CONN_UNBOUND
 *		|
 *		|
 *	       \|/
 *	EXS_CONN_ACCEPTING
 *		|
 *		|
 *	       \|/
 *	EXS_CONN_ACCEPT_SETUP
 *		|
 *		|
 *	       \|/	  RDMA_CM_EVENT_DISCONNECTED
 *	EXS_CONN_RUNNING ----------------------------->	EXS_CONN_TERMINATING
 *	 /		\					|
 *	/		 \ exs_close()				|
 *     /		  \					|
 * EXS_CONN_CLOSING	 EXS_CONN_DELAY_CLOSING			|
 *		\	  /					|
 *		 \	 / RDMA_CM_EVENT_DISCONNECTED		|
 *		  \	/					|
 *		EXS_CONN_FINISHED <-----------------------------+
 *
 * Client connection sequence:
 *		|
 *		|	exs_socket()
 *	       \|/
 *	EXS_CONN_UNOPENED
 *		|
 *		|
 *	       \|/
 *	EXS_CONN_UNBOUND
 *		|
 *		|	exs_connect()
 *	       \|/
 *	EXS_CONN_ADDRESSING
 *		|
 *		|
 *	       \|/
 *	EXS_CONN_ROUTING
 *		|
 *		|
 *	       \|/
 *	EXS_CONN_SETUP
 *		|
 *		|
 *	       \|/
 *	EXS_CONN_CONNECTING
 *		|
 *		|
 *	       \|/
 *	EXS_CONN_CLIENT_SETUP
 *		|
 *		|
 *	       \|/	  RDMA_CM_EVENT_DISCONNECTED
 *	EXS_CONN_RUNNING ----------------------------->	EXS_CONN_TERMINATING
 *	 /		\					|
 *	/		 \ exs_close()				| exs_close()
 *     /		  \					|
 * EXS_CONN_CLOSING	 EXS_CONN_DELAY_CLOSING			|
 *		\	  /					|
 *		 \	 / RDMA_CM_EVENT_DISCONNECTED		|
 *		  \	/					|
 *		EXS_CONN_FINISHED <-----------------------------+
 *
 */

/* These next 4 values need to be understood and tuned better! */
#define	RESPONDER_RESOURCES	4	/* no. of RDMA_READs we respond to */
#define	INITIATOR_DEPTH		4	/* no. of RDMA_READs we initiate */
#define	RETRY_COUNT		5	/* no. of retries before failure */
#define	RNR_RETRY_COUNT		5	/* no. of Recv Not Ready retries */

/*	default posted struct transaction_block items in each direction
 *	used to initialize posted_credits field in cb
 *	user can reduce these by fcntl(EXS_F_SETFLOWCONTROLCREDITS)
 *	the value actually used is negotiated at connect/accept time
 */
#define DEFAULT_FLOW_CONTROL_CREDITS	32	/* initial credit limit */

/* maximum allocated struct transaction_block items per connection:
 *
 *	to satisfy exs_send, we need locally:
 *	posted_credits recv blocks to catch remote advertisement
 *	posted_credits send_wwi blocks to push data to receiver
 *
 *	to satisfy exs_recv, we need locally:
 *	posted_credits send_advert blocks to send our advertisement
 *	posted_credits recv blocks to handle WWI
 *	1/3 * posted_credits recv blocks to handle ACK (SOCK_STREAM only)
 *
 * Thus, in total, we need:
 * 	2 * posted_credits recv blocks
 * 	1 * posted_credits send_wwi blocks
 * 	1 * posted_credits send_advert blocks
 */

/*	maximum send queue depth -- 1 * credits ADVERT, 1.5 * credits DATA */
#define SQ_DEPTH(cb)	(5 * ((cb)->posted_credits) / 2)

/*	maximum recv queue depth -- 1 * credits ADVERT, 2 * credits DATA */
#define RQ_DEPTH(cb)	(3 * ((cb)->posted_credits))

/*	maximum completion queue depth -- SQ_DEPTH + RQ_DEPTH */
#define CQ_DEPTH(cb)	(11 * ((cb)->posted_credits) / 2)

#define EXS_LISTEN_BACKLOG 12


/* default small packet max size */
#define SMALL_PACKET_MAX_SIZE		0

/* maximum number of immediate bytes we allow as immediate data */
#define MAX_SMALL_PACKET_MAX_SIZE	(64 * 1024 * 1024)


/* timeout values in seconds */

/* timeout used in our_delay_for_flushing */
#define FLUSHING_TIMEOUT_SECS	1
#define FLUSHING_TIMEOUT_NSECS	250000000	/* quarter of a second */

/* block state bits
 *
 * Bits used:
 * SOCK_SEQPACKET exs_send(): SEND_READY|ADVERT_READY
 * SOCK_SEQPACKET exs_recv(): (none)
 * SOCK_STREAM EXS_NODELAY exs_send():
 *        SEND_READY|ADVERT_READY|SEND_DONE|ACK_DONE
 * SOCK_STREAM EXS_NODELAY exs_recv(): RECV_READY|RECV_DONE|ACK_DONE */
enum block_state_bits {
	SEND_READY_BIT = 1 << 0,
	ADVERT_READY_BIT = 1 << 1,
	SEND_DONE_BIT = 1 << 2,
	ACK_DONE_BIT = 1 << 3,
	RECV_READY_BIT = 1 << 4,
	RECV_DONE_BIT = 1 << 5,
	ADVERT_SENDING_BIT = 1 << 6,
	ACK_PENDING_BIT = 1 << 7,
};

enum recv_flag {
	RECV_COPY_UNREG_BIT = 1 << 0,
	RECV_FLAG_MASK = ((1 << 1) - 1),
};

enum imm_flag_bits {
	IMM_INDEX_MASK = 0x7fffffff,
	IMM_DIRECT_BIT = 0x80000000,
	IMM_FLAG_MASK = 0x80000000,
};

/* all fields in an advertisement are stored, sent in network byte order */
struct our_advert
{
	uint64_t	ad_buf_seqno;
	uint64_t	ad_buf_addr;
	uint32_t	ad_buf_key;
	uint32_t	ad_buf_len;
	uint32_t	ad_buf_index;
	uint32_t	ad_buf_flags;
};	/* struct our_advert */

struct our_notify
{
	uint32_t	write_len;
	uint32_t	write_index;
	uint32_t	write_flags;
};	/* struct our_notify */

struct our_stream_ack {
	uint32_t	recv_buf_empty;
	uint32_t	recv_buf_ack_count;
};	/* struct our_stream_ack */


/* structure of the negotiation request/reply message
 * used only during negotiation exchange
 *
 * is limited to 56 bytes on RDMA_PS_TCP on InfiniBand
 */
struct our_nego_mess
{
	uint32_t		client_credits;
	uint32_t		reserved0;
	uint32_t		sp_max_size;
	uint8_t			version;
	uint8_t			subversion;
	uint8_t			subsubversion;
	uint8_t			is_seqpacket;
	uint32_t		fd_flags;
	uint64_t		recv_buf_addr;
	uint32_t		recv_buf_len;
	uint32_t		recv_buf_rkey;
};	/* struct our_nego_mess */


/* hd_type is in network byte order for send headers, but is converted to
 *	host byte order immediately upon reception into recv headers.
 * hd_seqno is in network byte order in send headers, but is converted to
 *	host byte order immediately upon reception into recv headers.
 */
struct our_header
{
	uint32_t	hd_type;
	uint32_t	hd_seqno;
};	/* struct our_header */


/* internal flag bits for transactions
 * these must be compatible with user-level flags that are defined as
 * user-visible constants for exs_recv() and exs_send() in exs.h
 * so far, these user-visible constants are:
 *	EXS_DONTWAIT
 *	EXS_BLOCK
 *	EXS_UNSIGNALED
 * and all smaller system-defined constants)
 *
 * These flags are stored in block->buffer_flags
 */
#define EXS_NOTIFY_ALREADY_DONE		(0x100000)


struct stream_advert_block
{
	struct our_advert		advert_buf;
	unsigned int			offset;
	unsigned int			advert_state;
	struct stream_advert_block	*next;
};

struct transaction_block
{
	int			expected_comp;	/* expected completion opcode --
						 * necessary because wc->opcode
						 * NOT valid unless wc->status
						 * is IBV_WC_SUCCESS
						 */
	uint32_t		buffer_flags;	/* for in-place buf & recycle*/
	our_qhandle_t		qhandle;	/*queue to post done event on*/
	exs_mhandle_t		mhandle;	/*mhandle supplied by the user*/

	/* "new" stuff */
	unsigned int		in_progress;
	struct transaction_block*next;
	sem_t			sem;
	int			index;
	uint64_t		wr_id;
	unsigned int		send_state;
	unsigned int		outstanding_completion_count;
	unsigned int		outstanding_wwi_count;
	unsigned int		recv_flags;

	struct ibv_mr		*block_mr;	/*whole block RO memory region*/

	struct ibv_send_wr	send_wr;	/* send work request */
	struct ibv_sge		send_sg_list[2];/* 2-element send s/g list */
	struct our_header	send_header;	/* 2-word header on sends */
	size_t			send_total_bytes;/* total data bytes sent */

	/* data to be sent in small packets may be copied into a separately
	 * allocated preregistered buffer that we keep track of here */
	void			*send_buf;	/* points to send buf itself */
	struct ibv_mr		*send_buf_mr;	/* send_buf memory region reg*/

	/* data to be received in small packets may be copied from a separately
	 * allocated preregistered buffer that we keep track of here */
	void			*recv_rdma_buf;
	struct ibv_mr		*recv_rdma_mr;

	struct ibv_mr		*send_unreg_mr;	/* UNREG buffer memory region*/

	struct our_advert	advert_buf;	/* holds the advertisement */
	struct our_stream_ack	ack_buf;	/* holds the SOCK_STREAM ack */

	struct ibv_recv_wr	recv_wr;	/* receive work request */
	struct ibv_sge		recv_sg_list[1];/* 1-element recv s/g list */
	struct exs_mhandle	*recv_mhandle;	/* recv_header/buf mem region*/
	volatile size_t		buf_offset;	/*offset to unused byt in buf*/
	size_t			recv_byte_len;  /* no. of bytes given to recv*/
	struct ibv_mr		*recv_unreg_mr;	/* UNREG buffer memory region*/

	/* fields used to set up work requests for RDMA WRITE requests */
	struct ibv_send_wr	writereq_wr;	/* write request work request */
	struct ibv_sge		writereq_sg_list[1];/* write request s/g list */

	struct ibv_send_wr	writeack_wr;	/* iWARP post-write ack */
	struct ibv_sge		writeack_sg_list[2];
	struct our_header	writeack_header;
	struct our_notify	writeack_buf;

	struct exs_event	*user_event_ptr;/* user-level completion info*/

	/* the next 2 fields must be contiguous, and must be at the end
	 * of "struct transaction_block" because we will receive untagged
	 * send packets directly into here off the wire, and any immediate
	 * data in those send packets immediately follows the header off
	 * the wire into this memory.  The size of the memory to catch this
	 * immediate data is added to the size of "struct transaction_block"
	 * when "struct transaction_block" is dynamically allocated, so that
	 * the "recv_buf" field will be big enough for all immediate data.
	 * This additional size is given by "cb->small_packet_max_size",
	 * and can be set by the user with the SMALL_PACKET_MAX_SIZE attribute.
	 */
	struct our_header	recv_header;	/* 2-word header on all recvs*/
	char			recv_buf[sizeof(struct our_nego_mess)];
};	/* struct transaction_block */

/* This pending_accept structure contains parameters and pointers to
 * allocated memory for pending accepts.
 *
 * These structures are created dynamically in exs_accept(), where
 * they are threaded onto the "pending_accept_list" of the listening
 * socket passed in as a parameter to exs_accept().
 *
 * When our_con_man_thread() gets an RDMA_CM_EVENT_CONNECT_REQUEST on
 * a listening socket, it calls our_merge_pending_accept() which
 * tries to remove one of these structures from the listening socket's
 * "pending_accept_list" and to initialize lots of stuff for the new
 * connection before calling rdma_accept() to actually accept the connection.
 *
 * If the "pending_accept_list" is empty, we instead store the event's cm_id
 * in a new (or recyled) struct conn_requested, add that struct onto
 * the listening socket's conn_requested_list, and then increment the listening
 * socket's conn_requested_backlog count.
 *
 * When our_con_man_thread() gets an RDMA_CM_EVENT_ESTABLISHED on
 * an agent socket, it calls our_finish_pending_accept() which
 * adds the embedded (now accepted) connection * to the master
 * "conn_list", posts the EXS_EVT_ACCEPT event and finally
 * frees this pending_accept structure.
 *
 * All accesses to the listening socket's
 *	pending_accept_list
 *	conn_requested_list
 *	conn_requested_free_list
 * are  protected by that listening socket's "event_mutex".
 */
struct pending_accept {
	struct exs_conn		*accepted_conn;
	struct exs_acceptaddr	*addrvec;
	struct exs_event	*evt_item;
	struct our_nego_mess	nego_send;
	our_qhandle_t		qhandle;
	int			err_numb;
};	/* struct pending_accept */

struct conn_requested {
	struct rdma_cm_event	cm_event;
	struct our_nego_mess	nego_recv;
};	/* struct conn_requested */

/* internal flags for conn_flags field in struct exs_conn */

/* turned on after completely successful EXS_ACCEPT or EXS_CONNECT
 * turned off on user call to exs_start_close()
 * tested in our_completion_thread()
 *	to prevent start of new EXS I/O
 */
#define EXS_NEED_LOCAL_CLOSE	(0x10000)

/* turned on after completely successful EXS_ACCEPT or EXS_CONNECT
 * turned off upon con_man's receipt of RDMA_CM_EVENT_DISCONNECTED,
 * also turned off just before calling rdma_disconnect()
 * tested before each call to ibv_post_send() and ibv_post_recv()
 *	to prevent start of new OFED I/O
 */
#define EXS_NEED_DISCONNECT	(0x20000)

/* turned on after completely successful EXS_ACCEPT or EXS_CONNECT
 * turned off on user call to exs_start_close()
 * tested in exs_recv(), exs_send(), exs_sendfile()
 *	to ensure start of new EXS I/O is allowed
 */
#define EXS_OK_TO_ISSUE_IO	(0x40000)

/* turned on after our_completion_thread is up and joinable
 * turned off by our_completion_thread before it calls pthread_exit()
 * turned off by our_con_man_thread before it calls pthread_cancel()
 * tested in our_con_man_thread before calling pthread_cancel()
 */
#define EXS_NEED_COMP_CANCEL	(0x80000)

#define EXS_ALL_CONN_FLAGS	(0xf0000)


enum disconnect_flag_values {
	DISCONNECT_NONE = 0,
	DISCONNECT_EVENT = 1,
	DISCONNECT_CLOSE_CONN = 2,
	DISCONNECT_FLUSH_ERR = 3,
	DISCONNECT_LISTEN = 4
};

typedef int (*exs_queue_send_function)(struct exs_conn *this_conn,
	struct transaction_block *this, const void *buffer, size_t length,
	our_qhandle_t *qptr, int *signo);
typedef int (*xmit_if_match_function)(struct exs_conn *this_conn,
			struct transaction_block *block,
			our_qhandle_t *qhandle, int *signo,
			const char *message);
typedef int (*handle_recv_advert_function)(struct exs_conn *cb,
				struct transaction_block *block,
				struct ibv_wc *wc,
				char *errbuf, const char *message,
				int *do_eof_broadcast, int *force_exit,
				our_qhandle_t *qhandle, int *signo);
typedef int (*handle_send_write_function)(struct exs_conn *cb,
				struct transaction_block *block,
				const char *message,
				our_qhandle_t *qhandle, int *signo);
typedef int (*handle_recv_write_function)(struct exs_conn *cb,
				uint32_t byte_len,
				uint32_t imm_data, uint32_t imm_flags,
				our_qhandle_t *qhandle, int *signo);


enum { EXS_MAX_STREAM_BUF_SIZE = 3000000000L, };

struct exs_buf {
	void		*ptr;
	size_t		size;
	struct ibv_mr	*mr;
};

struct exs_stream_send_buf {
	/* Invariants:
	 *  n_full + n_empty == size
	 * Guarded by cb->stream_sync_mutex
	 */

	struct exs_buf	buf;

	/* new send data is copied here */
	unsigned int	send_offset;

	/* send is complete up to here */
	unsigned int	done_offset;

	/* WWI starting from here */
	unsigned int	take_offset;

	unsigned int	n_full;
	unsigned int	n_empty;
	unsigned int	n_remote; /* empty bytes in receiver buffer */
};

struct exs_stream_recv_buf {
	/* Invariants:
	 *  n_full <= size
	 * Guarded by cb->stream_sync_mutex
	 */

	struct exs_buf	buf;

	/* data copied to application from here */
	unsigned int	take_offset;

	/* remote WWI from recv_offset, which is tracked only on sending side */

	unsigned int	n_full;
	unsigned int	n_empty;
	unsigned int	n_empty_threshold;
};

/* there are 6 pthread condition variables defined in struct exs_conn and 1
 * in struct evt_queue.  These all MUST be paired with a mutex in that same
 * structure that is locked when testing for the condition to be true, and
 * is unlocked automatically by the pthread_cond_wait() when waiting for a
 * signal.  In addition, most of these are also paired with an int counter
 * in the same struct that is positive when someone is waiting on the
 * pthread_cond_wait().  This counter is used by the signaller to know
 * whether or not issue the pthread_cond_broadcast().
 *
 *	pthread_cond_t		mutex_t		int counter
 *	----------------	-----------	----------------
 * struct exs_conn
 *	transmit_ok_notify	event_mutex	---
 * struct evt_queue
 *	full_notify		event_list_lock	waiting_for_notify
 *
 * We utilize this information in our_con_man_thread so that after it
 * detects and processes a RDMA_CM_EVENT_DISCONNECTED event, it will do a
 * broadcast on all condition variables whose counters are not zero.
 * This requires the condition test around the pthread_cond_wait() to test
 * for some disconnect condition as part of getting out of the wait loop.
 */

struct exs_conn
{
	struct transaction_block	*send_advert_array;
	struct transaction_block	*send_wwi_array;
	struct transaction_block	*recv_array;
	struct stream_advert_block	*recved_advert_array;

	struct ibv_mr			*send_advert_array_mr;
	struct ibv_mr			*send_wwi_array_mr;
	struct ibv_mr			*recv_array_mr;

	volatile unsigned int		pending_recv_count;
	struct transaction_block	*next_recv;
	struct transaction_block	*next_recv_primed;

	volatile unsigned int		pending_send_count;	/*all waiting*/
	struct transaction_block	*next_send;
	struct transaction_block	*next_send_primed;
	struct transaction_block	*next_send_unacked;

	struct stream_advert_block	*next_advert;
	struct stream_advert_block	*next_advert_primed;

	pthread_mutex_t			advert_sequence_mutex;
	pthread_mutex_t			data_sequence_mutex;
	pthread_mutex_t			stream_sync_mutex;

	volatile unsigned int		advert_signal_count;
	volatile unsigned int		data_signal_count;
	unsigned int			signal_limit;

	uint32_t			inline_limit;

	unsigned int			send_state;

	exs_queue_send_function		enqueue_send;
	xmit_if_match_function		xmit_if_match;
	handle_recv_advert_function	handle_recv_advert;
	handle_send_write_function	handle_send_write;
	handle_recv_write_function	handle_recv_write_notify;

	unsigned int			disconnect_state;
	unsigned int			disconnect_flag;

	unsigned int		posted_credits; /* limit on exs_send blocks*/
	TRACE_VAR_DECLARE(int, max_posted_send_blocks)
	TRACE_VAR_DECLARE(int, max_posted_recv_blocks)
	struct pending_connect	*conn_pending;	  /*to post completion events*/

	pthread_cond_t		transmit_ok_notify;  /* start completn thread*/

	struct pending_accept	*accept_pending;  /*to post completion events*/
	struct our_list_head	pending_accept_list; /* connections to accept*/
	struct exs_conn		*accept_parent;	  /* for agent, back pointer
						   * to listener "parent"; NULL
						   * for client, listener
						   */

	struct our_list_head	conn_requested_list; /*unmatched conn request*/
	struct our_list_head	conn_requested_free_list;
	int			conn_requested_backlog;	/* number on list */
	int			backlog_limit;		/* from exs_listen() */

	struct our_list_head	queue_cache_list;  /* struct evt_queue cache */

	struct rdma_cm_id	*cm_id;
	struct ibv_pd		*pd;
	int			sockno;	/* socket this connection represents */
	volatile enum conn_states	conn_state;	/* state of the connection */
	int			conn_flags;

	/* set to non-zero after a call to our_alloc_socket() when type of
	 * newly allocated socket is SOCK_SEQPACKET
	 */
	uint8_t			is_seqpacket;

	/* set by our_completion_thread when it is up and running
	 * initially clear
	 * protected by event_mutex while waiting in
	 * our_create_completion_thread
	 * on wait condition is transmit_ok_notify
	 */
	uint8_t			comp_thread_needs_joining;

	/* set by our_con_man_thread when it is up and running
	 * initially clear
	 * protected by event_mutex while waiting in create_our_con_man_thread
	 * on wait condition is transmit_ok_notify
	 */
	uint8_t			con_man_thread_is_active;

	uint8_t			filler_bytes[3];

	/* requested queue pair capabilities,
	 * and those actually obtained by successful rdma_create_qp()
	 */
	struct ibv_qp_cap	requested_cap;
	struct ibv_qp_cap	obtained_cap;

	int			requested_cq_nelements;
	int			obtained_cq_nelements;

	int			obtained_path_mtu_size;

	uint64_t		recv_seqno;	/* seq no of last data recv */
	uint64_t		source_seqno;	/* seq no of next data recv */
	uint64_t		send_seqno;	/* seq no of next data send */

	struct rdma_event_channel *event_channel;
	enum rdma_cm_event_type	event_expected;	/* to drive asynch sequence */
	pthread_mutex_t		event_mutex;	/* protects cm event CSs */
	pthread_t		event_thread;	/* processes asynch cm events*/

	struct ibv_cq		*cq;	/* 1 completion queue per connection */
	struct ibv_qp		*qp;	/* 1 queue pair per connection */
	struct ibv_comp_channel *comp_channel;
	struct ibv_wc		*wc_array; /* holds polled wcs */
	pthread_t		comp_thread;	/* processes completions */
	int			comp_thread_cpu; /*cpu binding for comp thread*/
	int			max_n_wc_slots;	/* no. of slots in wc_array */
	unsigned int		n_cq_events_that_need_ack;

	/* attributes that can be configured by a user */
	uint32_t		small_packet_max_size;
	size_t			stream_buf_size;
	uint32_t		fd_flags;	/* used by GETFD/SETFD */

	struct exs_stream_recv_buf	recv_buf;
	struct our_advert		remote_recv_buf;
	unsigned int			n_unacked_wwi;
	unsigned int			n_unacked_wwi_threshold;
	unsigned int			remote_buf_n_empty;
	unsigned int			remote_buf_offset;
	unsigned int			ack_seqno;

	struct our_nego_mess		nego_send;

	/* pointers to extra memory used only during initialization */
	struct transaction_block *nego_in_ptr, *nego_out_ptr;
};	/* struct exs_conn */


/* open_socket_table is the master hash array of all known "struct exs_conn"
 * connections of any type in any state, indexed by a hash of the fd
 *
 *	the table is initialized so that all entries are empty lists by calling
 *		our_init_open_socket_table()
 *	a connection is added to the end of conn_list by calling
 *		our_add_socket_to_list()
 *	a connection is removed from anywhere on conn_list by calling
 *		our_rem_socket_from_list()
 *	a connection with sockno equal to a given sockno is found by calling
 *		our_lookup_socket()
 */

/* use low-order 10 bits of the fd as the hash index into open_socket_table */
#define	OPEN_SOCKET_TABLE_POWER	10
#define OPEN_SOCKET_TABLE_SIZE	(1 << OPEN_SOCKET_TABLE_POWER)
#define OPEN_SOCKET_TABLE_MASK	(OPEN_SOCKET_TABLE_SIZE - 1)

struct hash_list {
	int sockno;
	struct exs_conn *conn;
	struct hash_list *next;
};

/* Hash table of open sockets indexed by (fd & OPEN_SOCKET_TABLE_MASK) */
static struct hash_list *open_socket_table[OPEN_SOCKET_TABLE_SIZE];


/* we define one protection domain (pd) per RDMA device on this system.
 * we assume that the number of RDMA devices will be small (in fact,
 * usually 1, and for a dual-ported system it will be 2).  Therefore, we use
 * a global fixed-size table, rather than a linked list, to store these pds.
 */
#define	MAX_PD_SLOTS	8

/* protects global protection domain table
 * is initialized in our_init_pd_table(), called from exs_init()
 */
static pthread_rwlock_t	pd_table_rwlock;

/* Table of protection domains, in order of creation.
 * Empty table slots are initialized to NULL.
 * New domains are created in the first NULL slot, starting at slot 0.
 * Searches start at slot 0 and continue until a slot value is NULL.
 * (Therefore, the table can only contain MAX_PD_SLOTS-1 actual pds.)
 *
 * Table entries are created (if needed) by our_get_pd(), called in 2 places:
 * 1.	exs_init() - for all the devices known at initialization time
 * 2.	our_setup_qp() - for each new connection
 *
 */

static struct ibv_pd *pd_table[MAX_PD_SLOTS] = {NULL, };


/* collect all forward declarations here at the beginning */
#include "forward_exs.h"


/* start declaring function bodies here */

/*
 *	exs_evt_type_to_string()
 *			- not in Interconnect Software Consortium Issue 1.0
 *		      was added by us
 *
 *	returns pointer to printable string for each exs_event_t value
 */
EXS_EXPORT const char *
exs_evt_type_to_string(int value)
{
	if (value < 0 || value > EXS_EVT_CLOSE)
		value = 0;
	return exs_evt_type_strings[value];
}	/* exs_evt_type_to_string */


/*
 * called from exs_accept(), exs_connect(), exs_send(), and exs_recv()
 *
 * gets pointer to next struct evt_queue on cb's queue_cache_list
 * if list is empty, tries to create a new struct evt_queue
 *
 * Returns pointer if struct exists or created ok, else NULL
 */
static inline struct evt_queue *
our_get_from_queue_cache(struct exs_conn *cb)
{
	our_qhandle_t		local_qhandle;

	local_qhandle = exs_p_list_del_head(&cb->queue_cache_list);
	if (!local_qhandle) {
		local_qhandle = _evt_queue_init(1);
		/* if couldn't create, local_qhandle NULL, errno already set */
	}

	return local_qhandle;		/* use our local qhandle */
}	/* our_get_from_queue_cache */


/*
 * called from exs_accept(), exs_connect(), exs_send(), and exs_recv()
 *
 * threads struct evt_queue pointed to by local_qhandle
 * onto end of cb's queue_cache_list
 */
static inline void
our_put_to_queue_cache(struct exs_conn *cb, struct evt_queue *local_qhandle)
{
	exs_p_list_add_tail(&cb->queue_cache_list, local_qhandle);
}	/* our_put_to_queue_cache */


/*
 * called from exs_accept(), our_close_accepting(),
 *	and our_undo_exs_socket()
 *
 * undo the allocations and initializations done by our_create_queue_cache()
 *	frees up all items on cb's queue_cache_list
 *	then destroys cb's queue_cache_mutex
 */
static void
our_destroy_queue_cache(struct exs_conn *cb, const char *message)
{
	int			j;
	struct evt_queue	*local_qhandle;
	(void)message;

	TRACE("\t\t Free all struct evt_queue on connection %p\n", cb);

	j = 0;
	while ((local_qhandle = exs_p_list_del_head(&cb->queue_cache_list))) {
		_evt_queue_deinit(local_qhandle);
		j++;
	}
	exs_p_list_destroy(&cb->queue_cache_list);

	TRACE("%s %s: Freed %d struct evt_queue from connection %p\n",
		_time_prefix(), message, j, cb);
}	/* our_destroy_queue_cache */


/*
 * called only from our_alloc_socket()
 *
 * initializes cb's queue_cache_list and cb's queue_cache_mutex
 *
 * returns 0 if all ok, else -1 on any error
 */
static int
our_create_queue_cache(struct exs_conn *cb, const char *message)
{
	int	ret;
	char	errbuf[_MAX_STRERROR_MALLOC];

	ret = exs_p_list_init(&cb->queue_cache_list, 0);
	if(ret)
	{
		if (strerror_r(ret, errbuf, _MAX_STRERROR_MALLOC)) {
			snprintf(errbuf, _MAX_STRERROR_MALLOC, "unknown");
		}
		TRACE_ERR("%s: queue_cache_list init failed, error %d: %s",
				message, ret, errbuf);
		errno = ret;
		ret = -1;
	}
	return ret;
}	/* our_create_queue_cache */


/*
 * called only from our_alloc_socket()
 */
static void
our_init_pending(struct exs_conn *this_conn)
{
	/* init list to track exs_accepts on this listening socket*/
	exs_p_list_init(&this_conn->pending_accept_list, 0);

	/* init 2 lists to track connect_request events (handled in
	 * our_con_man_thread()) with no pending accepts to match
	 */
	exs_p_list_init(&this_conn->conn_requested_list, 0);
	exs_p_list_init(&this_conn->conn_requested_free_list, 0);
}	/* our_init_pending */


/*
 * called from exs_init()
 *
 * initializes the global rwlock for the global pd_table
 * (the table itself is statically initialized to NULL)
 */
static inline void
our_init_pd_table(void)
{
	pthread_rwlock_init(&pd_table_rwlock, NULL);
}	/* our_init_socket_table */


/* finds the global protection domain (pd) for this context.
 * In most cases, this will already be defined.
 * However, if it is not defined, we will define it here
 * and add it to the global table.
 *
 * Normally, need pd's are created and added only when
 * this function is called from exs_init().
 *
 * Once added to this table, pd's are never deleted.
 */
static struct ibv_pd *
our_get_pd(struct ibv_context *context)
{
	struct ibv_pd	**ptr;

	TRACE("\t\t%4d:   RDLOCK pd_table_rwlock\n", __LINE__);
	pthread_rwlock_rdlock(&pd_table_rwlock);
	TRACE("\t\t%4d:     PAST pd_table_rwlock\n", __LINE__);

	for (ptr = &pd_table[0]; *ptr != NULL; ptr++) {
		if ((*ptr)->context == context)
			break;
	}

	TRACE("\t\t%4d:   UNLOCK pd_table_rwlock\n", __LINE__);
	pthread_rwlock_unlock(&pd_table_rwlock);

	if (*ptr != NULL) {
		TRACE("%s %s: returning pd %p\n",
			_time_prefix(), __func__, *ptr);
		return *ptr;
	}

	/* the pd for this context does not yet exist,
	 * create it if there is room in the table
	 */

	/* do the search again, this time with write locking so we
	 * have exclusive access to this table
	 */
	TRACE("\t\t%4d:   RDLOCK pd_table_rwlock\n", __LINE__);
	pthread_rwlock_wrlock(&pd_table_rwlock);
	TRACE("\t\t%4d:     PAST pd_table_rwlock\n", __LINE__);

	for (ptr = &pd_table[0]; *ptr != NULL; ptr++) {
		if ((*ptr)->context == context)
			break;
	}

	if (*ptr == NULL) {
		/* this context is still not in the table, try to add it */
		if (ptr >= &pd_table[MAX_PD_SLOTS-1]) {
			/* no room in the table for another pd */
			TRACE_ERR("global pd_table full, max of %d protection domains",
				MAX_PD_SLOTS - 1);
			errno = ENOMEM;
		} else {/* create a protection domain for this context and
			 * store it in this slot */
			*ptr = ibv_alloc_pd(context);
			if(*ptr == NULL) {
				OUR_PERROR("ibv_alloc_pd failed", errno);
			} else {
				TRACE("%s \tcreated ibv pd %p\n",
					_time_prefix(), *ptr);
			}
		}
	}

	TRACE("\t\t%4d:   UNLOCK pd_table_rwlock\n", __LINE__);
	pthread_rwlock_unlock(&pd_table_rwlock);

	TRACE("%s %s: returning pd %p\n",
			_time_prefix(), __func__, *ptr);
	return *ptr;
}	/* our_get_pd */


/*
 * so far called only by exs_init() in error situation
 *
 * deallocates all pd's in the global pd_table
 * and destroys the global rwlock for the global pd_table
 */
void
our_destroy_pd_table(void)
{
	struct ibv_pd	**ptr;

	TRACE("\t\t%4d:   RDLOCK pd_table_rwlock\n", __LINE__);
	pthread_rwlock_wrlock(&pd_table_rwlock);
	TRACE("\t\t%4d:     PAST pd_table_rwlock\n", __LINE__);

	for (ptr = &pd_table[0]; *ptr != NULL; ptr++) {
		ibv_dealloc_pd(*ptr);
		*ptr = NULL;
	}

	TRACE("\t\t%4d:   UNLOCK pd_table_rwlock\n", __LINE__);
	pthread_rwlock_unlock(&pd_table_rwlock);

	pthread_rwlock_destroy(&pd_table_rwlock);
}	/* our_init_socket_table */


/*
 * called from:
 *	exs_accept() when a new accept is setup for an agent socket
 *	exs_socket() when user creates a new listening or client socket
 *
 * allocates a new struct exs_conn, and partially initializes it:
 *	(any fields not explicitly set are initialized to 0 by default)
 *	allocates a new struct pending_connect and points conn_pending at it
 *	creates the queue_cache_list and queue_cache_mutex
 *
 * this is the ONLY place where a struct exs_conn is allocated!
 *
 * returns a pointer to a newly allocated struct exs_conn for this socket
 * or NULL if can't get memory for this struct
 */
static struct exs_conn *
our_alloc_socket( const char *message )
{
	struct exs_conn	*new_conn;
	char		errbuf[_MAX_STRERROR_MALLOC];

	/* allocate a new, zeroed-out struct exs_conn */
	new_conn = (struct exs_conn*)EXS_CALLOC(1, sizeof(struct exs_conn));
	if(new_conn == NULL)
	{	/* not enough memory for a new connection */
		if (strerror_r(ENOMEM, errbuf, _MAX_STRERROR_MALLOC)) {
			snprintf(errbuf, _MAX_STRERROR_MALLOC, "unknown");
		}
		TRACE_ERR("%s: calloc of %zu bytes for struct exs_conn failed: %s",
			message, sizeof(struct exs_conn), errbuf);
		errno = ENOMEM;
		goto err0;
	}

	/* initialize thread, state, flags */
	new_conn->conn_state = EXS_CONN_UNOPENED;
	new_conn->conn_flags = EXS_NEED_LOCAL_CLOSE;

	new_conn->conn_pending = (struct pending_connect *)
				EXS_CALLOC(1, sizeof(struct pending_connect));
	if(new_conn->conn_pending == NULL)
	{
		/* can't come up with enough memory */
		if (strerror_r(ENOMEM, errbuf, _MAX_STRERROR_MALLOC)) {
			snprintf(errbuf, _MAX_STRERROR_MALLOC, "unknown");
		}
		TRACE_ERR("%s: calloc of %zu bytes for struct pending_connect failed: %s",
			message, sizeof(struct pending_connect), errbuf);
		errno = ENOMEM;
		goto err1;
	}

	if(our_create_queue_cache(new_conn, message))
		goto err2;

	/* set up some empty lists that are actually used only by
	 * listening sockets
	 */
	our_init_pending(new_conn);

	new_conn->inline_limit = INT_MAX;
	new_conn->stream_buf_size = EXS_MAX_STREAM_BUF_SIZE;

	TRACE("%s %s: ALLOCATE %lu bytes for struct exs_conn %p\n",
		_time_prefix(), message,
		(long unsigned int)sizeof(*new_conn), new_conn);

	TRACE("%s %s: ALLOCATE %lu bytes for struct pending_connect %p"
		"\n", _time_prefix(), message,
		(long unsigned int)sizeof(struct pending_connect),
		new_conn->conn_pending);
	goto err0;

err2:
	EXS_FREE(new_conn->conn_pending);
err1:
	EXS_FREE(new_conn);
	new_conn = NULL;
err0:
	return new_conn;
}	/* our_alloc_socket */


/* undo the allocations and initializations taken by our_alloc_socket() */
void
our_unalloc_socket(struct exs_conn *cb, const char *message)
{
	our_free_all_pending_lists(cb, message);
	our_destroy_queue_cache(cb, message);
	EXS_FREE(cb->conn_pending);
	EXS_FREE(cb);
}	/* our_unalloc_socket */


/*
 * called from exs_init()
 *
 * initializes the open_socket_table so all entries are empty lists
 */
static void
our_init_socket_table(void)
{
	memset(open_socket_table, 0, sizeof(open_socket_table));
}	/* our_init_socket_table */


/*
 * called from our_finish_pending_accept() by an agent,
 *	and exs_socket() by a client or a listener
 *
 * adds a connection to end of conn_list
 *
 * returns 0 on success; errno on failure
 */
static int
our_add_socket_to_list(struct exs_conn *this_conn)
{
	unsigned int	index;
	struct hash_list *ptr;

	index = this_conn->sockno & OPEN_SOCKET_TABLE_MASK;

	for (ptr = open_socket_table[index]; ptr != NULL; ptr = ptr->next) {
		if ((int)atomic_swap_if_match(this_conn->sockno, -1,
					(unsigned int *)&ptr->sockno) == -1) {
			TRACE("%s %s Replace tombstone at %p with fd %d at index %d\n",
					_time_prefix(), __func__, ptr,
					this_conn->sockno, index);
			ptr->conn = this_conn;
			return 0;
		}
	}
	if (!(ptr = malloc(sizeof(*ptr)))) {
		return errno;
	}
	ptr->sockno = this_conn->sockno;
	ptr->conn = this_conn;
	while (1) {
		ptr->next = open_socket_table[index];
		if (atomic_swap_if_match_ptr(ptr, ptr->next, 
				(volatile void **)&open_socket_table[index])
				== ptr->next) {
			TRACE("%s %s Add new node %p with fd %d at index %d\n",
					_time_prefix(), __func__, ptr,
					this_conn->sockno, index);
			break;
		}
	}

	return 0;
}	/* our_add_socket_to_list */


/* returns a pointer to the struct exs_conn that corresponds to this socket
 * or NULL if no such connection exists yet
 */
static inline struct exs_conn *
our_lookup_socket( int sockno )
{
	struct hash_list *ptr;
	struct exs_conn		*this_conn;
	unsigned int		index;

	index = sockno & OPEN_SOCKET_TABLE_MASK;
	this_conn = NULL;

	for (ptr = open_socket_table[index]; ptr != NULL; ptr = ptr->next) {
		if (ptr->sockno == sockno) {
			/* found the connection that matches this socket */
			this_conn = ptr->conn;
			break;
		}
	}
	/* normal loop finish if found no connection matching this socket */

	return this_conn;

}	/* our_lookup_socket */


/*
 * called from
 *	our_undo_exs_socket() from our_start_close() for listener
 *				state = LISTENING
 *			  and from our_close_connecting() from
 *			      our_con_man_thread for client
 *				state = CLOSING
 *	our_close_accepting() from our_start_close() for server agent
 *				state = DISACCEPTING
 *	4 times from our_start_close()
 *		for client when state = CLIENT_SETUP or CONNECTED
 *		for client when state = TERMINATING or CONNECTING
 *		for server agent when = AGENT_SETUP or ACCEPTED
 *		for listener in state = LISTENING
 *
 * removes a connection from anywhere on conn_list
 */
static void
our_rem_socket_from_list(struct exs_conn *old_conn)
{
	struct hash_list *ptr;
	unsigned int index;

	index = old_conn->sockno & OPEN_SOCKET_TABLE_MASK;

	for (ptr = open_socket_table[index]; ptr != NULL; ptr = ptr->next) {
		if ((int)atomic_swap_if_match(-1, old_conn->sockno,
					(unsigned int *)&ptr->sockno)
				== old_conn->sockno) {
			ptr->conn = NULL;
			return;
		}
	}
}	/* our_rem_socket_from_list */


static const char *state_labels[] =
{
	"EXS_CONN_FINISHED",	/*  0 - after exs_close */
	"EXS_CONN_UNOPENED",	/*  1 - at end of good calloc() */
	"EXS_CONN_UNBOUND",	/*  2 - at end of good exs_socket() */
	"EXS_CONN_BOUND",	/*  3 - at end of good exs_bind() */
	"EXS_CONN_ADDRESSING",	/*  4 - at start of rdma_resolve_addr() */
	"EXS_CONN_ROUTING",	/*  5 - at start of rdma_resolve_route() */
	"EXS_CONN_SETUP",	/*  6 - at end of good our_setup_qp() */
	"EXS_CONN_CONNECTING",	/*  7 - at start of rdma_connect() */
	"EXS_CONN_RUNNING",	/*  8 - after good rdma_connect() */
	"EXS_CONN_LISTENING",	/*  9 - at end of good exs_listen() */
	"EXS_CONN_ACCEPTING",	/* 10 - after exs_accept and CONNECT_REQUEST */
	"EXS_CONN_CLOSING",	/* 11 - at start of rdma_disconnect() */
	"EXS_CONN_TERMINATING",	/* 12 - remotely triggered rdma_disconnect() */
	"EXS_CONN_UNLISTENING",	/* 13 - closed but owns still active accepts */
	"EXS_CONN_CLOSED",	/* 14 - old state during exs_close() */
	"EXS_CONN_CLIENT_SETUP",/* 15 - during client nego exchange */
	"EXS_CONN_AGENT_SETUP",	/* 16 - during agent nego exchange */
	"EXS_CONN_DELAY_CLOSING"/* 17 - after exs_close with pending sends */
};


/* routine to print the values of "enum rdma_cm_event_type" as defined in
 * /usr/include/rdma/rdma_cma.h
 */
static void
our_print_event_type(const char *message1, const char *message2,
			enum rdma_cm_event_type event_type)
{
	const char *ptr;

	ptr = rdma_event_str(event_type);
	TRACE_ERR("%s: %s %s(%d)", message1, message2, ptr, event_type);
}	/* our_print_event_type */


#if defined(ALLOW_EXS_TRACE)
/* routine to print the values of "enum ibv_wc_status" as defined in
 * /usr/include/infiniband/verbs.h
 */
static void
our_pstatus(const char *message, enum ibv_wc_status status)
{
	TRACE_ERR("%s: %s", message, ibv_wc_status_str(status));
}	/* our_pstatus */
#endif /* defined(ALLOW_EXS_TRACE) */


/* routine to trace the values of "enum rdma_cm_event_type" as defined in
 * /usr/include/rdma/rdma_cma.h
 */
static void
our_trace_event_type(struct exs_conn *cb, const char *message1,
		     const char *message2, enum rdma_cm_event_type event_type)
{
	const char *ptr;

	ptr = rdma_event_str(event_type);
	TRACE_ERR("%s state %s(%d) %s(%d): %s",
		message1, state_labels[cb->conn_state], cb->conn_state,
		message2, event_type, ptr);
}	/* our_trace_event_type */



static socklen_t
our_copy_src_addr(struct rdma_cm_id *cm_id, struct sockaddr *name,
			socklen_t namelen)
{
	int		this_family;
	socklen_t	n;

	if(cm_id->route.addr.src_addr.sa_family != PF_INET6)
		{
		this_family = PF_INET;
		n = sizeof(struct sockaddr_in);
		}
	else
		{
		this_family = PF_INET6;
		n = sizeof(struct sockaddr_in6);
		}
	if(n > namelen)
		n = namelen;
	if(n > 2)
		{
		memcpy(name, &cm_id->route.addr.src_addr, n);
		name->sa_family = this_family;
		}
	return n;
}	/* our_copy_src_addr */


static socklen_t
our_copy_dst_addr(struct rdma_cm_id *cm_id, struct sockaddr *name,
			socklen_t namelen)
{
	int		this_family;
	socklen_t	n;

	if(cm_id->route.addr.dst_addr.sa_family != PF_INET6)
		{
		this_family = PF_INET;
		n = sizeof(struct sockaddr_in);
		}
	else
		{
		this_family = PF_INET6;
		n = sizeof(struct sockaddr_in6);
		}
	if(n > namelen)
		n = namelen;
	if(n > 2)
		{
		memcpy(name, &cm_id->route.addr.dst_addr, n);
		name->sa_family = this_family;
		}
	return n;
}	/* our_copy_dst_addr */


/*
 * called only from our_con_man_thread(), 4 times
 *
 * When called, event_mutex MUST be locked.
 */
static void
our_deal_with_bad_event(struct exs_conn *cb, struct rdma_cm_event *event,
			const char *name, struct evt_queue **qhandle,
			int *signo)
{
	int	ret = 0, level;
	char	errbuf[_MAX_STRERROR_MALLOC];

	our_trace_event_type(cb,name, "got unexpected cm event type", event->event);
	our_trace_event_type(cb,name, "expected cm event type", cb->event_expected);

	TRACE_ERR("%s: cb %p error code in event status = %d",
		name, cb, event->status);
	if(event->status < 0)
		{
		ret = -event->status;
		if (strerror_r(ret, errbuf, _MAX_STRERROR_MALLOC)) {
			snprintf(errbuf, _MAX_STRERROR_MALLOC, "unknown");
		}
		TRACE_ERR("%s: cb %p bad event status, conn_state %s(%d), error %d: %s",
			name, cb,
			state_labels[cb->conn_state], cb->conn_state,
			ret, errbuf);
		}

	switch(cb->event_expected)
	{

	case RDMA_CM_EVENT_ADDR_RESOLVED:
		/* EADDRNOTAVAIL, Cannot assign requested address */
	case RDMA_CM_EVENT_ROUTE_RESOLVED:
		/* EADDRNOTAVAIL, Cannot assign requested address ??? */
		if(ret == 0)
			ret = EADDRNOTAVAIL;
		level = 1;
		break;

	case RDMA_CM_EVENT_ESTABLISHED:
		if(ret == 0)
			{
			if(event->event == RDMA_CM_EVENT_UNREACHABLE)
				ret = ENETUNREACH;  /* Network is unreachable */
			else	/* event->event == RDMA_CM_EVENT_REJECTED or
				 * event->event == RDMA_CM_EVENT_CONNECT_ERROR
				 */
				ret = ECONNREFUSED; /*Connection reset by peer*/
			}
		level = 4;
		break;

	default:
		our_print_event_type(name, "can't handle event type",
								event->event);
		if(ret == 0)
			ret = EIO;		    /* Generic I/O error */
		level = 1;
		break;
	}	/* switch */

	our_connect_returning(cb, level, ret, qhandle, signo);

}	/* our_deal_with_bad_event */


static inline int
our_ack_cm_event(struct rdma_cm_event *event, const char *message)
{
	int ret;

	ret = rdma_ack_cm_event(event);
	if(ret)
	{
		ret = errno;
		TRACE_ERR("%s: rdma_ack_cm_event failed, error %d",
			message, ret);
	}
	return ret;
}	/* our_ack_cm_event */


/*
 * called from:
 *		our_start_close()   (on client close after remote DISCONNECTED)
 *		our_merge_pending_accept()	(on an error)
 *		our_connect_returning()		(on an error)
 *		our_close_accepting()		(on agent remote DISCONNECTED
 *						and agent local exs_close())
 *		our_close_connecting()		(on client local DISCONNECTED)
 *
 * purpose is to cancel our_completion_thread(), as the name implies
 *
 * when called, event_mutex MUST be locked
 * to prevent a race with our_completion_thread()
 */
static void
our_cancel_completion_thread( struct exs_conn *this_conn, const char *message )
{
	int	ret;
	char	errbuf[_MAX_STRERROR_MALLOC];

	/* give completion thread a chance to get the "last" event */
	sched_yield();

	TRACE("%s %s: CANCELing comp_thread %lu on conn %p, "
		"conn_state %s(%d), conn_flags %06x, needs_joining %u\n",
		_time_prefix(),
		message, this_conn->comp_thread, this_conn,
		state_labels[this_conn->conn_state],
		this_conn->conn_state, this_conn->conn_flags,
		this_conn->comp_thread_needs_joining);

	if (!(this_conn->conn_flags & EXS_NEED_COMP_CANCEL))
	{
		/* our_completion_thread did pthread_exit(), no need to cancel*/
		TRACE("%s %s: pthread_cancel of comp_thread %lu "
			"not done, conn %p, conn_state %s(%d), "
			"conn_flags %06x, needs_joining %u\n", _time_prefix(),
			message, this_conn->comp_thread, this_conn,
			state_labels[this_conn->conn_state],
			this_conn->conn_state, this_conn->conn_flags,
			this_conn->comp_thread_needs_joining);
	} else {
		if ((ret = pthread_cancel(this_conn->comp_thread))) {
			if (strerror_r(ret, errbuf, _MAX_STRERROR_MALLOC)) {
				snprintf(errbuf, _MAX_STRERROR_MALLOC, "unknown");
			}
			TRACE("%s %s: pthread_cancel of comp_thread %lu"
				" failed, conn %p, conn_state %s(%d), conn_flags %06x, "
				"needs_joining %u, error %d: %s\n",
				_time_prefix(), message, this_conn->comp_thread,
				this_conn, state_labels[this_conn->conn_state],
				this_conn->conn_state, this_conn->conn_flags,
				this_conn->comp_thread_needs_joining,
				ret, errbuf);
			/* ignore this error, it means we just lost a race to do
			 * it */
		} else {
			TRACE("%s %s: CANCELed  comp_thread %lu on conn "
				"%p\n", _time_prefix(), message,
				this_conn->comp_thread, this_conn);
		}

		/* ensure this is off */
		this_conn->conn_flags &= ~(EXS_NEED_COMP_CANCEL);

		TRACE("%s %s: joining comp_thread %lu on conn "
			"%p, conn_state %s(%d), conn_flags %0x, "
			"needs_joining %u\n",
			_time_prefix(), message, this_conn->comp_thread,
			this_conn, state_labels[this_conn->conn_state],
			this_conn->conn_state, this_conn->conn_flags,
			this_conn->comp_thread_needs_joining);
		ret = pthread_join(this_conn->comp_thread, NULL);
		if(ret)
		{
			if (strerror_r(ret, errbuf, _MAX_STRERROR_MALLOC)) {
				snprintf(errbuf,_MAX_STRERROR_MALLOC,"unknown");
			}
			TRACE_ERR("%s: pthread_join of comp_thread %lu failed, error %d: %s",
				message, this_conn->comp_thread, ret, errbuf);
			/* ignore this error for now, maybe try again later? */
			this_conn->comp_thread_needs_joining = 0xff;
		} else {
			TRACE("%s %s: joined comp_thread %lu on conn "
				"%p, conn_state %s(%d), conn_flags %0x\n",
				_time_prefix(), message, this_conn->comp_thread,
				this_conn, state_labels[this_conn->conn_state],
				this_conn->conn_state, this_conn->conn_flags);
		}
	}
}	/* our_cancel_completion_thread */


/* used only in our_con_man_cancel_handler and our_con_man_thread */
struct con_man_cancel_params {
	struct exs_conn		*cb;
	struct rdma_cm_event	*event;
};


/*
 * called only if we get canceled in our_completion_thread()
 * when waiting to get the mutex, which is explicitly NOT a cancelation point!
 */
static void
our_con_man_cancel_handler(void *arg)
{
	struct con_man_cancel_params	*ptr;

	ptr = arg;
	TRACE("%s [%lu] %s Leave %s\n",
		_time_prefix(), pthread_self(), "our_con_man_thread",
		__func__);

	/* acknowledge the event */
	our_ack_cm_event(ptr->event, __func__);

	pthread_mutex_unlock(&ptr->cb->event_mutex);
}


/*
 *	our_con_man_thread - private, handles rdma connection manager events.
 *
 *	repeatedly waits for a connection manager event,
 *	then	locks the event_mutex
 *		dispatches the event
 *		broadcasts the event_completion
 *		unlocks the event_mutex
 *
 *	The arg passed in at creation is a pointer to the cb of
 *	either a client connection or a listening connection.
 *	It is never the cb of an agent connection.
 *
 */
static void *
our_con_man_thread(void *arg)
{
	int				ret, signo;
	unsigned int			our_disconnect_state;
	volatile int			done, do_shutdown;
	struct exs_conn			*cb = arg;
	struct exs_conn			*volatile child_conn;
	struct rdma_cm_event		*event;
	enum rdma_cm_event_type		event_type, prev_event_type;
	struct evt_queue		*qhandle;
	pthread_cond_t			*to_broadcast;
	char				*broadcast_name = "";
	char				message[100];
	struct con_man_cancel_params	params;

	sprintf(message, "[%lu] %s for cb %p",
		pthread_self(), __func__, cb);

	TRACE("%s %s Enter\n",
		_time_prefix(), message);

	/* tell our creator that this con_man_thread has started */
	TRACE("\t\t%4d:     LOCK cb %p event_mutex %p\n",
		__LINE__, cb, &cb->event_mutex);
	pthread_mutex_lock(&cb->event_mutex);
	TRACE("\t\t%4d:     PAST cb %p event_mutex %p\n",
		__LINE__, cb, &cb->event_mutex);

	cb->con_man_thread_is_active = 0xff;
	pthread_cond_signal(&cb->transmit_ok_notify);

	TRACE("\t\t%4d:   UNLOCK cb %p event_mutex %p\n",
		__LINE__, cb, &cb->event_mutex);
	pthread_mutex_unlock(&cb->event_mutex);

	done = 0;
	child_conn = NULL;
	do_shutdown = 0;
	event_type = RDMA_CM_EVENT_MULTICAST_ERROR;
	for( ; ; )
	{
		pthread_testcancel();

		/* wait for next connection management event on this channel */
		TRACE("%s %s: about to get cm event for cb %p\n",
			_time_prefix(), message, cb);

		pthread_testcancel();

		/* for debugging, remember last event type */
		prev_event_type = event_type;

		if((ret = rdma_get_cm_event(cb->event_channel, &event)))
		{
			ret = errno;
			TRACE_ERR("%s: rdma_get_cm_event failed on cb %p, error %d",
				message, cb, ret);
			errno = ret;
			pthread_exit(NULL);
		}
		event_type = event->event;
		TRACE_EVENT_TYPE(cb, message, "got cm event type", event_type);
		TRACE_EVENT_TYPE(cb, message, "prev cm event type",
							prev_event_type);

		/* push a cancel handler that unlocks event_mutex
		 *	and acknowledges the event
		 */
		params.event = event;
		params.cb = cb;
		pthread_cleanup_push(our_con_man_cancel_handler, &params);

		TRACE("\t\t%4d:     LOCK cb %p event_mutex %p\n",
			__LINE__, cb, &cb->event_mutex);
		pthread_mutex_lock(&cb->event_mutex);
		TRACE("\t\t%4d:     PAST cb %p event_mutex %p\n",
			__LINE__, cb, &cb->event_mutex);

		signo = -1;
		qhandle = NULL;
		to_broadcast = NULL;
		our_disconnect_state = cb->disconnect_state;
		switch(event_type)
		{
		case RDMA_CM_EVENT_CONNECT_REQUEST:
			/* executed by listening connection only */
			if (our_merge_pending_accept(cb, event) != 0) {
				/* reject the connection an any error */
				if (rdma_reject(event->id, NULL, 0) != 0)
					OUR_PERROR("rdma_reject CONNECT_REQUEST",
							errno);
			}
			break;

		case RDMA_CM_EVENT_ESTABLISHED:
			/* executed by both client and agent connections */
			if(cb->conn_state == EXS_CONN_ACCEPTING)
			{	/* this is an agent connection */
				if(our_finish_pending_accept(cb) == 0) {
					/* everything connected ok, allow the
					 * completion thread to get started */
					 to_broadcast = &cb->transmit_ok_notify;
					 broadcast_name = "transmit_ok_notify";
				}
				/* TO DO -- deal with error result */
			}
			else if(cb->conn_state == EXS_CONN_CONNECTING)
			{	/* this is a client connection */
				if( our_connect_finishing(cb, event, &qhandle,
							&signo) == 0) {
					/* everything connected ok, allow the
					 * completion thread to get started */
					 to_broadcast = &cb->transmit_ok_notify;
					 broadcast_name = "transmit_ok_notify";
				}
				/* TO DO -- deal with error result */
			} else {/* this connection is in an unexpected state */
				our_deal_with_bad_event(cb, event, message,
							&qhandle, &signo);
			}
			break;

		case RDMA_CM_EVENT_ADDR_RESOLVED:
			if(cb->conn_state == EXS_CONN_ADDRESSING)
			{	/* this is a client connection */
				our_connect_routing(cb, &qhandle, &signo);
				/* TO DO -- deal with result */
			} else {/* this connection is in an unexpected state */
				our_deal_with_bad_event(cb, event, message,
							&qhandle, &signo);
			}
			break;

		case RDMA_CM_EVENT_ROUTE_RESOLVED:
			if(cb->conn_state == EXS_CONN_ROUTING)
			{	/* this is a client connection */
				our_connect_resolving(cb, &qhandle, &signo);
			} else {/* this connection is in an unexpected state */
				our_deal_with_bad_event(cb, event, message,
							&qhandle, &signo);
			}
			break;

		case RDMA_CM_EVENT_DISCONNECTED:
			/* executed by both client and agent connections */
			our_disconnect_state
				= atomic_set_bits(0x1, &cb->disconnect_state);

			TRACE("%s %s: RDMA_CM_EVENT_DISCONNECTED, "
					"connection %p in state %s(%d) "
					"conn_flags %0x\n",
					_time_prefix(), message, cb,
					state_labels[cb->conn_state],
					cb->conn_state, cb->conn_flags);

			if(cb->conn_state == EXS_CONN_CLOSING) {
				/* a previous exs_close() finishes
				 * asynchronously here
				 * on local or remote disconnect
				 * make sure no one finds this conn again!
				 */
				cb->conn_state = EXS_CONN_FINISHED;
				/* rest done in common below */
				child_conn = cb;
			} else if (cb->conn_state == EXS_CONN_DELAY_CLOSING) {
				TRACE_DBG(EXS_LOG_UNSPECIFIED,
					"RDMA_CM_EVENT_DISCONNECTED, connection %p in state %s conn_flags %#x",
					cb, state_labels[cb->conn_state],
					cb->conn_flags);

				/* for now, turn on debugging */
				/*****_our_debug_flag = 1; *****/

				cb->conn_state = EXS_CONN_FINISHED;
				child_conn = cb;

				/* perform a local rdma_disconnect() to provoke
				 * FLUSH_ERR status on all local posted recv blocks
				 * (completion thread will handle these)
				 */
				cb->conn_flags &= ~(EXS_NEED_DISCONNECT
							| EXS_OK_TO_ISSUE_IO);

				if (atomic_swap_if_match(DISCONNECT_EVENT,
						DISCONNECT_NONE,
						&cb->disconnect_flag)
							== DISCONNECT_NONE) {
					ret = our_rdma_disconnect(cb, message);
					TRACE("%s %s: back from our_rdma_disconnect, "
						"ret %d, errno %d, conn_state %s(%d), "
						"conn_flags %06x\n",
						_time_prefix(), message, ret, errno,
						state_labels[cb->conn_state],
						cb->conn_state, cb->conn_flags);
				}

				do_shutdown = 1;
			} else {
				/* this connection was hit by disconnect
				 * from remote side before it issued
				 * local exs_close()
				 */

#if defined(ALLOW_PARANOID_CHECKING)
				if(cb->conn_state != EXS_CONN_RUNNING)
				{	/* should never happen */
					TRACE_ERR("%s: unexpected RDMA_CM_EVENT_DISCONNECTED!",
						message);
				}
#endif
				TRACE("%s %s: RDMA_CM_EVENT_DISCONNECTED, "
					"connection %p in state %s(%d) "
					"conn_flags %0x\n",
					_time_prefix(), message, cb,
					state_labels[cb->conn_state],
					cb->conn_state, cb->conn_flags);

				/* for now, turn on debugging */
				/*****_our_debug_flag = 1; *****/

				cb->conn_state = EXS_CONN_TERMINATING;

				/* perform a local rdma_disconnect() to provoke
				 * FLUSH_ERR status on all local posted recv blocks
				 * (completion thread will handle these)
				 */
				cb->conn_flags &= ~(EXS_NEED_DISCONNECT
							| EXS_OK_TO_ISSUE_IO);

				if (atomic_swap_if_match(DISCONNECT_EVENT,
						DISCONNECT_NONE,
						&cb->disconnect_flag)
							== DISCONNECT_NONE) {
					ret = our_rdma_disconnect(cb, message);
					TRACE("%s %s: back from our_rdma_disconnect, "
						"ret %d, errno %d, conn_state %s(%d), "
						"conn_flags %06x\n",
						_time_prefix(), message, ret, errno,
						state_labels[cb->conn_state],
						cb->conn_state, cb->conn_flags);
				}

				do_shutdown = 1;
			}
			break;

		case RDMA_CM_EVENT_TIMEWAIT_EXIT:
			/* the QP associated with a connection has exited its
			 * timewait state and is now ready to be re-used.
			 * After a QP has been disconnected, it is maintained
			 * in a timewait state to allow any in flight packets
			 * to exit the network.
			 * After the timewait state has completed, the
			 * rdma_cm will report this event.
			 *
			 * My observations:
			 * 1. This only happens with IB interfaces, not iWARP,
			 *    and only for server agent connections.
			 * 2. This happens a few seconds after a server agent
			 *    is terminated
			 * 3. The event id appears to be that of the server
			 *    agent -- it is NOT that of the listener
			 */
			TRACE("%s %s: RDMA_CM_EVENT_TIMEWAIT_EXIT "
				"for cb %p, conn_state %s(%d), "
				"conn_flags %06x, cm_id %p, event id %p\n",
				_time_prefix(), message, cb,
				state_labels[cb->conn_state],
				cb->conn_state, cb->conn_flags,
				cb->cm_id, event->id);
			break;
		case RDMA_CM_EVENT_ADDR_ERROR:
		case RDMA_CM_EVENT_ROUTE_ERROR:
		case RDMA_CM_EVENT_CONNECT_ERROR:
		case RDMA_CM_EVENT_CONNECT_RESPONSE:
		case RDMA_CM_EVENT_UNREACHABLE:
		case RDMA_CM_EVENT_REJECTED:
		case RDMA_CM_EVENT_DEVICE_REMOVAL:
		case RDMA_CM_EVENT_MULTICAST_JOIN:
		case RDMA_CM_EVENT_MULTICAST_ERROR:
		case RDMA_CM_EVENT_ADDR_CHANGE:
			our_deal_with_bad_event(cb, event, message,
						&qhandle, &signo);
			break;
		}	/* switch */

		TRACE_EVENT_TYPE(cb, message, "acknowledging cm event type",
								event_type);
		ret = our_ack_cm_event(event, message);
		if(ret)
		{
			TRACE_ERR("%s: our_ack_cm_event returned %d, event_type %d, done %d, signo %d, qhandle %p",
				message, ret,
				event_type, done, signo, qhandle);
			/* now what ?? */
		}

		TRACE("\t\t%4d:   UNLOCK cb %p event_mutex %p\n",
			__LINE__, cb, &cb->event_mutex);
		pthread_mutex_unlock(&cb->event_mutex);

		/* pop the cancellation handler */
		pthread_cleanup_pop(0);		/* i.e., do NOT do the unlock */

		/* at this point no locks are locked */
		if (our_disconnect_state == 0x7) {
			/* if get here, child_conn == cb of client or agent */
			TRACE("%s %s doing final cleanup\n",
					_time_prefix(), message);
			child_conn->conn_flags &= ~(EXS_NEED_DISCONNECT
							| EXS_OK_TO_ISSUE_IO);

			/* cb is socket whose con-man thread
			 * should terminate now because it just
			 * got disconnect event and exs_close() was
			 * already called by user
			 */
			our_close_connecting(cb, message, &qhandle, &signo);
			TRACE("%s %s Leave due to"
					" RDMA_CM_EVENT_DISCONNECTED\n",
					_time_prefix(), message);
			/* cb is now invalid */
			do_shutdown = 1;
		}
		if(signo >= 0 && qhandle != NULL)
			_evt_queue_do_notify(qhandle, signo);

		if(to_broadcast != NULL)
		{
			pthread_cond_broadcast(to_broadcast);
		}
		if(do_shutdown)
		{	/* self-terminate this con_man_thread */
			TRACE("%s %s Leave\n",
				_time_prefix(), message);
			pthread_exit(NULL);
			/* con-man thread's flow never goes beyond this */
		}
	}	/* for ever */
}	/* our_con_man_thread */


/*
 * called only from exs_socket() and our_continue_merge_pending_accept()
 */
static int
create_our_con_man_thread(struct exs_conn *cb, const char *message)
{
	int	ret;
	char	errbuf[_MAX_STRERROR_MALLOC];

	ret = pthread_create(&cb->event_thread, NULL, our_con_man_thread, cb);
	if(ret)
	{
		if (strerror_r(ret, errbuf, _MAX_STRERROR_MALLOC)) {
			snprintf(errbuf, _MAX_STRERROR_MALLOC, "unknown");
		}
		TRACE_ERR("%s: pthread_create of event thread failed, error %d: %s",
			message, ret, errbuf);
		return ret;

	}
	TRACE("%s %s: CREATE event_thread %lu "
		"on cb %p\n", _time_prefix(), message, cb->event_thread, cb);

	ret = pthread_detach(cb->event_thread);
	if(ret)
	{
		if (strerror_r(ret, errbuf, _MAX_STRERROR_MALLOC)) {
			snprintf(errbuf, _MAX_STRERROR_MALLOC, "unknown");
		}
		TRACE_ERR("%s: pthread_detach of event thread failed, error %d: %s",
			message, ret, errbuf);
		/* ignore this error */
		ret = 0;
	}

	if (cb->con_man_thread_is_active == 0) {
		/* con_man_thread not up yet, wait for it to get started */
		TRACE("\t\t%4d:     LOCK cb %p event_mutex %p\n",
			__LINE__, cb, &cb->event_mutex);
		pthread_mutex_lock(&cb->event_mutex);
		TRACE("\t\t%4d:     PAST cb %p event_mutex %p\n",
			__LINE__, cb, &cb->event_mutex);

		while (cb->con_man_thread_is_active == 0) {
			pthread_cond_wait(&cb->transmit_ok_notify,
							&cb->event_mutex);
		}

		TRACE("\t\t%4d:   UNLOCK cb %p event_mutex %p\n",
			__LINE__, cb, &cb->event_mutex);
		pthread_mutex_unlock(&cb->event_mutex);
	}

	return ret;
}	/* create_our_con_man_thread */


/*
 * called from our_undo_exs_socket() case 7
 * which is only called from our_start_close() when exs_socket() on this
 * conn has completed ok and
 * either it has only progressed through bind before the close
 * or it is a listener, not a client or an agent
 *
 * when called, caller has locked event_mutex
 */
static void
our_terminate_con_man_thread(struct exs_conn *cb, const char *message)
{
	int	ret;
	char	errbuf[_MAX_STRERROR_MALLOC];
	(void)message;

	TRACE("%s %s: CANCELing our_con_man_thread %lu on cb %p\n",
		_time_prefix(), message, cb->event_thread, cb);

	ret = pthread_cancel(cb->event_thread);
	if(ret)
	{
		if (strerror_r(ret, errbuf, _MAX_STRERROR_MALLOC)) {
			snprintf(errbuf, _MAX_STRERROR_MALLOC, "unknown");
		}
		TRACE("%s %s: pthread_cancel of our_con_man_thread "
			"failed, error %d: %s\n", _time_prefix(), message,
			ret, errbuf);
		/* ignore this error */
	} else {
		TRACE(
			"%s %s: CANCELed  our_con_man_thread %lu on cb %p\n",
			_time_prefix(), message, cb->event_thread, cb);
		cb->event_thread = 0;
	}

	sched_yield();
}	/* our_terminate_con_man_thread */


/* global statistics counters */
#if defined(ALLOW_EXS_STATISTICS)

	static uint64_t statistics_counters[max_statistics_counter_type] = {0};

	static char* statistics_counter_labels[max_statistics_counter_type] = {
		"total completion loops",
		"total ibv poll cqs",
		"empty ibv poll cqs",
		"total ibv get cq events",
		"total completions",
		"send advert completions",
		"send rdma_write completions",
		"recv advert completions",
		"recv rdma_with_imm completions",
		"recv eof completions",
		"setup completions",
		"matched recv await",
		"unmatched recv await",
		"send done after sending",
		"send done after WWI completion",
		"total setup sends",
		"advert blocks sent",
		"signaled inline advert blocks sent",
		"unsignaled inline advert blocks sent",
		"signaled inline rdma_write blocks sent",
		"unsignaled inline rdma_write blocks sent",
		"rdma_write blocks sent",
		"rdma_write bytes sent",
		"ack_blocks sent",
		"recv ack completions",
	};

/* NOTE:	advertisements are always inline (if the CA supports it)
		rdma_writes can be inline or not

		advertisement blocks can be signaled or unsignaled
		rdma_writes can be signaled or unsignaled
 */

#endif


/*
 *	exs_reset_counters() - not in Interconnect Software Consortium Issue 1.0
 *		      was added by us
 *
 *	resets current value of statistics counters
 */
EXS_EXPORT void
exs_reset_counters(void)
{
#if defined(ALLOW_EXS_STATISTICS)
	enum statistics_counter_type	i;

	TRACE("exs_reset_counters\n");
	for (i = completion_loops; i < max_statistics_counter_type; i++) {
		statistics_counters[i] = 0;
	}
#endif
}	/* exs_reset_counters */


/*
 *	exs_print_counters() - not in Interconnect Software Consortium Issue 1.0
 *		      was added by us
 *
 *	prints current value of statistics counters
 */
EXS_EXPORT void
exs_print_counters(FILE *fptr)
{
	(void)fptr;
#if defined(ALLOW_EXS_STATISTICS)
	enum statistics_counter_type	i;

	for (i = completion_loops; i < max_statistics_counter_type; i++) {
		fprintf(fptr, "%2d %20llu %s\n", i,
			(long long unsigned)statistics_counters[i],
			statistics_counter_labels[i]);
	}
	fflush(fptr);
#endif

}	/* exs_print_counters */


/*
 *	exs_dump_counters() - not in Interconnect Software Consortium Issue 1.0
 *		      was added by us
 *
 *	prints current value of statistics counters
 */
EXS_EXPORT void
exs_dump_counters(void)
{
#if defined(ALLOW_EXS_STATISTICS)
	exs_print_counters(stderr);
#endif
}	/* exs_dump_counters */


/*
 *	exs_print_calloc_report() - not in Interconnect Software Consortium Issue 1.0
 *		      was added by us
 *
 *	prints out any remaining unfreed heap allocations
 */
EXS_EXPORT void
exs_print_calloc_report(FILE *fptr)
{
	(void)fptr;
	EXS_CALLOC_REPORT(fptr);
}	/* exs_print_calloc_report */


static struct timeval	base_tval;
static char		time_string[16];


/* store current time of day in global base_tval once at beginning of run */
static void
mark_time_base(void)
{
	gettimeofday(&base_tval, NULL);
	TRACE("%d.%06d: set base time\n", (int)base_tval.tv_sec,
		(int)base_tval.tv_usec);
}	/* mark_time_base */


/* create a time-stamp for debug trace printing in global time_string array */
char *
_time_prefix(void)
{
	struct timeval	cur_tval;

	gettimeofday(&cur_tval, NULL);
	if((int)base_tval.tv_usec > (int)cur_tval.tv_usec) {
		cur_tval.tv_usec += 1000000;
		cur_tval.tv_sec -= 1;
	}
	cur_tval.tv_usec -= base_tval.tv_usec;
	cur_tval.tv_sec -= base_tval.tv_sec;
	sprintf(time_string, "%3d.%06d:", (int)cur_tval.tv_sec,
		(int)cur_tval.tv_usec);
	return time_string;
}	/* _time_prefix */

void
exs_log(int module, int priority,
		const char *file, int line, const char *function,
		const char *format, ...)
{
	FILE *fptr;
	va_list args;
	(void)module;
	(void)priority;

	fptr = stderr; /* to get around git pre-commit hook */
	flockfile(stderr);
	fprintf(fptr, "%s %s:%d %s ",
			_time_prefix(), file, line, function);
	va_start(args, format);
	vfprintf(fptr, format, args);
	va_end(args);
	putc_unlocked('\n', stderr);
	funlockfile(stderr);
}


/* holds the pid of the last process that successfully called exs_init */
static pid_t global_owner = 0;


/*
 *	exs_init - page 21 of Interconnect Software Consortium Issue 1.0
 *		extended socket API initialization
 *		Must be called once per process
 *		prior to calling any other exs_xxx function.
 *
 *	returns 0 if all ok, else -1 with errno set to error value
 */
EXS_EXPORT int
exs_init(int version)
{
	pid_t			current;
	int			num_devices, count, n, ret;
	struct ibv_context	**list, **lptr;
	struct ibv_device_attr	dev_attrs;
	char			*tptr;

	ret = 0;
	mark_time_base();

	current = getpid();
	TRACE("%s pid %d start %s, version %d\n",
		_time_prefix(), current, __func__, version);

	if(version != EXS_VERSION)
	{
		errno = ENOTSUP;
		return -1;
	}

	if(global_owner == current)
	{
		/* already called once by this process */
		errno = EALREADY;	/* Operation already in progress */
		return -1;
	}

	TRACE_INFO(EXS_LOG_UNSPECIFIED,
		"UNH-EXS version %s compiled on %s at %s",
		our_version, __DATE__, __TIME__);

	num_devices = 0;
	list = rdma_get_devices(&num_devices);
	if(list == NULL || num_devices == 0)
	{
		TRACE_ERR("No RDMA devices found");
		errno = ENXIO;		/* No such device or address */
		return -1;
	}

	TRACE("%s pid %d %s: found %d RDMA devices\n",
		_time_prefix(), current, __func__, num_devices);

	our_init_pd_table();
	our_init_socket_table();

	for(count = 0, lptr = list; *lptr != NULL; lptr++, count++)
	{
		switch ((*lptr)->device->transport_type)
		{
		case IBV_TRANSPORT_IB:
			tptr = "InfiniBand";
			break;
		case IBV_TRANSPORT_IWARP:
			tptr = "iWARP";
			break;
		default:
			tptr = "Unknown transport type";
			break;
		}	/* switch */
		TRACE("%s pid %d %s: %d %s, %s\n",
			_time_prefix(), current, __func__, count,
			(*lptr)->device->name, tptr);
		if((n = ibv_query_device(*lptr, &dev_attrs)) != 0)
		{
			if(n < 0)
				n = -n;
			OUR_PERROR("ibv_query_device returned error", errno);
		continue;
		}
		TRACE("%d: RDMA device vendor id %u, part id %u, "
			"hardware version %u\n", count, dev_attrs.vendor_id,
			dev_attrs.vendor_part_id, dev_attrs.hw_ver);
		TRACE("\tMaximum number of supported queue pairs - "
			"max_qp %d\n", dev_attrs.max_qp);
		TRACE("\tMaximum number of outstanding work requests "
			"- max_qp_wr %d\n", dev_attrs.max_qp_wr);
		TRACE("\tMaximum number of scatter-gather items per "
			"work request for non-registered queue pairs"
			"- max_sge %d\n", dev_attrs.max_sge);
		TRACE("\tMaximum number of scatter-gather items per "
			"work request for registered queue pairs"
			"- max_sge_rd %d\n", dev_attrs.max_sge_rd);
		TRACE("\tMaximum number of supported completion "
			"queues - max_cq %d\n", dev_attrs.max_cq);
		TRACE("\tMaximum number of elements per completion "
			"queue - max_cqe %d\n", dev_attrs.max_cqe);
		TRACE("\tMaximum number of supported memory "
			"regions - max_mr %d\n", dev_attrs.max_mr);
		TRACE("\tMaximum number of supported protection "
			"domains - max_pd %d\n", dev_attrs.max_pd);
		if (our_get_pd(*lptr) == NULL) {
			ret = errno;
			our_destroy_pd_table();
			errno = ret;
			ret = -1;
			goto errout;
		}
	}	/* for */

	/* Mark that exs_init has been called successfully by this process.
	 * This allows all other exs_xxx functions to be called by this process.
	 */
	global_owner = current;

errout:
	rdma_free_devices(list);

	TRACE("%s pid %d %s: finished, ret %d\n",
		_time_prefix(), current, __func__, ret);
	sched_yield();
	return ret;
}	/* exs_init */


struct exs_mhandle
{
	void			*addr;		/* address registered by user*/
	size_t			size;		/* size registered by user */
	enum ibv_access_flags	access_flags;
	pthread_rwlock_t	mr_table_rwlock;
	struct ibv_mr		*recent_mr;
	struct ibv_mr		*mr_table[MAX_PD_SLOTS];
};	/* struct exs_mhandle */


/*
 * called from exs_mregister(), and our_common_send()
 */
static inline struct ibv_mr *
our_call_ibv_reg_mr(struct ibv_pd *pd, void *addr, size_t size, int flags,
			const char *mess)
{
	struct ibv_mr	*mr_ptr;

	TRACE("%s \tcall ibv_reg_mr: pd %p, addr %p, size %zu, "
			"flags %0x for %s\n",
			_time_prefix(), pd, addr, size, flags, mess);

	if((mr_ptr = ibv_reg_mr(pd, addr, size, flags)) == NULL)
	{
		TRACE_ERR("failed ibv_reg_mr: pd %p, addr %p, size %zu, flags 0x%x, errno %d for %s",
			pd, addr, size, flags, errno, mess);
	}
	else
	{
		TRACE("%s \tdone ibv_reg_mr: pd %p, addr %p, size %zu, "
			"flags 0x%x for %s at mr %p\n",
			_time_prefix(), pd, addr, size, flags, mess, mr_ptr);
	}
	return mr_ptr;
}	/* our_call_ibv_reg_mr */


/*
 * called from and our_common_send()
 */
static inline struct ibv_mr *
our_ensure_buffer_registered(struct exs_conn *this_conn, exs_mhandle_t mhandle,
				const char *message)
{
	struct ibv_mr	*mr, **mr_ptr;

	TRACE("%s %s: mhandle %p should be registered, "
		"mhandle->addr %p\n", _time_prefix(), message,
		mhandle, mhandle->addr);

	/* see if most recently used mr for this mhandle fits this conn */
	if ((mr = mhandle->recent_mr) != NULL && mr->pd == this_conn->pd) {
		return mr;
	}

	/* most recent use of this mhandle does not fit this conn's pd
	 * search the mr_table to find one that does
	 */

	OUR_MEMBER_RWLOCK_RDLOCK(mhandle, mr_table_rwlock);

	for (mr_ptr = &mhandle->mr_table[0]; *mr_ptr != NULL; mr_ptr++) {
		if ((*mr_ptr)->pd == this_conn->pd)
			break;
	}

	OUR_MEMBER_RWLOCK_UNLOCK(mhandle, mr_table_rwlock);

	if (*mr_ptr != NULL) {
		/* we found a fit, cache it as most recently used mr */
		mhandle->recent_mr = *mr_ptr;
		goto fit_ok;
	}

	/* this should rarely happen, but sometime after this mhandle was
	 * created, a new device must have come on-line, so we now need to add
	 * a registration for that device to this (old) mhandle's mr_table
	 */

	OUR_MEMBER_RWLOCK_WRLOCK(mhandle, mr_table_rwlock);

	/* make sure it is not now in the mhandle's mr_table */
	for (mr_ptr = &mhandle->mr_table[0]; *mr_ptr != NULL; mr_ptr++) {
		if ((*mr_ptr)->pd == this_conn->pd)
			break;
	}

	if (*mr_ptr != NULL) {
		/* we found a fit, cache it as most recently used mr */
		mhandle->recent_mr = *mr_ptr;
	} else if (mr_ptr >= &mhandle->mr_table[MAX_PD_SLOTS-1]) {
			/* no room in this mr_table for another mr */
			TRACE_ERR("%s: mhandle %p mr_table full, max of %d pd-specific registrations",
				message, mhandle, MAX_PD_SLOTS - 1);
			errno = ENOMEM;
	} else {
		/* bad news -- there is no valid mr_table entry for conn's pd
		 * try to get one (creating it if necessary) from global table
		 */
		if (this_conn->pd == our_get_pd(this_conn->pd->context)) {
			/* ok, the global pd is set up */
			*mr_ptr = our_call_ibv_reg_mr(this_conn->pd,
						mhandle->addr, mhandle->size,
						mhandle->access_flags, message);
			if (*mr_ptr != NULL) {
				/* we registered it too, so cache it as
				 * most recently used mr
				 */
				mhandle->recent_mr = *mr_ptr;
			} else {
				errno = ENXIO;	/* No such device or address */
			}
		} else {
			TRACE_ERR("%s: mhandle %p has no matching mr_table for conn %p pd %p",
				message, mhandle, this_conn, this_conn->pd);
			errno = ENXIO;		/* No such device or address */
		}
	}

	OUR_MEMBER_RWLOCK_UNLOCK(mhandle, mr_table_rwlock);

fit_ok:
	TRACE("%s %s: returning mr %p\n",
			_time_prefix(), __func__, *mr_ptr);
	return *mr_ptr;
}	/* our_ensure_buffer_registered */


static exs_mhandle_t
our_mregister(void *addr, size_t size, int flags, const char *message)
{
	exs_mhandle_t	mhandle;
	struct ibv_pd	**pd_ptr;
	struct ibv_mr	**mr_ptr;

	/***** NEED to calloc storage for mhandle to point to *****/
	if((mhandle = EXS_CALLOC(1, sizeof(*mhandle))) == NULL)
	{
		return EXS_MHANDLE_INVALID;
	}

	mhandle->addr = addr;
	mhandle->size = size;

	if((flags & EXS_ACCESS_ALL) == EXS_ACCESS_ALL)
	{
	mhandle->access_flags	|= IBV_ACCESS_LOCAL_WRITE
					|  IBV_ACCESS_REMOTE_READ
					|  IBV_ACCESS_REMOTE_WRITE;
	}
	else
	{
		if((flags & EXS_LOCAL_WRITE))
		mhandle->access_flags	|= IBV_ACCESS_LOCAL_WRITE;
		if((flags & EXS_REMOTE_READ))
		mhandle->access_flags	|= IBV_ACCESS_REMOTE_READ;
		if((flags & EXS_REMOTE_WRITE))
		mhandle->access_flags	|= IBV_ACCESS_REMOTE_WRITE
						|  IBV_ACCESS_LOCAL_WRITE;
	}

	pthread_rwlock_init(&mhandle->mr_table_rwlock, NULL);

	/* fill the mr_table by registering this region with all known pds */

	OUR_MEMBER_RWLOCK_RDLOCK(mhandle, mr_table_rwlock);

	for (pd_ptr = &pd_table[0], mr_ptr = &mhandle->mr_table[0];
					*pd_ptr != NULL; pd_ptr++, mr_ptr++) {
		*mr_ptr = our_call_ibv_reg_mr(*pd_ptr, addr, size + 1,
						mhandle->access_flags, message);
		if (*mr_ptr == NULL) {
			break;
		}
	}

	/* set the recent_mr to the first item in mr_table */
	if ((mhandle->recent_mr = mhandle->mr_table[0]) == NULL) {
		/* must always have a recent_mr */
		pthread_rwlock_destroy(&mhandle->mr_table_rwlock);
		EXS_FREE(mhandle);
		errno = ENXIO;		/* No such device or address */
		mhandle = EXS_MHANDLE_INVALID;
	}

	OUR_MEMBER_RWLOCK_UNLOCK(mhandle, mr_table_rwlock);

	return mhandle;
}	/* our_mregister */


/*
 *	exs_mregister - page 25 of Interconnect Software Consortium Issue 1.0
 *		register application memory
 *
 *	returns handle if all ok,
 *		else EXS_MHANDLE_INVALID with errno set to error value
 */
EXS_EXPORT exs_mhandle_t
exs_mregister(void *addr, size_t size, int flags)
{
	exs_mhandle_t		ret;

#if defined(ALLOW_PID_CHECKING)
	pid_t			current;

	current = getpid();
	TRACE("%s pid %d start %s, addr %p, size %d, "
		"flags 0x%x\n", _time_prefix(), current, __func__, addr,
		(int)size, flags);

	if(global_owner != current)
#else
	if(global_owner == 0)
#endif
	{
		errno = EPERM;		/* Operation not permitted */
		return EXS_MHANDLE_INVALID;
	}

	if(size <= 0 || flags > EXS_ALL_FLAGS || addr == NULL)
	{
		errno = EINVAL;		/* Invalid argument */
		return EXS_MHANDLE_INVALID;
	}

	ret = our_mregister(addr, size, flags, __func__);

#if defined(ALLOW_PID_CHECKING)
	TRACE("%s %s: finished, returning %p, "
		"flags 0x%x\n", _time_prefix(), __func__, ret, flags);
#endif

	return ret;
}	/* exs_mregister */


/*
 *	exs_mmodify - page 23 of Interconnect Software Consortium Issue 1.0
 *		modify registered application memory
 *
 *	returns 0 if all ok, else -1 with errno set to error value
 */
EXS_EXPORT int
exs_mmodify(exs_mhandle_t mhandle, size_t size, int flags)
{

#if defined(ALLOW_PID_CHECKING)
	pid_t		current;

	current = getpid();
	TRACE("%s pid %d start %s, size %d, "
		"flags 0x%x\n", _time_prefix(), current, __func__,
		(int)size, flags);

	if(global_owner != current)
#else
	if(global_owner == 0)
#endif
	{
		errno = EPERM;		/* Operation not permitted */
		return -1;
	}

	if(size <= 0
		|| flags > EXS_ALL_FLAGS
		|| mhandle == EXS_MHANDLE_INVALID
		|| mhandle == EXS_MHANDLE_UNREGISTERED
		|| mhandle == NULL)
	{
		errno = EINVAL;		/* Invalid argument */
		return -1;
	}
	mhandle->size = size;
	mhandle->access_flags = 0;
	if((flags & EXS_ACCESS_ALL) == EXS_ACCESS_ALL)
	{
	mhandle->access_flags |= IBV_ACCESS_LOCAL_WRITE
				      |  IBV_ACCESS_REMOTE_READ
				      |  IBV_ACCESS_REMOTE_WRITE;
	}
	else
	{
		if((flags & EXS_LOCAL_WRITE))
		mhandle->access_flags |= IBV_ACCESS_LOCAL_WRITE;
		if((flags & EXS_REMOTE_READ))
		mhandle->access_flags |= IBV_ACCESS_REMOTE_READ;
		if((flags & EXS_REMOTE_WRITE))
		mhandle->access_flags |= IBV_ACCESS_REMOTE_WRITE
					      |  IBV_ACCESS_LOCAL_WRITE;
	}

#if defined(ALLOW_PID_CHECKING)
	TRACE("%s %s: finished ok\n",
		_time_prefix(), __func__);
#endif

	return 0;
}	/* exs_mmodify */


static void
our_mderegister(exs_mhandle_t mhandle)
{
	struct ibv_mr	**mr_ptr;

	/* unregister all items in the mr_table
	 * here we assume that the mr_table array ends with a NULL entry */
	for (mr_ptr = &mhandle->mr_table[0]; *mr_ptr != NULL; mr_ptr++) {
		ibv_dereg_mr(*mr_ptr);
	}

	mhandle->recent_mr = NULL;

#if defined(ALLOW_FULL_TRACE)
	memset(mhandle, 0, sizeof(*mhandle));	/* for debugging only */
#endif

	pthread_rwlock_destroy(&mhandle->mr_table_rwlock);

	EXS_FREE(mhandle);
}	/* our_mderegister */


/*
 *	exs_mderegister - page 22 of Interconnect Software Consortium Issue 1.0
 *		deregister application memory
 *
 *	returns 0 if all ok, else -1 with errno set to error value
 */
int
EXS_EXPORT exs_mderegister(exs_mhandle_t mhandle, int flags)
{

#if defined(ALLOW_PID_CHECKING)
	pid_t		current;

	current = getpid();
	TRACE("%s pid %d start %s, mhandle %p flags 0x%x\n",
		_time_prefix(), current, __func__, mhandle, flags);

	if(global_owner != current)
#else
	if(global_owner == 0)
#endif
	{
		errno = EPERM;		/* Operation not permitted */
		return -1;
	}

	/* Flags are unused in Issue 1.0 of the specification */
	(void) flags;

	if(mhandle == EXS_MHANDLE_INVALID ||
	   mhandle == EXS_MHANDLE_UNREGISTERED ||
	   mhandle == NULL)
	{
		errno = EINVAL;		/* Invalid argument */
		return -1;
	}
	our_mderegister(mhandle);

#if defined(ALLOW_PID_CHECKING)
	TRACE("%s %s: finished ok\n",
		_time_prefix(), __func__);
#endif

	return 0;
}	/* exs_mdregister */


/*
 *	exs_qcreate - page 33 of Interconnect Software Consortium Issue 1.0
 *		create an event queue
 *
 *	returns handle if all ok,
 *		else EXS_QHANDLE_INVALID with errno set to error value
 */
EXS_EXPORT exs_qhandle_t
exs_qcreate(int depth)
{
	our_qhandle_t		queue;

#if defined(ALLOW_PID_CHECKING)
	pid_t			current;

	current = getpid();
	TRACE("%s pid %d start %s, depth %d\n",
		_time_prefix(), current, __func__, depth);

	if(global_owner != current)
#else
	if(global_owner == 0)
#endif
	{
		errno = EPERM;		/* Operation not permitted */
		return EXS_QHANDLE_INVALID;
	}

	if((queue = _evt_queue_init(depth)) == NULL)
	{	/* EXS standard requires this return value */
		return EXS_QHANDLE_INVALID;
	}

#if defined(ALLOW_PID_CHECKING)
	TRACE("%s %s: finished, returning queue %p\n",
		_time_prefix(), __func__, queue);
#endif

	return _cnv_ptr_to_qnumb(queue);
}	/* exs_qcreate */


/*
 *	exs_qdelete - page 35 of Interconnect Software Consortium Issue 1.0
 *		delete an event queue
 *
 *	returns 0 if all ok, else -1 with errno set to error value
 */
EXS_EXPORT int
exs_qdelete(exs_qhandle_t qhandle)
{
	our_qhandle_t		qptr;
#if defined(ALLOW_PID_CHECKING)
	pid_t			current;

	current = getpid();
	TRACE("%s pid %d start %s, qhandle %p\n",
		_time_prefix(), current, __func__, qhandle);

	if(global_owner != current)
#else
	if(global_owner == 0)
#endif
	{
		errno = EPERM;		/* Operation not permitted */
		return -1;
	}

	if((qptr = _cnv_qnumb_to_ptr(qhandle)) == NULL)
	{	/* Invalid argument */
		return -1;
	}
	return _evt_queue_deinit(qptr);	/* errno is set */
}	/* exs_qdelete */


/*
 *	exs_qmodify - page 38 of Interconnect Software Consortium Issue 1.0
 *		modify event queue attributes
 *
 *	returns 0 if all ok, else -1 with errno set to error value
 */
EXS_EXPORT int
exs_qmodify(exs_qhandle_t qhandle, int attr_type, void *attr_value,
		 size_t attr_length)
{
	our_qhandle_t		qptr;
#if defined(ALLOW_PID_CHECKING)
	pid_t			current;

	current = getpid();
	TRACE("%s pid %d start %s, qhandle %p\n",
		_time_prefix(), current, __func__, qhandle);

	if(global_owner != current)
#else
	if(global_owner == 0)
#endif
	{
		errno = EPERM;		/* Operation not permitted */
		return -1;
	}

	if((qptr = _cnv_qnumb_to_ptr(qhandle)) == NULL)
	{	/* Invalid argument */
		return -1;
	}
	if(attr_type == EXS_QATTR_SIGNAL)
	{
		if(attr_length != sizeof(exs_signal_t*))
		{
			errno = EINVAL;		/* Invalid argument */
			return -1;
		}
		if(((exs_signal_t *)attr_value)->exs_sigstate==EXS_SIG_DISABLE)
		{
			return _evt_queue_disable_signal(qptr);
		}
		else if(((exs_signal_t *)attr_value)->exs_sigstate == EXS_SIG_ENABLE)
		{
			return _evt_queue_enable_signal(qptr,
				((exs_signal_t *)attr_value)->exs_signo);
		}
		else
		{
			errno = EINVAL;		/* Invalid argument */
			return -1;
		}
	}
	else if(attr_type == EXS_QATTR_DEPTH)
	{
		/* currently not supported */
		errno = EINVAL;		/* Invalid argument */
		return -1;
	}
	else
	{	
		errno = EINVAL;		/* Invalid argument */
		return -1;
	}

#if defined(ALLOW_PID_CHECKING)
	TRACE("%s %s: finished ok\n",
		_time_prefix(), __func__);
#endif

	return 0;
}	/* exs_qmodify */


/*
 *	exs_qdequeue - page 36 of Interconnect Software Consortium Issue 1.0
 *		retrieve events from an event queue
 *
 *	returns number of events dequeued >= 0 if all ok,
 *		else -1 with errno set to error value
 */
EXS_EXPORT int
exs_qdequeue(exs_qhandle_t qhandle, exs_event_t *evtvec, int evtvec_cnt,
		const struct timeval *timeout)
{
	our_qhandle_t		qptr;
	int			ndequeued;

#if defined(ALLOW_PID_CHECKING)
	pid_t		current;

	current = getpid();
	TRACE("%s pid %d start %s, qhandle %p, "
		"evtvec_cnt %d\n",
		_time_prefix(), current, __func__, qhandle, evtvec_cnt);

	if(global_owner != current)
#else
	if(global_owner == 0)
#endif
	{
		errno = EPERM;		/* Operation not permitted */
		return -1;
	}

	if((qptr = _cnv_qnumb_to_ptr(qhandle)) == NULL)
	{	/* Invalid argument */
		return -1;
	}
	if(evtvec_cnt < 1)
	{
		errno = EINVAL;		/* Invalid argument */
		return -1;
	}

	ndequeued = _evt_queue_dequeue_vector(qptr,evtvec,evtvec_cnt,timeout);

	TRACE("%s %s: finished, qhandle %p, returning %d, "
		"evtvec_cnt %d\n", _time_prefix(), __func__, qhandle, ndequeued,
		evtvec_cnt);

	return ndequeued;
}	/* exs_qdequeue */


/*
 *	exs_qstatus - page 40 of Interconnect Software Consortium Issue 1.0
 *		retrieve event queue attributes
 *
 *	returns 0 if all ok, else -1 with errno set to error value
 */
EXS_EXPORT int
exs_qstatus(exs_qhandle_t qhandle, int attr_type, void *attr_value,
		size_t attr_length)
{
	our_qhandle_t		qptr;
#if defined(ALLOW_PID_CHECKING)
	pid_t			current;

	current = getpid();
	TRACE("%s pid %d start %s, qhandle %p\n",
		_time_prefix(), current, __func__, qhandle);

	if(global_owner != current)
#else
	if(global_owner == 0)
#endif
	{
		errno = EPERM;		/* Operation not permitted */
		return -1;
	}

	if((qptr = _cnv_qnumb_to_ptr(qhandle)) == NULL)
	{	/* Invalid argument */
		return -1;
	}

	if(attr_type == EXS_QATTR_DEPTH)
	{
		if(attr_length != sizeof(int))
		{
			errno = EINVAL;		/* Invalid argument */
			return -1;
		}
		return _get_evt_queue_depth(qptr, (int*)attr_value);
	}
	else if(attr_type == EXS_QATTR_SIGNAL)
	{
		if(attr_length != sizeof(exs_signal_t*))
		{
			errno = EINVAL;		/* Invalid argument */
			return -1;
		}
		return _get_evt_queue_signal(qptr, (exs_signal_t *)attr_value);
		}
	else if(attr_type == EXS_QATTR_EVENTS)
	{
		if(attr_length != sizeof(int))
		{
			errno = EINVAL;		/* Invalid argument */
			return -1;
		}
		return _get_evt_queue_n_events(qptr, (int*)attr_value);
	}
	else
	{
		errno = EINVAL;		/* Invalid argument */
		return -1;
	}

#if defined(ALLOW_PID_CHECKING)
	TRACE("%s %s: finished ok\n",
		_time_prefix(), __func__);
#endif

	return 0;
}	/* exs_qstatus */


#if defined(ALLOW_EXS_TRACE)

static void
our_dump_buf(void *buffer, int length)
{
	int	i, actual_length;
	char	*ptr, dumpbuf[120];

	if(_our_debug_flag)
	{
		/* dump only min(length, 500) bytes */
		actual_length = length;
		if(length > 500)
			length = 500;
		/* dump the buffer in hex, 16 bytes per line */
		ptr = dumpbuf + sprintf(dumpbuf, "%11d:", 0);;
		for(i = 0; i < length; i++)
		{
			ptr = ptr + sprintf(ptr, " %02x",
					*(((unsigned char *)buffer) + i));
			if(((i + 1) & 0x0f) == 0)
			{
			fprintf(stderr, "%s\n", dumpbuf);
			ptr = dumpbuf + sprintf(dumpbuf, "%11d:", i+1);;
			}
		}
		fprintf(stderr, "%s\n", dumpbuf);
		if(length < actual_length)
			fprintf(stderr, "%12s %d remaining bytes not dumped\n",
				"", actual_length - length);
	}
}	/* our_dump_buf */

#endif


/*
 *	exs_cancel - page 15 of Interconnect Software Consortium Issue 1.0
 *		cancel pending asynchronous operations
 *
 *	returns 0 if all ok, else -1 with errno set to error value
 */
EXS_EXPORT int
exs_cancel(int flags, int fildes, exs_ahandle_t ahandle)
{
	struct exs_conn		*this_conn;

#if defined(ALLOW_PID_CHECKING)
	pid_t			current;

	current = getpid();
	TRACE("%s start %s, pid %d\n",
		_time_prefix(), __func__, current);

	if(global_owner != current)
#else
	if(global_owner == 0)
#endif
	{
		errno = EPERM;		/* Operation not permitted */
		return -1;
	}

	/* verify that this socket has been created */
	if((this_conn = our_lookup_socket(fildes)) == NULL)
	{
		errno = EBADF;		/* Bad file number */
		return -1;
	}

#if defined(ALLOW_PID_CHECKING)
	TRACE("%s %s: socket %d, this_conn %p\n",
		_time_prefix(), __func__, fildes, this_conn);
#endif

	errno = ENOSYS;		/* Function not implemented */
	(void)flags;
	(void)ahandle;
	return -1;
}	/* exs_cancel */



/* called from:
 *   1.	our_exs_recv_done - to pass a user's recv buffer back to the user
 *   2. our_rest_of_exs_recv - to pass an in-place system recv buffer to user
 *
 * queue up this block's completion event for caller of exs_recv or exs_send
 * by removing it from this block and then adding it to tail of this block's
 * qhandle's full_list; then replacing it in this block by removing head
 * of this block's qhandle's empty list
 *
 * Returns signo >= 0 to trigger full_notify.
 */
static inline int
our_post_user_event(struct transaction_block *block, size_t amount)
{
	int	signo;

	/* first fill in actual number of bytes transfered
	 * NOTE -- this assignment assumes that exs_evt_length is at the same
	 * offset in ALL relevant overlays of the exs_evt_union!
	 * These overlays are:
	 *		exs_evt_xfer
	 *		exs_evt_xferfile
	 *		exs_evt_xfermsg		(so far unused)
	 */
	block->user_event_ptr->exs_evt_union.exs_evt_xfer.exs_evt_length
								= amount;

	TRACE_DBG(EXS_LOG_UNSPECIFIED,
		"start block %p with event %p amount %zu buffer_flags %#x\n",
		(void *)block, (void *)block->user_event_ptr, amount,
		block->buffer_flags);

	/* then make the moves */
	signo = _swap_evt_to_full_evt_queue(block->qhandle,
							&block->user_event_ptr);

	/* mark that this block has notified the user */
	block->buffer_flags |= EXS_NOTIFY_ALREADY_DONE;

	TRACE_DBG(EXS_LOG_UNSPECIFIED,
		"done block %p with event %p signo %d buffer_flags %#x\n",
		(void *)block, (void *)block->user_event_ptr, signo,
		block->buffer_flags);

	return signo;
}	/* our_post_user_event */


/* called from:
 *   1.	our_completion_thread itself when RECV_RDMA_WITH_IMM completes;
 *   2. our_rest_of_exs_recv on EOF
 *
 * Returns signo >=0 and qhandle is set to the notification event queue
 */
static int
our_exs_recv_done(struct exs_conn *this_conn,
		struct transaction_block *block,
		our_qhandle_t *qhandle,
		int keep_posted_recv_slot)
{
	int	signo;

	TRACE_DBG(EXS_LOG_UNSPECIFIED,
		"this_conn %p, block %p, seqno %" PRIu32 ", amount %zu",
		this_conn, block, block->recv_header.hd_seqno,
		block->recv_byte_len);

	if(block->recv_unreg_mr != NULL) {
		TRACE_DBG(EXS_LOG_UNSPECIFIED,
			"deregister recv_unreg_mr addr %p, size %zu",
			block->recv_unreg_mr->addr,
			block->recv_unreg_mr->length);
		ibv_dereg_mr(block->recv_unreg_mr);
		block->recv_unreg_mr = NULL;
	} else if (block->recv_flags & RECV_COPY_UNREG_BIT) {
		memcpy(block->user_event_ptr->exs_evt_union
						.exs_evt_xfer.exs_evt_buffer,
				block->recv_rdma_buf, block->recv_byte_len);
		block->recv_flags &= ~RECV_COPY_UNREG_BIT;
	}

	if (block->buffer_flags & EXS_NOTIFY_ALREADY_DONE) {
		qhandle = NULL;
		signo = -1;
	} else if ((block->qhandle == NULL)
			|| ((block->buffer_flags & EXS_UNSIGNALED)
				&& block->user_event_ptr->exs_evt_errno == 0)) {
		/* the user didn't request a signal */
		block->buffer_flags |= EXS_NOTIFY_ALREADY_DONE;
		atomic_add_reset(-1, UINT_MAX, &this_conn->pending_recv_count);
		*qhandle = NULL;
		signo = -1;
	} else {
		/* return the qhandle because block will be freed after event
		 * posting */
		atomic_add_reset(-1, UINT_MAX, &this_conn->pending_recv_count);
		*qhandle = block->qhandle;
		signo = our_post_user_event(block, block->recv_byte_len);
	}

	/* we are finally done with the block for this recv transaction */
	if (!keep_posted_recv_slot) {
		block->in_progress = 0;
		OUR_MEMBER_SEM_POST(block, sem);
	}

	return signo;
}	/* our_exs_recv_done */


/*
 * called only from our_all_done_with_block()
 *
 * when block is the last in the list for send/sendfile
 *
 * Returns signo == 0 notification but no signal should be given
 *	   signo > 0 if notification and a signal should be given
 *	   signo < 0 if notification was already done previously
 *			or user doesn't want to know about this
 *
 *	*qhandle is set to block->qhandle in all cases
 */
static inline int
our_last_block_sent(struct exs_conn *cb, struct transaction_block *block,
			our_qhandle_t *qhandle)
{
	int	signo;

	TRACE_DBG(EXS_LOG_UNSPECIFIED,
		"our_last_block_sent block %p, event %p, qhandle %p, buffer_flags %#x",
		(void *)block, (void *)block->user_event_ptr,
		(void *)*qhandle, block->buffer_flags);

	if(block->buffer_flags & EXS_NOTIFY_ALREADY_DONE) {
		/* user was already told about this event */
		signo = -1;
	} else if(((*qhandle = block->qhandle) == NULL)
			|| ((block->buffer_flags & EXS_UNSIGNALED)
				&& block->user_event_ptr->exs_evt_errno == 0)) {
		/* user doesn't want to know about this event */
		atomic_add_reset(-1, UINT_MAX, &cb->pending_send_count);
		signo = -1;
	} else {
		/* user wants to know about this event, and was not previously
		 * told about it, post it now
		 */
		/* our_post_user_event fills in exs_evt_length */
		atomic_add_reset(-1, UINT_MAX, &cb->pending_send_count);
		signo = our_post_user_event(block, block->send_total_bytes);
	}
	return signo;
}	/* our_last_block_sent */


/* the ibv_post_send was completed and
 * we got the completion, so block will be freed.
 */
static int
our_all_done_with_block(struct exs_conn *this_conn,
			struct transaction_block *block,
			our_qhandle_t *qhandle,
			int keep_posted_send_slot,
			const char *message)
{
	int	signo, ret;
	char				errbuf[_MAX_STRERROR_MALLOC];

	signo = -1;	/* assume we will not notify the user */

	/* unregister source-side user buffer
	 * if we registered it just for this send only
	 */
	TRACE("\t\t%4d: %s block %p, event %p, "
		"send_unreg_mr %p\n", __LINE__, message, block,
		block->user_event_ptr, block->send_unreg_mr);
	if(block->send_unreg_mr != NULL)
	{
		ret = ibv_dereg_mr(block->send_unreg_mr);
		if(ret)
		{
			if (strerror_r(ret, errbuf, _MAX_STRERROR_MALLOC)) {
				snprintf(errbuf,_MAX_STRERROR_MALLOC,"unknown");
			}
			TRACE_ERR("%s: ibv_dereg_mr failed, error %d: %s",
				message, ret, errbuf);
			/* now what ?? */
		}
		block->send_unreg_mr = NULL;
	}

	/* sets qhandle because block will be freed */
	signo = our_last_block_sent(this_conn, block, qhandle);
	TRACE("%s done with send transaction block %p, event %p\n",
			_time_prefix(), block, block->user_event_ptr);

	if (!keep_posted_send_slot) {
		block->in_progress = 0;
		OUR_MEMBER_SEM_POST(block, sem);
	}

	return signo;
}	/* our_all_done_with_block */


/*
 * called from:
 *	our_send_rdma_write() with block->expected_comp == IBV_WC_RDMA_WRITE
 * or
 *	our_common_send() with block->expected_comp == IBV_WC_RDMA_WRITE
 *
 * When called to post send of advert or immediate,
 *	block->send_total_bytes is set up (by convention - not needed here)
 *
 * returns	-1 with errno set if ibv_post_send fails (block NOT SENT)
  *		0  if ibv_post_send succeeds (block IS SENT)
 */
static int
our_post_send(struct exs_conn *this_conn, struct ibv_send_wr *wr)
{
	int			ret;
	struct ibv_send_wr	*send_bad_wr = NULL;
	struct our_header	*hd_ptr;
	char			errbuf[_MAX_STRERROR_MALLOC];

	TRACE("%s\t call %s: wr %p, wr.wr_id %llx\n",
		_time_prefix(), __func__, wr, (long long unsigned)wr->wr_id);

	if(!(this_conn->conn_flags & EXS_NEED_DISCONNECT))
	{
		TRACE("%s %s ibv_post_send not done due to "
			"conn_state %s(%d), conn_flags %0x\n",
			_time_prefix(), __func__,
			state_labels[this_conn->conn_state],
			this_conn->conn_state, this_conn->conn_flags);
		errno = ENOTCONN;	/* Transport endpoint is not connected*/
		return -1;
	}
	ret = ibv_post_send(this_conn->qp, wr, &send_bad_wr);
	if(ret){
		/* ibv_post_send failed */
		if (strerror_r(-ret, errbuf, _MAX_STRERROR_MALLOC)) {
			snprintf(errbuf, _MAX_STRERROR_MALLOC, "unknown");
		}
		TRACE_ERR("failed ibv_post_send, wr %p, error %d: %s",
			wr, ret, errbuf);

		print_send_wr(stderr, send_bad_wr);

		hd_ptr = (struct our_header *)wr->sg_list[0].addr;
		TRACE("%s\t opcode %d, num_sge %d, length %d, "
			"addr %p\n", _time_prefix(), wr->opcode, wr->num_sge,
			wr->sg_list[0].length, hd_ptr);
		if (wr->opcode == IBV_WR_SEND) {
			TRACE("%s\t hd_type %0x, hd_seqno %d\n",
				_time_prefix(),
				ntohl(hd_ptr->hd_type),
				ntohl(hd_ptr->hd_seqno));
		}
		TRACE("%s\t conn_state %s(%d), conn_flags %0x\n",
			_time_prefix(),
			state_labels[this_conn->conn_state],
			this_conn->conn_state,
			this_conn->conn_flags);
		/* there was an input/output error, is there a better errno? */
		errno = EIO;		/* I/O error */
		return -1;
	}
	TRACE("%s\t done %s: wr %p, wr.wr_id %llx\n",
		_time_prefix(), __func__, wr, (long long unsigned)wr->wr_id);

	return ret;
}	/* our_post_send */


/*
 * called from exs_send() when user sends data
 */
static inline void
our_setup_send_state(struct transaction_block *block)
{
	/* send transactions keep total number of data bytes sent */
	block->send_total_bytes = 0;
	block->buf_offset = 0;

	/* assume we don't need to register memory */
	block->send_unreg_mr = NULL;
}	/* our_setup_send_state */


/*
 * Called only in our_completion_thread() context
 * called from our_post_transaction_blocks()
 *	and our_completion_thread() on an IBV_WC_RECV
 *
 * Returns 0 if all ok, else errno on error.
 */
static int
our_try_to_post_recv_block(struct exs_conn *cb, struct transaction_block *block,
				const char *message)
{
	int			ret;
	struct ibv_recv_wr	*recv_bad_wr;
	struct our_header	*hd_ptr;
	char			errbuf[_MAX_STRERROR_MALLOC];

	/* set expected completion code RECV */
	block->expected_comp = IBV_WC_RECV;

	if(cb->conn_state == EXS_CONN_FINISHED
		|| (cb->conn_state >= EXS_CONN_CLOSING
			&& cb->conn_state <= EXS_CONN_CLOSED))
	{
		TRACE("%s %s ibv_post_recv not done due to "
			"conn_state %s(%d), conn_flags %0x\n",
			_time_prefix(), message,
			state_labels[cb->conn_state],
			cb->conn_state, cb->conn_flags);
		ret = ENOTCONN;		/* Transport endpoint is not connected*/
	}
	else if((ret = ibv_post_recv(cb->qp, &block->recv_wr, &recv_bad_wr)))
	{
		if (strerror_r(-ret, errbuf, _MAX_STRERROR_MALLOC)) {
			snprintf(errbuf, _MAX_STRERROR_MALLOC, "unknown");
		}
		TRACE_ERR("%s failed ibv_post_recv, wr %p, error %d: %s",
			message, &block->recv_wr, -ret, errbuf);

		hd_ptr = (struct our_header *)block->recv_wr.sg_list->addr;
		TRACE("%s\t num_sge %d, length %d, "
			"addr %p\n", _time_prefix(), block->recv_wr.num_sge,
			block->recv_wr.sg_list->length, hd_ptr);
		TRACE("%s\t conn_state %s(%d), conn_flags %0x\n",
			_time_prefix(),
			state_labels[cb->conn_state],
			cb->conn_state,
			cb->conn_flags);
	}
	else
	{	
		TRACE("\t\t ibv_post_recv ok for block %p, "
			"recv_wr.wr_id %llx\n", block,
			(long long unsigned)block->recv_wr.wr_id);
	}
	return ret;
}	/* our_try_to_post_recv_block */


/*
 * Called only in our_completion_thread() context
 * called only from our_create_and_post_blocks()
 *
 * Tries to call ibv_post_recv() 2 * posted_credits times
 *	(via calls to our_try_to_post_recv_block())
 *
 * Returns 0 if all ok, else errno on error.
 */
static int
our_post_transaction_blocks(struct exs_conn *this_conn)
{
	int				ret, j, n_blocks_to_post;

	/* post all necessary receive work requests, minus 1
	 * (which has been posted as part of the initial negotiation) */
	n_blocks_to_post = RQ_DEPTH(this_conn) - 1;
	ret = j = 0;
	for (j = 0; j < n_blocks_to_post; ++j) {
		ret = our_try_to_post_recv_block(this_conn,
					&this_conn->recv_array[j], __func__);
		if(ret)
		{
			/* for now, we go with just the blocks already posted */
			break;
		}
	}
	if (j > 0) {
		this_conn->recv_array[j - 1].next = &this_conn->recv_array[0];
	}

	TRACE("%s %s: posted %d of %d recv transaction blocks on cb %p\n",
		_time_prefix(), __func__, j, n_blocks_to_post, this_conn);

	if(ret == 0 && j == 0)
		ret = ENOMEM;

	return ret;
}	/* our_post_transaction_blocks	*/


/*
 * called from our_init_transaction_block()
 *
 * to create a separate buffer to hold unregistered small packets
 *	and then register it's memory area
 *
 * Returns 0 if memory allocated and registered ok, else -1
 */
static int
our_calloc_and_register_recv_buf(struct exs_conn *cb,
				struct transaction_block *block, size_t size)
{
	if(size <= 7)
	{	/* always allocate and register at least 8 bytes */
		size = 8;
	}

	block->recv_rdma_buf = EXS_CALLOC(1, size);
	if(block->recv_rdma_buf == NULL)
	{
		OUR_PERROR("alloc transaction recv_rdma_buf", errno);
		return -1;
	}

	TRACE("%s \tblock %p: allocated %zu bytes for recv_rdma_buf %p\n",
			_time_prefix(), block, size, block->recv_rdma_buf);

	/* register our recv buffer from which we can copy small packet data */
	if((block->recv_rdma_mr = our_call_ibv_reg_mr(cb->pd, block->recv_rdma_buf,
			size, IBV_ACCESS_LOCAL_WRITE|IBV_ACCESS_REMOTE_WRITE,
			"recv rdma buf")) == NULL)
	{
		OUR_PERROR("register transaction recv_rdma_buf", errno);
		EXS_FREE(block->recv_rdma_buf);
		block->recv_rdma_buf = NULL;		/* don't free it again */
		return -1;
	}
	return 0;
}	/* our_calloc_and_register_recv_buf */

/*
 * called from our_common_send() and our_create_transaction_block()
 *
 * to create a separate buffer to hold unregistered small packets
 *	and then register it's memory area
 *
 * Returns 0 if memory allocated and registered ok, else -1
 */
static int
our_calloc_and_register_send_buf(struct exs_conn *cb,
				struct transaction_block *block, int size)
{
	if(size <= 7)
	{	/* always allocate and register at least 8 bytes */
		size = 8;
	}

	block->send_buf = EXS_CALLOC(1, size);
	if(block->send_buf == NULL)
	{
		OUR_PERROR("alloc transaction send_buf", errno);
		return -1;
	}

	TRACE("%s \tblock %p: allocated %d bytes for send_buf %p\n",
			_time_prefix(), block, size, block->send_buf);

	/* register our send buffer into which we can copy immediate data */
	if((block->send_buf_mr = our_call_ibv_reg_mr(cb->pd, block->send_buf,
			size, IBV_ACCESS_REMOTE_READ, "send buf")) == NULL)
	{
		OUR_PERROR("register transaction send_buf", errno);
		EXS_FREE(block->send_buf);
		block->send_buf = NULL;		/* don't free it again */
		return -1;
	}
	return 0;
}	/* our_calloc_and_register_send_buf */


/* create and initialize a transaction block
 *
 * called from our_create_and_post_blocks() and our_nego_init()
 *
 * no locks needed for this function
 *
 * Returns pointer to newly created and initialized block if all ok, else NULL
 */
static struct transaction_block *
our_create_transaction_block(struct exs_conn *cb, uint32_t sp_size,
						uint32_t copy_unreg_on_send)
{
	struct transaction_block *block;
	struct ibv_mr            *mr;

	block = EXS_CALLOC(1, sizeof(struct transaction_block) + sp_size);
	if(block == NULL)
	{
		OUR_PERROR("alloc transaction block", errno);
		goto err0;
	}

	TRACE("%s \tblock %p: allocated %lu bytes\n", _time_prefix(),
			block, sizeof(struct transaction_block) + sp_size);

	/* register this whole block as a local-read-only memory region */
	if((mr = our_call_ibv_reg_mr(cb->pd,
			block, sizeof(struct transaction_block) + sp_size,
			0, "block")) == NULL)
	{
		goto err1;
	}

	if (!our_init_transaction_block(cb, sp_size, copy_unreg_on_send, block, mr)) {
		goto err2;
	}

	return block;
err2:
	ibv_dereg_mr(block->block_mr);
err1:
	EXS_FREE(block);
err0:
	return NULL;
}

static struct transaction_block *
our_init_transaction_block(struct exs_conn *cb, uint32_t sp_size,
		uint32_t copy_unreg_on_send, struct transaction_block *block,
		struct ibv_mr *mr)
{
	struct ibv_mr		*recv_mr;
	block->block_mr = mr;

	if (copy_unreg_on_send)
	{	/* we now allocate and preregister a send buffer for
		 * either immediate data sent in small packets
		 */
		if(our_calloc_and_register_send_buf(cb, block, sp_size) != 0)
			goto err2;
		if(our_calloc_and_register_recv_buf(cb, block, sp_size) != 0)
			goto err2;
	}
	/* else both send_buf and send_buf_mr remain 0 (NULL) */

	/* set up send work request to point to 2-element s/g list */
	block->send_wr.wr_id = (uint64_t)(unsigned long)block;
	block->send_wr.sg_list = block->send_sg_list;
	block->send_wr.num_sge = 2;
	block->send_wr.opcode = IBV_WR_SEND;

	/* immediates are always signaled, but can be inline or not,
	 * depending on size, so send_wr.send_flags are set dynamically to
	 * either	IBV_SEND_SIGNALED | IBV_SEND_INLINE
	 * or		IBV_SEND_SIGNALED
	 */

	TRACE("%s \tblock %p: send_wr %p points at sg_list %p, "
		"with %lu byte send_header %p\n", _time_prefix(), block,
		&block->send_wr, block->send_sg_list,
		sizeof(block->send_header), &block->send_header);

	/* point send s/g list element to point to send header */
	block->send_sg_list[0].addr
			= (uint64_t)(unsigned long)&block->send_header;
	block->send_sg_list[0].length = sizeof(block->send_header);
	block->send_sg_list[0].lkey = block->block_mr->lkey;

	/* point 2nd send s/g list element at advert buf for now */
	block->send_sg_list[1].addr
			= (uint64_t)(unsigned long)&block->advert_buf;
	block->send_sg_list[1].length = sizeof(block->advert_buf);
	block->send_sg_list[1].lkey = block->block_mr->lkey;

	TRACE("%s \tblock %p: recv_wr %p points at sg_list %p, "
		"with %lu byte recv_header %p, %lu bytes of data\n",
		_time_prefix(), block, &block->recv_wr, block->recv_sg_list,
		sizeof(block->recv_header), &block->recv_header,
		sizeof(struct our_advert) + sp_size);

	/* set up receive work request to point to 1-element s/g list */
	block->recv_wr.wr_id = (uint64_t)(unsigned long)block;
	block->recv_wr.sg_list = block->recv_sg_list;
	block->recv_wr.num_sge = 1;

	/* register our contiguous recv header/buffer */
	block->recv_mhandle = our_mregister(&block->recv_header,
			sizeof(block->recv_header)
				+ sizeof(block->recv_buf)
				+ sp_size,
			EXS_LOCAL_WRITE, "recv header");
	if(block->recv_mhandle == EXS_MHANDLE_INVALID)
	{
		goto err3;
	}
	recv_mr = our_ensure_buffer_registered(cb, block->recv_mhandle, __func__);
	if (recv_mr == NULL) {
		/* should be able to register buffer but couldn't */
		goto err4;
	}
	/* set up recv s/g list element to point at recv header/buffer */
	block->recv_sg_list[0].addr
			= (uint64_t)(unsigned long)&block->recv_header;
	block->recv_sg_list[0].length = block->recv_mhandle->size;
	block->recv_sg_list[0].lkey = recv_mr->lkey;

	/* set up write request work request to point to a 1-element s/g list */

	TRACE("%s \tblock %p: writereq_wr %p points at sg_list %p\n",
		_time_prefix(),block,&block->writereq_wr,block->writereq_sg_list);

	if (!(cb->fd_flags & EXS_FD_NO_WRITE_WITH_IMM)) {
		block->writereq_wr.wr_id = (uint64_t)(unsigned long)block;
		block->writereq_wr.sg_list = block->writereq_sg_list;
		block->writereq_wr.num_sge = 1;
		block->writereq_wr.opcode = IBV_WR_RDMA_WRITE_WITH_IMM;
		block->writereq_wr.send_flags = IBV_SEND_SIGNALED;
	} else {
		block->writereq_wr.wr_id = (uint64_t)(unsigned long)block;
		block->writereq_wr.sg_list = block->writereq_sg_list;
		block->writereq_wr.num_sge = 1;
		block->writereq_wr.opcode = IBV_WR_RDMA_WRITE;
		block->writereq_wr.send_flags = IBV_SEND_SIGNALED;
		block->writereq_wr.next = &block->writeack_wr;

		block->writeack_wr.wr_id = (uint64_t)(unsigned long)block;
		block->writeack_wr.sg_list = block->writeack_sg_list;
		block->writeack_wr.num_sge = 2;
		block->writeack_wr.opcode = IBV_WR_SEND;
		block->writeack_wr.send_flags = IBV_SEND_SIGNALED;

		block->writeack_sg_list[0].addr
			= (uint64_t)(unsigned long)&block->writeack_header;
		block->writeack_sg_list[0].length = sizeof(block->writeack_header);
		block->writeack_sg_list[0].lkey = block->block_mr->lkey;

		block->writeack_sg_list[1].addr
			= (uint64_t)(unsigned long)&block->writeack_buf;
		block->writeack_sg_list[1].length = sizeof(block->writeack_buf);
		block->writeack_sg_list[1].lkey = block->block_mr->lkey;
	}

	/* allocate an event for this block */
	if((block->user_event_ptr = EXS_CALLOC(1, sizeof(struct exs_event)))
									== NULL)
	{
		OUR_PERROR("our_create_transaction_block: exs_event", errno);
		goto err4;
	}

	TRACE("%s %s on cb %p, "
		"return block %p with %lu byte event %p\n", _time_prefix(),
		__func__, cb, block, sizeof(struct exs_event),
		block->user_event_ptr);
	return block;

err4:
	our_mderegister(block->recv_mhandle);
err3:
	if(block->send_buf_mr)
	{
		ibv_dereg_mr(block->send_buf_mr);
		EXS_FREE(block->send_buf);
	}
	if(block->recv_rdma_mr) {
		ibv_dereg_mr(block->recv_rdma_mr);
		EXS_FREE(block->recv_rdma_buf);
	}
err2:
	return NULL;
}	/* our_create_transaction_block */


/* Allocates and registers an EXS internal buffer.  Returns 0 on success or the
 * value of errno on failure. */
static int
exs_buf_alloc(struct exs_conn *cb, struct exs_buf *buf, size_t size, char *message)
{
	int ret;

	if (buf->ptr != NULL) {
		TRACE_ERR("Attempt to allocate EXS buffer \"%s\" twice",
				message);
	}

	TRACE("%s %s calloc buffer %s size %zu\n",
			_time_prefix(), __func__, message, size);
	if (!(buf->ptr = EXS_CALLOC(1, size))) {
		ret = ENOMEM;
		goto out;
	}
	buf->size = size;

	if ((buf->mr = our_call_ibv_reg_mr(cb->pd, buf->ptr, buf->size,
					IBV_ACCESS_LOCAL_WRITE
					|IBV_ACCESS_REMOTE_WRITE,
					message)) == NULL) {
		ret = errno;
		EXS_FREE(buf->ptr);
		buf->ptr = NULL;
		goto out;
	}
	ret = 0;

out:
	return ret;
}	/* exs_buf_alloc */


/* Deregisters and frees an EXS internal buffer.  Returns 0 on success or the
 * value of errno on failure. */
static int
exs_buf_free(struct exs_buf *buf)
{
	if (buf->ptr != NULL) {
		int ret;
		ret = ibv_dereg_mr(buf->mr);
		EXS_FREE(buf->ptr);
		buf->ptr = NULL;
		return ret;
	} else {
		return 0;
	}
}	/* exs_buf_free */


/*
 * Called only in our_completion_thread() context
 * called from our_client_finishing() and our_agent_finishing()
 *
 * Assumes that send_wwi_array_mr, send_advert_array_mr, and recv_array_mr are
 * all NULL when this function is called.  They will be non-NULL if this
 * function returns successfully.
 *
 * Returns 0 if all ok, errno on error.
 */
static int
our_create_and_post_blocks(struct exs_conn *cb, const char *message)
{
	long i;
	struct ibv_mr *mr;
	(void)message;

	TRACE("%s %s: small_packet_max_size %d\n",
			_time_prefix(), message,
			cb->small_packet_max_size);

	TRACE("%s %s: fd_flags & EXS_FD_SP_COPY_UNREG_ON_SEND %d\n",
			_time_prefix(), message,
			cb->fd_flags & EXS_FD_SP_COPY_UNREG_ON_SEND);

	TRACE("\t\t%25s %d\n", "posted_credits", cb->posted_credits);

	/* register this whole set of blocks as a local-read-only memory region */
	if (!(cb->send_wwi_array = EXS_CALLOC(cb->posted_credits,
			sizeof(*cb->send_wwi_array)))) {
		goto errout;
	}
	if ((mr = our_call_ibv_reg_mr(cb->pd, cb->send_wwi_array,
			sizeof(struct transaction_block) * cb->posted_credits,
			0, "send_wwi_array")) == NULL) {
		goto errout;
	}
	for (i = 0; i < cb->posted_credits; ++i) {
		struct transaction_block *block = &cb->send_wwi_array[i];
		block->index = i;
		block->send_state = 0;
		if (i < cb->posted_credits - 1) {
			block->next = &cb->send_wwi_array[i + 1];
		} else {
			block->next = &cb->send_wwi_array[0];
		}
		if (sem_init(&block->sem, 0, 1) < 0) {
			ibv_dereg_mr(mr);
			goto errout;
		}
		if ((our_init_transaction_block(cb,
				cb->small_packet_max_size,
				cb->fd_flags & (EXS_FD_SP_COPY_UNREG_ON_SEND
						|EXS_FD_NODELAY),
				block, mr)) == NULL) {
			ibv_dereg_mr(mr);
			goto errout;
		}
	}
	cb->next_send = &cb->send_wwi_array[0];
	cb->next_send_primed = &cb->send_wwi_array[0];
	cb->next_send_unacked = &cb->send_wwi_array[0];
	cb->send_wwi_array_mr = mr;
	TRACE("%s %s: created %ld send_wwi transaction blocks on cb %p\n",
		_time_prefix(), message, i, cb);

	if (!(cb->send_advert_array = EXS_CALLOC(cb->posted_credits,
			sizeof(*cb->send_advert_array)))) {
		goto errout;
	}
	if((mr = our_call_ibv_reg_mr(cb->pd, cb->send_advert_array,
			sizeof(struct transaction_block) * cb->posted_credits,
			0, "send_advert_array")) == NULL) {
		goto errout;
	}
	for (i = 0; i < cb->posted_credits; ++i) {
		struct transaction_block *block = &cb->send_advert_array[i];
		block->index = i;
		if (i < cb->posted_credits - 1) {
			block->next = &cb->send_advert_array[i + 1];
		} else {
			block->next = &cb->send_advert_array[0];
		}
		if (sem_init(&block->sem, 0, 1) < 0) {
			ibv_dereg_mr(mr);
			goto errout;
		}
		if ((our_init_transaction_block(cb,
				cb->small_packet_max_size,
				cb->fd_flags & (EXS_FD_SP_COPY_UNREG_ON_SEND
						|EXS_FD_NODELAY),
				block, mr)) == NULL) {
			ibv_dereg_mr(mr);
			goto errout;
		}
	}
	cb->send_advert_array_mr = mr;
	TRACE("%s %s: created %ld send_advert transaction blocks on cb %p\n",
		_time_prefix(), message, i, cb);

	if (!(cb->recv_array = EXS_CALLOC(RQ_DEPTH(cb),
					sizeof(*cb->recv_array)))) {
		goto errout;
	}
	if((mr = our_call_ibv_reg_mr(cb->pd, cb->recv_array,
			sizeof(struct transaction_block) * RQ_DEPTH(cb),
			0, "recv_array")) == NULL) {
		goto errout;
	}
	for (i = 0; i < RQ_DEPTH(cb); ++i) {
		if ((our_init_transaction_block(cb,
				cb->small_packet_max_size,
				cb->fd_flags & (EXS_FD_SP_COPY_UNREG_ON_SEND
						|EXS_FD_NODELAY),
				&cb->recv_array[i], mr)) == NULL) {
			ibv_dereg_mr(mr);
			goto errout;
		}
	}
	cb->next_recv = &cb->send_advert_array[0];
	cb->next_recv_primed = &cb->send_advert_array[0];
	cb->recv_array_mr = mr;
	TRACE("%s %s: created %ld recv transaction blocks on cb %p\n",
		_time_prefix(), message, i, cb);

	if (!cb->is_seqpacket) {
		if (!(cb->recved_advert_array = EXS_CALLOC(cb->posted_credits,
					sizeof(*cb->recved_advert_array)))) {
			goto errout;
		}
		for (i = 0; i < cb->posted_credits; ++i) {
			struct stream_advert_block *block;
			block = &cb->recved_advert_array[i];
			if (i < cb->posted_credits - 1) {
				block->next = &cb->recved_advert_array[i + 1];
			} else {
				block->next = &cb->recved_advert_array[0];
			}
		}
		cb->next_advert = &cb->recved_advert_array[0];
		cb->next_advert_primed = &cb->recved_advert_array[0];
	}

	return our_post_transaction_blocks(cb);

errout:
	if (cb->send_wwi_array_mr) {
		ibv_dereg_mr(cb->send_wwi_array_mr);
		for (i = 0; i < cb->posted_credits; ++i) {
			sem_destroy(&cb->send_wwi_array[i].sem);
			our_uninit_transaction_block(&cb->send_wwi_array[i]);
		}
		EXS_FREE(cb->send_wwi_array);
	}
	if (cb->send_advert_array_mr) {
		ibv_dereg_mr(cb->send_advert_array_mr);
		for (i = 0; i < cb->posted_credits; ++i) {
			sem_destroy(&cb->send_wwi_array[i].sem);
			our_uninit_transaction_block(&cb->send_advert_array[i]);
		}
		EXS_FREE(cb->send_advert_array);
	}
	if (cb->recv_array_mr) {
		ibv_dereg_mr(cb->recv_array_mr);
		for (i = 0; i < 2 * cb->posted_credits; ++i) {
			our_uninit_transaction_block(&cb->recv_array[i]);
		}
		EXS_FREE(cb->recv_array);
	}
	if (cb->recved_advert_array) {
		EXS_FREE(cb->recved_advert_array);
	}
	return ENOMEM;
}	/* our_create_and_post_blocks */


static inline void
our_set_send_credit_limits(struct exs_conn *cb, int value)
{
	if((cb->posted_credits = value) < 2)
		cb->posted_credits = 2;
	if ((cb->signal_limit = value >> 2) < 1)
		cb->signal_limit = 1;
	TRACE("%s %s for cb %p set posted_credits %d, "
		"signal_limit %d\n", _time_prefix(), __func__, cb,
		cb->posted_credits, cb->signal_limit);
}	/* our_set_send_credit_limits */


/*
 * Called only in our_completion_thread() context
 * twice from our_completion_thread() when
 *	a) in state EXS_CONN_CLIENT_SETUP
 *	b) cb->nego_in_ptr is not NULL
 *	c) cb->nego_out_ptr is not NULL
 *
 * OR state onto cb's nego_state_info
 * and test for both flags set -- if both are set
 *	use results sent back from server
 *	allocate the blocks and post some of them
 *	post completion to the exs_connect()
 *	free the init memory
 *
 * Returns 0 if all ok, errno on error.
 */
static int
our_client_finishing(struct exs_conn *cb, int state,
			our_qhandle_t *qhandle, int *signo)
{
	int				ret;
	unsigned int			uvalue;
	struct transaction_block	*nego_in_ptr, *nego_out_ptr;
	struct our_nego_mess		*nego_mess;

	nego_in_ptr = cb->nego_in_ptr;
	nego_out_ptr = cb->nego_out_ptr;
	atomic_set_bits(state, &nego_out_ptr->send_state);
	if((nego_out_ptr->send_state & (SEND_DONE_BIT|ACK_DONE_BIT))
					== (SEND_DONE_BIT|ACK_DONE_BIT))
	{	/* got IBV_WC_SEND complete for OUR_SETUP_REQUEST
		 * and IBV_WC_RECV complete for OUR_SETUP_REPLY
		 */
		/* check values server sent back in recv_buf */
		nego_mess = (struct our_nego_mess *)nego_in_ptr->recv_buf;

		uvalue = ntohl(nego_mess->client_credits);
		if (cb->posted_credits != uvalue) {
			TRACE_ERR("client's credits %u, expected %u",
				uvalue, cb->posted_credits);
			ret = EPROTO;
			goto exit0;
		}

		uvalue = ntohl(nego_mess->sp_max_size);
		if (cb->small_packet_max_size != uvalue) {
			TRACE_ERR("client's small packet_max size %u, expected %u",
				uvalue, cb->small_packet_max_size);
			ret = EPROTO;
			goto exit0;
		}

		nego_mess = (struct our_nego_mess *)nego_in_ptr->recv_buf;
		our_set_send_credit_limits(cb,
					ntohl(nego_mess->client_credits));
		cb->small_packet_max_size = ntohl(nego_mess->sp_max_size);

		our_connect_done(cb, qhandle, signo);
	}

	ret = 0;

exit0:
	return ret;
}	/* our_client_finishing */


static inline void
our_fill_client_nego_mess(struct exs_conn *cb, struct our_nego_mess *nego_send)
{
	nego_send->client_credits = htonl(cb->posted_credits);
	nego_send->reserved0 = 0;
	nego_send->sp_max_size = htonl(cb->small_packet_max_size);
	nego_send->version = EXS_VERSION;
	nego_send->subversion = EXS_SUBVERSION;
	nego_send->subsubversion = EXS_SUBSUBVERSION;
	nego_send->is_seqpacket = cb->is_seqpacket;
	nego_send->fd_flags = cb->fd_flags;
	if (!cb->is_seqpacket && !(cb->fd_flags & EXS_FD_NODELAY)) {
		nego_send->recv_buf_addr
			= htonll((uint64_t)cb->recv_buf.buf.ptr);
		nego_send->recv_buf_rkey = htonl(cb->recv_buf.buf.mr->rkey);
		nego_send->recv_buf_len = htonl(cb->recv_buf.buf.size);
	}
}	/* our_fill_client_nego_mess */


/*	agent's recv credits become client's send credits
 *	agent's send credits become client's recv credits
 */
static inline void
our_setup_agent_nego_mess(struct exs_conn *cb, struct our_nego_mess *nego_send)
{
	nego_send->client_credits = htonl(cb->posted_credits);
	nego_send->reserved0 = 0;
	nego_send->sp_max_size = htonl(cb->small_packet_max_size);
	nego_send->version = EXS_VERSION;
	nego_send->subversion = EXS_SUBVERSION;
	nego_send->subsubversion = EXS_SUBSUBVERSION;
	nego_send->is_seqpacket = cb->is_seqpacket;
	nego_send->fd_flags = cb->fd_flags;
	if (!cb->is_seqpacket && !(cb->fd_flags & EXS_FD_NODELAY)) {
		nego_send->recv_buf_addr
			= htonll((uint64_t)cb->recv_buf.buf.ptr);
		nego_send->recv_buf_rkey = htonl(cb->recv_buf.buf.mr->rkey);
		nego_send->recv_buf_len = htonl(cb->recv_buf.buf.size);
	}
}	/* our_setup_agent_nego_mess */


/*
 * called only from our_connect_finishing()
 *
 * post recv into recv_buf,
 * then store parameter values into the send_nego_mess and send it
 *
 * Returns 0 if all ok, else the error number
 */
static int
our_client_nego_launch(struct exs_conn *cb)
{
	int				ret;
	struct transaction_block	*nego_in_ptr, *nego_out_ptr;
	struct our_nego_mess		*nego_send;
	struct ibv_send_wr		*send_bad_wr;

	nego_in_ptr = cb->nego_in_ptr;
	nego_out_ptr = cb->nego_out_ptr;

	/* mark that this send is a setup request */
	nego_out_ptr->recv_header.hd_type = OUR_SETUP_REQUEST;
	nego_out_ptr->send_header.hd_type = htonl(OUR_SETUP_REQUEST);
	nego_out_ptr->send_header.hd_seqno = 0;

	nego_send = (struct our_nego_mess*)nego_out_ptr->send_buf;
	our_fill_client_nego_mess(cb, nego_send);

	/* dump everything being sent -- in hex, 16 bytes/line */
	TRACE("\t\t%4d: dump of %d nego header bytes being sent "
			"from addr %p\n", __LINE__,
			nego_out_ptr->send_sg_list[0].length,
			(void *)nego_out_ptr->send_sg_list[0].addr);
	DUMP_BUF((void *)(unsigned long)nego_out_ptr->send_sg_list[0].addr,
					nego_out_ptr->send_sg_list[0].length);

	TRACE("\t\t%4d: dump of %d nego data bytes being sent "
			"from addr %p\n", __LINE__,
			nego_out_ptr->send_sg_list[1].length,
			(void *)nego_out_ptr->send_sg_list[1].addr);
	DUMP_BUF((void *)(unsigned long)nego_out_ptr->send_sg_list[1].addr,
					nego_out_ptr->send_sg_list[1].length);

	/* set expected completion code SEND,
	 * always send this signaled, inline
	 */
	nego_out_ptr->expected_comp = IBV_WC_SEND;
	nego_out_ptr->send_wr.send_flags = IBV_SEND_SIGNALED | IBV_SEND_INLINE;

	/* Before sending the negotiation packet, we must post recvs to catch
	 * advertisements for exs_recvs on the remote side */
	ret = our_create_and_post_blocks(cb, __func__);
	if (ret != 0) {
		OUR_PERROR("failed create and post blocks", ret);
		return ret;
	}

	if((ret = ibv_post_send(cb->qp, &nego_out_ptr->send_wr, &send_bad_wr)))
	{	/* ibv_post_send failed */
		OUR_PERROR("failed ibv_post_send", ret);
		return ret;
	}
	STATISTICS_INCREMENT(total_setup_sends);
	TRACE("%s %s: ibv_post_send ok for nego_out_ptr->send_wr %p, "
		"send_wr.wr_id %llx\n", _time_prefix(), __func__,
		&nego_out_ptr->send_wr,
		(long long unsigned)nego_out_ptr->send_wr.wr_id);
	return 0;
}	/* our_client_nego_launch */


/*
 * Called only in our_completion_thread() context for a server connection
 *
 * called only from our_pending_accept_done()
 *	(which is only called directly from our_completion_thread()
 *	when it gets successful IBV_WC_SEND
 *	of OUR_SETUP_REPLY that agent sent to client)
 *
 * fill in and post event structure for the exs_accept
 * then free new_pending
 *
 * Returns 0 if all ok, -1 on error.
 */
static void
our_fill_in_accept_event(struct pending_accept *new_pending, int listen_socket)
{

	TRACE("%s %s: Building completion event %p for "
		"accepted connection %p in state %s(%d)\n", _time_prefix(),
		__func__, new_pending->evt_item, new_pending->accepted_conn,
		state_labels[new_pending->accepted_conn->conn_state],
		new_pending->accepted_conn->conn_state);

	memset(new_pending->evt_item, 0, sizeof(struct exs_event));
	new_pending->evt_item->exs_evt_type = EXS_EVT_ACCEPT;
	new_pending->evt_item->exs_evt_socket = listen_socket;
	new_pending->evt_item->exs_evt_union.exs_evt_accept.exs_evt_new_socket
					= new_pending->accepted_conn->sockno;
	TRACE("%s %s: listen_socket %d, accepted_socket %d, "
		"addrvec %p\n", _time_prefix(), __func__,
		listen_socket, new_pending->accepted_conn->sockno,
		new_pending->addrvec);
	if(new_pending->addrvec != NULL)
	{
		TRACE("%s %s: exs_ahandle %p, exs_addr %p, "
			"exs_addrlen %d\n", _time_prefix(), __func__,
			new_pending->addrvec->exs_ahandle,
			new_pending->addrvec->exs_addr,
			new_pending->addrvec->exs_addrlen);
		new_pending->evt_item->exs_evt_ahandle
					= new_pending->addrvec->exs_ahandle;
		if((new_pending->evt_item->
				exs_evt_union.exs_evt_accept.exs_evt_addr
				= new_pending->addrvec->exs_addr) != NULL)
		{
			/* get correct sockaddr of remote client */
			new_pending->evt_item->
				exs_evt_union.exs_evt_accept.exs_evt_addrlen
				= our_copy_dst_addr(
					new_pending->accepted_conn->cm_id,
					new_pending->addrvec->exs_addr,
					new_pending->addrvec->exs_addrlen);
		}
	}

	TRACE("%s %s: Posting completion event %p for "
			"accepted connection %p\n", _time_prefix(), __func__,
			new_pending->evt_item, new_pending->accepted_conn);

	_put_full_evt_queue_event(new_pending->qhandle, new_pending->evt_item);

	EXS_FREE(new_pending);

}	/* our_fill_in_accept_event */


/*
 * Called only in our_completion_thread() context for an agent connection
 * called only from our_completion_thread()
 *
 * when it gets successful IBV_WC_SEND
 * of OUR_SETUP_REPLY that agent sent to client
 *
 * Returns 0 if all ok, -1 on error.
 */
static int
our_pending_accept_done(struct exs_conn *cb)
{
	struct pending_accept	*new_pending;

	TRACE("%s %s:  in, cb %p, state %s(%d)\n", _time_prefix(),
		__func__, cb, state_labels[cb->conn_state], cb->conn_state);

	new_pending = cb->accept_pending;
#if defined(ALLOW_PARANOID_CHECKING)
	if(new_pending == NULL)
	{	/* should never happen */
		TRACE_ERR("cb %p, state %s(%d), accept_pending NULL",
			cb, state_labels[cb->conn_state],
			cb->conn_state);
		return -1;
	}
#endif

	if(cb->nego_out_ptr != NULL)
	{	/* need to free nego block */
		TRACE("%s %s: free nego out transaction block %p\n",
			_time_prefix(), __func__, cb->nego_out_ptr);
		our_free_transaction_block(cb->nego_out_ptr);
		cb->nego_out_ptr = NULL;	/* so we don't free it again */
	}


	cb->conn_state = EXS_CONN_RUNNING;
	cb->event_expected = RDMA_CM_EVENT_DISCONNECTED;

	/* fills in and posts accept event, then frees new_pending */
	our_fill_in_accept_event(new_pending, cb->accept_parent->sockno);
	cb->accept_pending = NULL;		/* so we don't free it again */
	TRACE("%s %s: out, cb %p, state %s(%d), returning 0\n",
		_time_prefix(), __func__, cb, state_labels[cb->conn_state],
		cb->conn_state);

	return 0;				/* all ok */
}	/* our_pending_accept_done */


/*
 * Called only in our_completion_thread() context
 * called only from our_completion_thread() when
 *	a) got IBV_WC_RECV complete for OUR_SETUP_REQUEST
 *	b) in state EXS_CONN_AGENT_SETUP
 *	c) cb->nego_in_ptr is not NULL
 *	d) cb->nego_out_ptr is not NULL
 *
 * Uses negotiated values sent by client to calculate minimums
 * Then uses these minimums to allocate empty transaction blocks
 *	and to post some of them
 * Then sets up and sends OUR_SETUP_REPLY back to client with new values
 *
 * Returns 0 if all ok, else error number
 */
static int
our_agent_finishing(struct exs_conn *cb )
{
	int				ret;
	unsigned int			uvalue;
	struct transaction_block	*nego_in_ptr, *nego_out_ptr;
	struct our_nego_mess		*nego_mess;
	struct ibv_send_wr		*send_bad_wr;

	nego_in_ptr = cb->nego_in_ptr;
	nego_mess = (struct our_nego_mess *)nego_in_ptr->recv_buf;

	/* client's send credits should equal agent's recv credits */
	uvalue = ntohl(nego_mess->client_credits);
	if (cb->posted_credits != uvalue) {
		TRACE_ERR("client's credits %u, expected %u",
			uvalue, cb->posted_credits);
		ret = EPROTO;
		goto exit0;
	}

	uvalue = ntohl(nego_mess->sp_max_size);
	if (cb->small_packet_max_size != uvalue) {
		TRACE_ERR("client's small packet_max size %u, expected %u",
			uvalue, cb->small_packet_max_size);
		ret = EPROTO;
		goto exit0;
	}

	/* now set up and send the SETUP_REPLY packet */
	nego_out_ptr = cb->nego_out_ptr;

	/* mark that this send is a setup_reply */
	nego_out_ptr->recv_header.hd_type = OUR_SETUP_REPLY;
	nego_out_ptr->send_header.hd_type = htonl(OUR_SETUP_REPLY);
	nego_out_ptr->send_header.hd_seqno = 0;

	nego_mess = (struct our_nego_mess *)nego_out_ptr->send_buf;
	our_setup_agent_nego_mess(cb, nego_mess);

	/* dump everything being sent -- in hex, 16 bytes/line */
	TRACE("\t\t%4d: dump of %d nego header bytes being sent "
			"from addr %p\n", __LINE__,
			nego_out_ptr->send_sg_list[0].length,
			(void *)nego_out_ptr->send_sg_list[0].addr);
	DUMP_BUF((void *)(unsigned long)nego_out_ptr->send_sg_list[0].addr,
					nego_out_ptr->send_sg_list[0].length);

	TRACE("\t\t%4d: dump of %d nego data bytes being sent "
			"from addr %p\n", __LINE__,
			nego_out_ptr->send_sg_list[1].length,
			(void *)nego_out_ptr->send_sg_list[1].addr);
	DUMP_BUF((void *)(unsigned long)nego_out_ptr->send_sg_list[1].addr,
					nego_out_ptr->send_sg_list[1].length);

	/* set expected completion code SEND, always send this signaled */
	nego_out_ptr->expected_comp = IBV_WC_SEND;
	nego_out_ptr->send_wr.send_flags = IBV_SEND_SIGNALED | IBV_SEND_INLINE;

	if((ret = ibv_post_send(cb->qp, &nego_out_ptr->send_wr, &send_bad_wr)))
	{	/* ibv_post_send failed */
		OUR_PERROR("failed ibv_post_send", ret);
		goto exit0;
	}
	STATISTICS_INCREMENT(total_setup_sends);
	TRACE("%s %s: ibv_post_send ok for nego_out_ptr->send_wr %p, "
		"send_wr.wr_id %llx\n", _time_prefix(), __func__,
		&nego_out_ptr->send_wr,
		(long long unsigned)nego_out_ptr->send_wr.wr_id);

	ret = 0;
exit0:
	return ret;
}	/* our_agent_finishing */


/*
 * Called only in our_con_man_thread() context
 * called twice from our_setup_qp()
 *
 * to allocate and initialize one negotiation exchange block
 *
 * no locks needed for this function
 *
 * returns 0 if all ok, else the errno if things go bad
 */
static int
our_nego_init(struct exs_conn *cb, struct transaction_block **cb_nego_ptr)
{
	struct transaction_block	*nego_ptr;

	TRACE("%s start %s, cb %p, state %s(%d)\n",
			_time_prefix(), __func__, cb,
			state_labels[cb->conn_state], cb->conn_state);

	TRACE("%s %s: small_packet_max_size %lu\n",
			_time_prefix(), __func__, sizeof(struct our_nego_mess));

	TRACE("%s %s: sp_copy_unreg_on_send %d\n",
			_time_prefix(), __func__, EXS_FD_SP_COPY_UNREG_ON_SEND);

	nego_ptr = our_create_transaction_block(cb,
		sizeof(struct our_nego_mess),
				(EXS_FD_SP_COPY_UNREG_ON_SEND|EXS_FD_NODELAY));
	if(nego_ptr == NULL)
	{
		return ENOMEM;
	}
	*cb_nego_ptr = nego_ptr;

	/* point 2nd send s/g list element to point to send_buf message */
	nego_ptr->send_sg_list[1].addr
		= (uint64_t)(unsigned long)nego_ptr->send_buf;
	nego_ptr->send_sg_list[1].length
		= sizeof(struct our_nego_mess);
	nego_ptr->send_sg_list[1].lkey
		= nego_ptr->send_buf_mr->lkey;

	return 0;
}	/* our_nego_init */


/*
 * called only from our_completion_thread() and our_try_to_disconnect()
 *
 * returns 0 on success, errno on error after printing error message
 */
static int
our_rdma_disconnect(struct exs_conn *this_conn, const char *message)
{
	int	ret;
	char	errbuf[_MAX_STRERROR_MALLOC];

	TRACE("%s [%lu] %s: calling rdma_disconnect(%p) for cb %p,"
		" new state %s(%d), new conn_flags %0x\n",
		_time_prefix(), pthread_self(), message, this_conn->cm_id,
		this_conn, state_labels[this_conn->conn_state],
		this_conn->conn_state, this_conn->conn_flags);

	ret = rdma_disconnect(this_conn->cm_id);
	if(ret != 0)
	{	/* the rdma_disconnect() didn't work, find out why */
		TRACE("%s %s: rdma_disconnect cb %p returned %d, "
			"errno %d, state %s(%d), conn_flags %0x\n",
			_time_prefix(), message, this_conn, ret, errno,
			state_labels[this_conn->conn_state],
			this_conn->conn_state, this_conn->conn_flags);

		if(errno == EAGAIN || errno == EINVAL)
		{	/* our_con_man_thread waiting to disconnect */
			ret = 0;
		}
		else
		{
			if (strerror_r(errno, errbuf, _MAX_STRERROR_MALLOC)) {
				snprintf(errbuf,_MAX_STRERROR_MALLOC,"unknown");
			}
			TRACE_ERR("%s: rdma_disconnect cm_id %p failed, state %s(%d), conn_flags %0x, error %d: %s",
				message, this_conn->cm_id,
				state_labels[this_conn->conn_state],
				this_conn->conn_state,
				this_conn->conn_flags,
				errno, errbuf);
		}
	}

	return ret;
}	/* our_rdma_disconnect */


/*
 * called only if we get canceled in our_completion_thread()
 * when waiting to get the mutex, which is explicitly NOT a cancelation point!
 */
void
our_completion_cancel_handler(void *ptr)
{
	TRACE("%s [%lu] %s Leave %s\n",
		_time_prefix(), pthread_self(), "our_completion_thread",
		__func__);

	pthread_mutex_unlock((pthread_mutex_t *)ptr);
}	/* our_completion_cancel_handler */


static int
our_repost_recv_block(struct exs_conn *cb, struct transaction_block *block,
		struct ibv_wc *wc, char *errbuf, const char *message,
		int *do_eof_broadcast)
{
	int ret = 0;

	if (wc->status == IBV_WC_SUCCESS) {
		/* status must be IBV_WC_SUCCESS, post
		 * another recv block NOW, before we
		 * send any ACKs that might provoke
		 * further sends that we have to recv */

		if(cb->conn_flags & EXS_NEED_DISCONNECT) {
			ret = our_try_to_post_recv_block(cb, block,
							message);
			if (ret) {
				if (strerror_r(ret, errbuf,
						_MAX_STRERROR_MALLOC)) {
					snprintf(errbuf,
						_MAX_STRERROR_MALLOC,
						"unknown");
				}
				TRACE_ERR("%s our_try_to_post_recv_blk failed, error %d: %s",
					message, ret, errbuf);

				/* wake up any thread waiting to close
				 * down */
				*do_eof_broadcast = 1;
			}
		} else {
			TRACE("%s %s our_try_to_post"
					"_recv_blk not called "
					"due to connection "
					"disconnected, "
					"state %d, "
					"conn_flags %0x\n",
					_time_prefix(), message,
					cb->conn_state,
					cb->conn_flags);
		}
	}

	return ret;
}

static int
our_handle_recv_advert_seqpacket(struct exs_conn *cb,
				struct transaction_block *block,
				struct ibv_wc *wc,
				char *errbuf, const char *message,
				int *do_eof_broadcast, int *force_exit,
				our_qhandle_t *qhandle, int *signo)
{
	struct our_advert		*ad_ptr;
	struct transaction_block	*this;
	int				ret;

	ret = 0;

	/* just got an advert */
	STATISTICS_INCREMENT(recv_advert_completions);
	ad_ptr = (struct our_advert *)block->recv_buf;

	this = cb->next_send_primed;
	this->advert_buf.ad_buf_addr = ntohll(ad_ptr->ad_buf_addr);
	this->advert_buf.ad_buf_key = ntohl(ad_ptr->ad_buf_key);
	this->advert_buf.ad_buf_len = ntohl(ad_ptr->ad_buf_len);
	this->advert_buf.ad_buf_index = ntohl(ad_ptr->ad_buf_index);

	ret = our_repost_recv_block(cb, block, wc, errbuf, message,
						do_eof_broadcast);

	if (atomic_set_bits(ADVERT_READY_BIT, &this->send_state)
					& SEND_READY_BIT) {
		TRACE("%s %s advert and send ready\n",
				_time_prefix(), message);
		if (cb->xmit_if_match(cb, this,
				qhandle, signo, message)) {
			OUR_PERROR("our_common_send error", errno);
			*force_exit = 1;
		}
	} else {
		TRACE("advert ready only\n");
	}

	TRACE("%s %s advance next_send_primed\n",
			_time_prefix(), __func__);
	cb->next_send_primed = this->next;

	return ret;
}	/* our_handle_recv_advert_seqpacket */

static int
our_handle_recv_advert_stream(struct exs_conn *cb,
				struct transaction_block *block,
				struct ibv_wc *wc,
				char *errbuf, const char *message,
				int *do_eof_broadcast, int *force_exit,
				our_qhandle_t *qhandle, int *signo)
{
	struct our_advert		*ad_ptr;
	struct stream_advert_block	*this;
	int				ret;

	ret = 0;

	/* just got an advert */
	STATISTICS_INCREMENT(recv_advert_completions);
	ad_ptr = (struct our_advert *)block->recv_buf;

	this = cb->next_advert;
	cb->next_advert = this->next;
	this->advert_buf.ad_buf_seqno = ntohll(ad_ptr->ad_buf_seqno);
	this->advert_buf.ad_buf_addr = ntohll(ad_ptr->ad_buf_addr);
	this->advert_buf.ad_buf_key = ntohl(ad_ptr->ad_buf_key);
	this->advert_buf.ad_buf_len = ntohl(ad_ptr->ad_buf_len);
	this->advert_buf.ad_buf_index = ntohl(ad_ptr->ad_buf_index);
	this->advert_buf.ad_buf_flags = ntohl(ad_ptr->ad_buf_flags);
	this->offset = 0;

	ret = our_repost_recv_block(cb, block, wc, errbuf, message,
						do_eof_broadcast);

	atomic_set_bits(ADVERT_READY_BIT, &this->advert_state);
	if (cb->next_send_primed->send_state & SEND_READY_BIT) {
		TRACE("%s %s send %p advert %p advert and send ready\n",
				_time_prefix(), message, cb->next_send_primed,
				this);
		if (cb->xmit_if_match(cb, cb->next_send_primed, qhandle,
							signo, message)) {
			OUR_PERROR("our_common_send error", errno);
			*force_exit = 1;
		}
	} else {
		TRACE("%s %s send %p advert %p ready only\n",
				_time_prefix(), message, cb->next_send_primed,
				this);
	}

	return ret;
}	/* our_handle_recv_advert_stream */

static int
our_handle_recv_advert_streamcopy(struct exs_conn *cb,
				struct transaction_block *block,
				struct ibv_wc *wc,
				char *errbuf, const char *message,
				int *do_eof_broadcast, int *force_exit,
				our_qhandle_t *qhandle, int *signo)
{
	struct our_advert		*ad_ptr;
	struct stream_advert_block	*this;
	uint64_t			seqno, send_seqno;
	int				use_advert;
	int				ret;

	ret = 0;

	/* just got an advert */
	STATISTICS_INCREMENT(recv_advert_completions);
	ad_ptr = (struct our_advert *)block->recv_buf;

	seqno = ntohll(ad_ptr->ad_buf_seqno);
	send_seqno = cb->send_seqno;
	TRACE_DBG(EXS_LOG_STREAM,
		"got advert with seqno=%" PRIu64 "; current seqno=%" PRIu64,
		seqno, send_seqno);
	use_advert = (seqno >= send_seqno);
	if (use_advert) {
		this = cb->next_advert;
		cb->next_advert = this->next;
		this->advert_buf.ad_buf_seqno = seqno;
		this->advert_buf.ad_buf_addr = ntohll(ad_ptr->ad_buf_addr);
		this->advert_buf.ad_buf_key = ntohl(ad_ptr->ad_buf_key);
		this->advert_buf.ad_buf_len = ntohl(ad_ptr->ad_buf_len);
		this->advert_buf.ad_buf_index = ntohl(ad_ptr->ad_buf_index);
		this->advert_buf.ad_buf_flags = ntohl(ad_ptr->ad_buf_flags);
		this->offset = 0;
		atomic_set_bits(ADVERT_READY_BIT, &this->advert_state);
	}

	ret = our_repost_recv_block(cb, block, wc, errbuf, message,
						do_eof_broadcast);

	if (!use_advert) {
		goto out0;
	}

	if (cb->next_send_primed->send_state & SEND_READY_BIT) {
		TRACE_DBG(EXS_LOG_STREAM, "advert and send ready");
		if (cb->xmit_if_match(cb, cb->next_send_primed, qhandle,
							signo, message)) {
			OUR_PERROR("our_common_send error", errno);
			*force_exit = 1;
		}
	} else {
		TRACE_DBG(EXS_LOG_STREAM, "advert ready only");
	}


out0:
	return ret;
}	/* our_handle_recv_advert_streamcopy */

static int
our_handle_send_write_seqpacket(struct exs_conn *cb,
		struct transaction_block *block, const char *message,
		our_qhandle_t *qhandle, int *signo)
{
	STATISTICS_INCREMENT(send_done_after_completion);
	*signo = our_all_done_with_block(cb, block, qhandle, 0, message);
	return 0;
}	/* our_handle_send_write_seqpacket */

static int
our_handle_send_write_stream(struct exs_conn *cb,
		struct transaction_block *block, const char *message,
		our_qhandle_t *qhandle, int *signo)
{
	if (atomic_add_reset(-1, UINT_MAX,
			&block->outstanding_completion_count) == 0) {
		/* Post the semaphore on this block to open it up for
		 * another send if and only if there are no remaining
		 * outstanding completions for this block and we have
		 * transmitted all data for this block. */
		if (block->buf_offset == block->user_event_ptr
					->exs_evt_union.exs_evt_xfer
					.exs_evt_length) {
			STATISTICS_INCREMENT(send_done_after_completion);
			*signo = our_all_done_with_block(cb, block, qhandle,
						0, message);
		}
	}

	return 0;
}	/* our_handle_send_write_stream */

static int
our_handle_send_write_streamcopy(struct exs_conn *cb,
		struct transaction_block *block, const char *message,
		our_qhandle_t *qhandle, int *signo)
{
	unsigned int all_done;

	if (atomic_add_reset(-1, UINT_MAX,
			&block->outstanding_completion_count) == 0) {
		/* Post the semaphore on this block to open it up for
		 * another send if and only if there are no remaining
		 * outstanding completions for this block and we have
		 * transmitted all data for this block. */
		TRACE("%s %s %s outstanding_completion_count %d buf_offset %zu exs_evt_length %zu\n",
				_time_prefix(), __func__, message,
				block->outstanding_completion_count,
				block->buf_offset,
				block->user_event_ptr->exs_evt_union
					.exs_evt_xfer.exs_evt_length);

		if (block->buf_offset == block->user_event_ptr
					->exs_evt_union.exs_evt_xfer
					.exs_evt_length) {
			STATISTICS_INCREMENT(send_done_after_completion);
			atomic_set_bits(SEND_DONE_BIT, &block->send_state);
		}
	}

	all_done = (SEND_DONE_BIT|ACK_DONE_BIT);
	if ((block->send_state & all_done) == all_done) {
		*signo = our_all_done_with_block(cb, block, qhandle,
							0, message);
		atomic_clear_bits(SEND_DONE_BIT|ACK_DONE_BIT,
							&block->send_state);
	}

	return 0;
}	/* our_handle_send_write_streamcopy */


static int
our_send_ack(struct exs_conn *cb, struct transaction_block *ablock,
		int *send_flags, const char *message)
{
	(void)message;

	atomic_set_bits(ACK_PENDING_BIT, &ablock->send_state);
	if (ablock->send_state & ADVERT_SENDING_BIT) {
		/* Can't send the ACK since the ADVERT is still possibly in the
		 * send queue!  Set the signaled bit as a hack since we are
		 * waiting for a signal even if it isn't the ACK signal. */
		*send_flags |= IBV_SEND_SIGNALED;
		return 0;
	}
	atomic_clear_bits(ACK_PENDING_BIT, &ablock->send_state);

	TRACE_DBG(EXS_LOG_STREAM,
		"[via %s] send ACK seqno %u n_empty %u/%u n_unacked_wwi %u/%u\n",
		message, cb->ack_seqno + cb->recv_buf.n_empty,
		cb->recv_buf.n_empty,
		cb->recv_buf.n_empty_threshold,
		cb->n_unacked_wwi,
		cb->n_unacked_wwi_threshold);

	cb->ack_seqno += cb->recv_buf.n_empty;

	/* point 2nd send s/g list element to point to ack_buf message */
	ablock->send_sg_list[1].addr = (uint64_t)(uintptr_t)&ablock->ack_buf;
	ablock->send_sg_list[1].length = sizeof(ablock->ack_buf);
	ablock->send_sg_list[1].lkey = ablock->block_mr->lkey;

	ablock->recv_header.hd_type = OUR_ACK;
	ablock->send_header.hd_type = htonl(OUR_ACK);
	ablock->send_header.hd_seqno = htonl(cb->ack_seqno);
	ablock->ack_buf.recv_buf_empty = htonl(cb->recv_buf.n_empty);
	ablock->ack_buf.recv_buf_ack_count = htonl(cb->n_unacked_wwi);

	ablock->send_wr.send_flags = *send_flags;

	if ((sizeof(ablock->send_header) + sizeof(ablock->ack_buf))
						<= cb->inline_limit) {
		ablock->send_wr.send_flags |= IBV_SEND_INLINE;
		if ((ablock->send_wr.send_flags & IBV_SEND_SIGNALED)
				|| atomic_add_reset(1, cb->signal_limit,
					&cb->advert_signal_count) == 0) {
			/* time to force a periodic signal */
			TRACE("%s force a signal, advert_signal_count %d, "
				"signal_limit %d\n", __func__,
				cb->advert_signal_count, cb->signal_limit);
			ablock->send_wr.send_flags |= IBV_SEND_SIGNALED;
		}
	} else {
		ablock->send_wr.send_flags |= IBV_SEND_SIGNALED;
	}

	/* now post send of this advert, then leave */
	STATISTICS_INCREMENT(ack_blocks_sent);
	TRACE_DBG(EXS_LOG_UNSPECIFIED,
			"send ack: seqno %" PRIu64 " n_empty %u unacked %u",
			cb->source_seqno,
			cb->recv_buf.n_empty, cb->n_unacked_wwi);
	if (our_post_send(cb, &ablock->send_wr) < 0) {
		/* got an input/output error, so ACK was NOT SENT.
		 * Post the event for this block anyways. */
		return -1;
	}

	*send_flags = ablock->send_wr.send_flags;

	cb->recv_buf.n_empty = 0;
	cb->n_unacked_wwi = 0;

	return 0;
}	/* our_send_ack */


/* Copies up to length bytes from the receive buffer into the given user buffer.
 * Meant to be called directly from exs_recv().  Returns the number of bytes
 * copied, which may be 0, in which case nothing was copied because either no
 * data was available or MSG_WAITALL was set and insufficient data was
 * available. */
static ssize_t
our_stream_copy_recv(struct exs_conn *cb, void *buffer, size_t length,
		int flags, const char *message)
{
	size_t min_length;
	size_t copy_length;
	(void)message;

	min_length = (flags & MSG_WAITALL) ? length : 1;
	if (cb->conn_state != EXS_CONN_RUNNING) {
		TRACE_DBG(EXS_LOG_STREAM,
			"[via %s] conn_state %#x, changing min_length from %zu to 1",
			message, cb->conn_state, min_length);
		min_length = 1;
	}

	copy_length = length;
	if (copy_length > cb->recv_buf.n_full) {
		TRACE_DBG(EXS_LOG_STREAM,
			"[via %s] copy_length=%zu > n_full=%u",
			message, copy_length, cb->recv_buf.n_full);
		copy_length = cb->recv_buf.n_full;
	}
	if (copy_length > (cb->recv_buf.buf.size
					- cb->recv_buf.take_offset)) {
		TRACE_DBG(EXS_LOG_STREAM,
				"[via %s] memcpy would wrap around",
				message);
		copy_length = cb->recv_buf.buf.size
					- cb->recv_buf.take_offset;
	}
	if (copy_length < min_length) {
		TRACE_DBG(EXS_LOG_STREAM,
			"[via %s] copy_length %zu < min_length %zu!!! seqno=%u n_full=%u requested_length=%zu",
			message, copy_length, min_length,
			cb->ack_seqno, cb->recv_buf.n_full,
			length);
		return 0;
	}
	TRACE_DBG(EXS_LOG_STREAM,
			"[via %s] memcpy %zu bytes from %p+%#x to %p, n_full %u",
			message, copy_length,
			cb->recv_buf.buf.ptr, cb->recv_buf.take_offset,
			buffer, cb->recv_buf.n_full);

	memcpy(buffer, cb->recv_buf.buf.ptr + cb->recv_buf.take_offset,
		copy_length);

	/* Update loop/buffer variables */
	cb->recv_buf.n_full -= copy_length;
	cb->recv_buf.n_empty += copy_length;
	cb->recv_buf.take_offset += copy_length;
	if (cb->recv_buf.take_offset >= cb->recv_buf.buf.size) {
		/* Should never exceed, but putting greater than just in
		 * case */
		cb->recv_buf.take_offset -= cb->recv_buf.buf.size;
	}

	return copy_length;
}	/* our_stream_copy_recv */


/* Satisfies all pending recv blocks which may be satisfied by data in the
 * receive buffer. */
static int
our_stream_recv_loop(struct exs_conn *cb, our_qhandle_t *qhandle, int *signo,
		const char *message)
{
	struct transaction_block *ablock;
	struct exs_evt_xfer *evt;
	ssize_t copied_length;
	int send_flags;
	int ret;

	/* Satisfy any receives we can with this new chunk of data */
	OUR_MEMBER_MUTEX_LOCK(cb, stream_sync_mutex);
	ablock = cb->next_recv_primed;
	while (cb->recv_buf.n_full > 0 && (ablock->send_state & RECV_READY_BIT)) {
		evt = &ablock->user_event_ptr->exs_evt_union.exs_evt_xfer;
		copied_length = our_stream_copy_recv(cb, evt->exs_evt_buffer,
				evt->exs_evt_length - ablock->recv_byte_len,
				ablock->buffer_flags, message);
		if (copied_length <= 0) {
			/* Nothing was copied either because nothing was
			 * available or MSG_WAITALL was set */
			break;
		}

		/* Update the block state since the block was satisfied */
		atomic_clear_bits(RECV_READY_BIT, &ablock->send_state);
		ablock->recv_byte_len += copied_length;

		/* Send an ACK if we need to */
		send_flags = 0;
		if ((cb->recv_buf.n_empty > cb->recv_buf.n_empty_threshold)
				|| (cb->n_unacked_wwi
					> cb->n_unacked_wwi_threshold)) {
			/* We will free the transaction block after we get the
			 * SEND completion for this ACK */
			ret = our_send_ack(cb, ablock, &send_flags, message);
			if (ret < 0 && errno != ENOTCONN) {
				OUR_PERROR("Could not send ACK", errno);
				*signo = -1;
				break;
			}
		}

		*signo = our_exs_recv_done(cb, ablock, qhandle,
				send_flags & IBV_SEND_SIGNALED);

		cb->next_recv_primed = ablock->next;
		ablock = ablock->next;
	}
	OUR_MEMBER_MUTEX_UNLOCK(cb, stream_sync_mutex);

	return 0;
}	/* our_stream_recv_loop */

static int
our_handle_recv_write_notify_seqpacket(struct exs_conn *cb,
		uint32_t byte_len, uint32_t imm_data, uint32_t imm_flags,
		our_qhandle_t *qhandle, int *signo)
{
	struct transaction_block *this;
	(void)imm_data;
	(void)imm_flags;

	this = cb->next_recv_primed;
	this->recv_byte_len = byte_len;

	/* TODO: fix error reporting */
	*signo = our_exs_recv_done(cb, this, qhandle, 0);
	cb->next_recv_primed = this->next;

	return *signo;
}	/* our_handle_recv_write_notify_seqpacket */

static int
our_handle_recv_write_notify_stream(struct exs_conn *cb,
		uint32_t byte_len, uint32_t imm_data, uint32_t imm_flags,
		our_qhandle_t *qhandle, int *signo)
{
	struct transaction_block *this;
	struct exs_evt_xfer *evt;
	(void)imm_data;
	(void)imm_flags;

	this = cb->next_recv_primed;
	evt = &this->user_event_ptr->exs_evt_union.exs_evt_xfer;
	this->recv_byte_len += byte_len;
	cb->recv_seqno += byte_len;

	if (!(this->buffer_flags & MSG_WAITALL)
			|| this->recv_byte_len >= evt->exs_evt_length) {
		*signo = our_exs_recv_done(cb, this, qhandle, 0);
		cb->next_recv_primed = this->next;
		return *signo;
	} else {
		return (*signo = 0);
	}
}	/* our_handle_recv_write_notify_stream */

static int
our_handle_recv_write_notify_streamcopy(struct exs_conn *cb,
		uint32_t byte_len, uint32_t imm_data, uint32_t imm_flags,
		our_qhandle_t *qhandle, int *signo)
{
	struct transaction_block *this;
	uint32_t last_seqno = cb->recv_seqno & IMM_INDEX_MASK;
	struct exs_evt_xfer *evt;
	int send_flags;

	if (imm_data != last_seqno) {
		TRACE_ERR("seqno of received RDMA_WRITE %"PRIu32" does not match expected value %"PRIu32,
				imm_data, last_seqno);
	}
	cb->recv_seqno += byte_len;

	if (imm_flags & IMM_DIRECT_BIT) {
		/* FIXME: check the sequence number in imm_data */
		this = cb->next_recv_primed;
		evt = &this->user_event_ptr->exs_evt_union.exs_evt_xfer;
		this->recv_byte_len += byte_len;

		OUR_MEMBER_MUTEX_LOCK(cb, stream_sync_mutex);
		++cb->n_unacked_wwi;
		if (cb->n_unacked_wwi > cb->n_unacked_wwi_threshold) {
			/* We will free the transaction block after we
			 * get the SEND completion for this ACK */
			send_flags = 0;
			int ret = our_send_ack(cb, this, &send_flags, __func__);
			if (ret < 0 && errno != ENOTCONN) {
				OUR_PERROR("Could not send ACK", errno);
				*signo = -1;
			} else if (ret >= 0) {
				*signo = 0;
			}
		}
		OUR_MEMBER_MUTEX_UNLOCK(cb, stream_sync_mutex);

		if (!(this->buffer_flags & MSG_WAITALL)
				|| this->recv_byte_len >= evt->exs_evt_length) {
			atomic_clear_bits(RECV_READY_BIT, &this->send_state);
			cb->next_recv_primed = this->next;
			*signo = our_exs_recv_done(cb, this, qhandle,
					send_flags & IBV_SEND_SIGNALED);
			return *signo;
		} else {
			if (*signo >= 0) {
				return (*signo = 0);
			} else {
				return *signo;
			}
		}
	} else {
		OUR_MEMBER_MUTEX_LOCK(cb, stream_sync_mutex);
		cb->recv_buf.n_full += byte_len;
		++cb->n_unacked_wwi;
		OUR_MEMBER_MUTEX_UNLOCK(cb, stream_sync_mutex);

		return our_stream_recv_loop(cb, qhandle, signo, __func__);
	}
}	/* our_handle_recv_write_notify_streamcopy */


/* Clean up recved ACKs. */
static int
our_handle_recv_ack(struct exs_conn *cb, uint64_t seqno,
		struct our_stream_ack *recv_ack,
		struct transaction_block *recved,
		struct ibv_wc *wc,
		char *errbuf, const char *message,
		int *do_eof_broadcast, int *force_exit,
		our_qhandle_t *qhandle, int *signo)
{
	struct transaction_block *block;
	int ret;
	uint32_t new_n_empty = ntohl(recv_ack->recv_buf_empty);
	unsigned int initial_n_acks = ntohl(recv_ack->recv_buf_ack_count);
	unsigned int n_acks = initial_n_acks;
	unsigned int old_remote_buf_n_empty, tmp_outstanding_wwi_count;
	(void)force_exit;
	(void)seqno;

	TRACE_DBG(EXS_LOG_STREAM,
		"%s received ACK seqno %" PRIu64 " empty %" PRIu32 " ack_count %u",
		message, seqno, new_n_empty, n_acks);

	ret = our_repost_recv_block(cb, recved, wc, errbuf, message,
						do_eof_broadcast);

	/* Process all outstanding WWI ACKs */
	while (n_acks > 0) {
		block = cb->next_send_unacked;
		TRACE_DBG(EXS_LOG_STREAM,
			"processing ack for block %p send_state %u",
			block, block->send_state);
		if (!block->in_progress) {
			TRACE_ERR("block %p in_progress was 0 now %u send_state %u",
					block, block->in_progress, block->send_state);
			TRACE_ERR("outstanding_wwi_count %u", block->outstanding_wwi_count);
			TRACE_ERR("acking %u of %u", initial_n_acks, n_acks);
			break;
		}
		if (block->send_state & ACK_DONE_BIT) {
			TRACE_ERR("block %p already acked in_progress %u send_state %u",
					block, block->in_progress, block->send_state);
			TRACE_ERR("outstanding_wwi_count %u", block->outstanding_wwi_count);
			TRACE_ERR("acking %u of %u", initial_n_acks, n_acks);
			break;
		}
		tmp_outstanding_wwi_count = block->outstanding_wwi_count;
		if (n_acks >= tmp_outstanding_wwi_count) {
			n_acks -= tmp_outstanding_wwi_count;
			if ((atomic_add_reset(-(int)(tmp_outstanding_wwi_count),
						UINT_MAX,
						&block->outstanding_wwi_count)
						== 0)
					&& (block->buf_offset == block->user_event_ptr
						->exs_evt_union.exs_evt_xfer
						.exs_evt_length)) {
				atomic_set_bits(ACK_DONE_BIT, &block->send_state);
				if (block->send_state & SEND_DONE_BIT) {
					atomic_clear_bits(ACK_DONE_BIT
							|SEND_DONE_BIT,
						&block->send_state);
					*signo = our_all_done_with_block(cb, block,
							qhandle, 0, message);
				}
				cb->next_send_unacked = block->next;
			}
		} else {
			block->outstanding_wwi_count -= n_acks;
			n_acks = 0;
		}
	}

	/* Process data buffer ACK */
	OUR_MEMBER_MUTEX_LOCK(cb, stream_sync_mutex);
	old_remote_buf_n_empty = cb->remote_buf_n_empty;
	cb->remote_buf_n_empty += new_n_empty;
	OUR_MEMBER_MUTEX_UNLOCK(cb, stream_sync_mutex);
	if (old_remote_buf_n_empty == 0 && new_n_empty > 0) {
		cb->xmit_if_match(cb, cb->next_send_primed,
					qhandle, signo, message);
	}

	return ret;
}


static int
our_deal_with_completion(struct ibv_wc *wc, struct exs_conn *cb, char *errbuf,
								char *message)
{
	int			force_exit;
	int			ret, signo, err;
	int			do_eof_broadcast;
	int			send_flags;
	struct transaction_block *block;
	our_qhandle_t		qhandle;
	enum our_message_header_types rtype;
	struct our_notify	*notify_ptr;
	uint32_t		byte_len;	/*initialized from wc.byte_len*/

	/* initialize force_exit to 0.  This will be set to 1 if the completion
	 * handling loop needs to exit for some reason. */
	force_exit = 0;

	/* once in the critical section, do not allow
	 * this thread to be cancelled
	 */
	err = pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
	if (err) {
		OUR_PERROR("pthread_setcancelstate(PTHREAD_CANCEL_DISABLE): %s",
				err);
	}

	PRINT_STATUS("our_completion_thread ibv_poll_cq", wc->status);

	/* get transaction block for this completion */
	block = (void *)(unsigned long)wc->wr_id;

#if defined(ALLOW_PARANOID_CHECKING)
	if(block == NULL)
	{	/* should never happen */
		TRACE_ERR("%s wc->wr_id = NULL", message);
		force_exit = 1;
		goto early_exit;
	}
#endif
	TRACE("%s %s wc->wr_id is block %p, event %p\n",
			_time_prefix(), message, block, block->user_event_ptr);

	if(wc->status != IBV_WC_SUCCESS) {
		/* completion is reporting some error */
		if(wc->status != IBV_WC_WR_FLUSH_ERR) {
			TRACE_ERR("%s ibv_poll_cq returned: %s",
					message, ibv_wc_status_str(wc->status));
			TRACE_ERR("expected completion type %d",
					block->expected_comp);
			TRACE_ERR("send header type %"PRIu32,
					block->recv_header.hd_type);
			/* now force this to be a flush err to end gracefully */
			wc->status = IBV_WC_WR_FLUSH_ERR;
		} else {
			TRACE("%s %s fd %d IBV_WC_WR_FLUSH_ERR\n",
				_time_prefix(), message, cb->sockno);
		}

		cb->conn_flags &= ~(EXS_OK_TO_ISSUE_IO);
		if (atomic_swap_if_match(DISCONNECT_FLUSH_ERR,
					DISCONNECT_NONE,
					&cb->disconnect_flag)
				== DISCONNECT_NONE) {
			rdma_disconnect(cb->cm_id);
		}
		byte_len = 0;			/* EOF */
	} else {
		byte_len = wc->byte_len;	/* success */
	}

	/* wc->status is now either SUCCESS or FLUSH_ERR */

	TRACE("%s %s got ibv_poll_cq opcode %d, expected opcode %d\n",
					_time_prefix(), message, wc->opcode,
					block->expected_comp);
	if (block->expected_comp == IBV_WC_RECV
				&& wc->status == IBV_WC_SUCCESS
				&& wc->opcode == IBV_WC_RECV_RDMA_WITH_IMM) {
		/* No way to sort these; whatever gets it gets it. */
		block->expected_comp = IBV_WC_RECV_RDMA_WITH_IMM;
	}

	do_eof_broadcast = 0;
	qhandle = NULL;
	signo = -1;

	STATISTICS_INCREMENT(good_completions);

	switch (block->expected_comp)
	{
	case IBV_WC_SEND:	/* 0 */
		/* a previously posted send just completed */
		TRACE("\twc->opcode = IBV_WC_SEND\n");
		TRACE("\twc->byte_len = %d\n", byte_len);
		TRACE("\tsend_header.hd_seqno = %d\n",
					ntohl(block->send_header.hd_seqno));
		TRACE("\tsend_header.hd_type = %d\n",
					ntohl(block->send_header.hd_type));
		TRACE("\trecv_header.hd_seqno = %d\n",
					block->recv_header.hd_seqno);
		rtype = block->recv_header.hd_type & HEADER_TYPE_MASK;
		TRACE("\trecv_header.hd_type = %d\n", rtype);

		switch(rtype)
		{
		case OUR_ADVERTISEMENT:
			/* this send was advert.  It is normally a no-op since
			 * this should only happen if we need to flush the
			 * completion queue.  However, if we tried to send an
			 * ACK earlier and couldn't, then we need to do that
			 * now. */
			STATISTICS_INCREMENT(send_advert_completions);
			atomic_clear_bits(ADVERT_SENDING_BIT,
					&block->send_state);
			if (block->send_state & ACK_PENDING_BIT) {
				OUR_MEMBER_MUTEX_LOCK(cb, stream_sync_mutex);
				send_flags = IBV_SEND_SIGNALED;
				our_send_ack(cb, block, &send_flags, message);
				OUR_MEMBER_MUTEX_UNLOCK(cb, stream_sync_mutex);
			}
			break;
		case OUR_WRITE_NOTIFY:
			/* this is an iWARP WRITE notification */
			TRACE("\twriteack_header.hd_seqno = %d\n",
					ntohl(block->writeack_header.hd_seqno));

			cb->handle_send_write(cb, block, message,
							&qhandle, &signo);

			break;
		case OUR_SETUP_REQUEST:
			/* should only happen on client */
#if defined(ALLOW_PARANOID_CHECKING)
			if(cb->conn_state!=EXS_CONN_CLIENT_SETUP
					|| cb->nego_out_ptr == NULL) {
				/* should never happen */
				TRACE_ERR("block %p has type %d, but state is %d, expected %d, nego_in_ptr is %p, nego_out_ptr is %p",
						block, rtype, cb->conn_state,
						EXS_CONN_CLIENT_SETUP,
						cb->nego_in_ptr,
						cb->nego_out_ptr);
				force_exit = 1;
				goto early_exit;
			}
#endif
			STATISTICS_INCREMENT(total_setup_completions);
			if ((err = our_client_finishing(cb, SEND_DONE_BIT,
						&qhandle, &signo)) != 0) {
				TRACE_ERR("negotiation failed");
				force_exit = 1;
				goto around_again;
			}
			break;
		case OUR_SETUP_REPLY:
			/* should only happen on agent */
#if defined(ALLOW_PARANOID_CHECKING)
			if(cb->conn_state!=EXS_CONN_AGENT_SETUP
						|| cb->nego_out_ptr == NULL) {
				/* should never happen */
				TRACE_ERR("block %p has type %d, but state is %d, expected %d, nego_in_ptr is %p, nego_out_ptr is %p",
						block, rtype, cb->conn_state,
						EXS_CONN_AGENT_SETUP,
						cb->nego_in_ptr,
						cb->nego_out_ptr);
				force_exit = 1;
				goto early_exit;
			}
#endif
			STATISTICS_INCREMENT(total_setup_completions);
			if ((err = our_pending_accept_done(cb)) != 0) {
				TRACE_ERR("negotiation failure!");
				force_exit = 1;
			}
			goto around_again;

		case OUR_ACK:
			if (!(block->buffer_flags & MSG_WAITALL)
				|| block->recv_byte_len >= block->user_event_ptr
					->exs_evt_union.exs_evt_xfer.exs_evt_length) {
				/* we are finally done with the block for this
				 * recv transaction */
				signo = our_exs_recv_done(cb, block, &qhandle, 0);
			}
			break;

		case OUR_EOF:	/* should never happen */
			break;

#if defined(ALLOW_PARANOID_CHECKING)
		default:
			/* should never happen */
			TRACE_ERR("block %p has illegal type %d", block, rtype);
			force_exit = 1;
			goto early_exit;
#endif
		}	/* switch(rtype) */
		break;

	case IBV_WC_RECV:	/* 128 */
		TRACE("\twc->opcode = IBV_WC_RECV(%d)\n", IBV_WC_RECV);
		TRACE("\twc->byte_len = %d\n", byte_len);
		/* send_header gets newly recvd block's hd_seqno
		 *	in network byte order
		 * recv_header gets newly recvd block's hd_seqno
		 *	in host byte order
		 */
		block->send_header.hd_seqno = block->recv_header.hd_seqno;
		block->recv_header.hd_seqno =
					ntohl(block->recv_header.hd_seqno);
		TRACE("\trecv_header.hd_seqno = %d\n",
						block->recv_header.hd_seqno);
		block->send_header.hd_type = block->recv_header.hd_type;
		if (wc->status == IBV_WC_SUCCESS) {
			block->recv_header.hd_type = ntohl(block->recv_header.hd_type);
			rtype = (enum our_message_header_types)
				(block->recv_header.hd_type & HEADER_TYPE_MASK);
			TRACE("\trecv_header.hd_type = %d\n", rtype);
			TRACE("\trecv_header.hd_type flags = 0x%08x\n",
				block->recv_header.hd_type ^ rtype);
			TRACE("\tsend_header.hd_seqno = %d\n",
				ntohl(block->send_header.hd_seqno));
			TRACE("\tsend_header.hd_type = %d\n",
				ntohl(block->send_header.hd_type));
		} else {
			/* synchronize with closing threads, don't post another
			 * recv block */
			block->recv_header.hd_type = rtype = OUR_EOF;

			/*wake up thread waiting to close down*/
			do_eof_broadcast = 1;
		}

		/* Do pre-processing of received data.  We must copy data out of
		 * this block *before* it gets reposted. */
		switch(rtype)
		{
		case OUR_ADVERTISEMENT:
#if defined(ALLOW_PARANOID_CHECKING)
			if(byte_len != sizeof(block->send_header)
					+ sizeof(block->advert_buf)) {
				/* should never happen */
				TRACE_ERR("got advert wc->byte_len %d expected %d",
					byte_len,
					(int)(sizeof(block->send_header)
						+ sizeof(block->advert_buf)));
			}
#endif
			cb->handle_recv_advert(cb, block, wc, errbuf,
						message, &do_eof_broadcast,
						&force_exit, &qhandle, &signo);
			goto after_common;
		case OUR_SETUP_REQUEST:
			/* should only happen on agent */
#if defined(ALLOW_PARANOID_CHECKING)
			if(cb->conn_state!=EXS_CONN_AGENT_SETUP
					|| cb->nego_out_ptr == NULL) {
				/* should never happen */
				TRACE_ERR("block %p has "
					"type %d, but state is "
					"%d, expected %d, "
					"nego_in_ptr is %p, "
					"nego_out_ptr is %p",
					block, rtype,
					cb->conn_state,
					EXS_CONN_AGENT_SETUP,
					cb->nego_in_ptr,
					cb->nego_out_ptr);
				force_exit = 1;
				goto early_exit;
			}
#endif
			STATISTICS_INCREMENT(total_setup_completions);
			cb->nego_in_ptr = block;
			if ((err = our_agent_finishing(cb)) != 0) {
				TRACE_ERR("negotiation failure");
				force_exit = 1;
				goto around_again;
			}
			break;
		case OUR_SETUP_REPLY:
			/* should only happen on client */
#if defined(ALLOW_PARANOID_CHECKING)
			if(cb->conn_state!=EXS_CONN_CLIENT_SETUP
				|| cb->nego_out_ptr == NULL) {
				/* should never happen */
				TRACE_ERR("block %p has type %d, but state is %d, expected %d, nego_in_ptr is %p, nego_out_ptr is %p",
					block, rtype,
					cb->conn_state,
					EXS_CONN_CLIENT_SETUP,
					cb->nego_in_ptr,
					cb->nego_out_ptr);
				force_exit = 1;
				goto early_exit;
			}
#endif
			STATISTICS_INCREMENT(total_setup_completions);
			cb->nego_in_ptr = block;
			if ((err = our_client_finishing(cb, ACK_DONE_BIT,
						&qhandle, &signo)) != 0) {
				TRACE_ERR("negotiation failed");
				force_exit = 1;
				goto around_again;
			}
			break;

		case OUR_WRITE_NOTIFY:
			if (byte_len != sizeof(block->send_header)
					+ sizeof(block->writeack_buf)) {
				/* should never happen */
				TRACE_ERR("got WRITE notify wc->byte_len %" PRIu32 " expected %zu",
					byte_len, sizeof(block->send_header)
						+ sizeof(block->writeack_buf));
			}
			notify_ptr = (struct our_notify *)block->recv_buf;
			TRACE("%s %s Got WRITE_NOTIFY index=%" PRIu32 "\n",
					_time_prefix(), __func__,
					ntohl(notify_ptr->write_index));
			cb->handle_recv_write_notify(cb,
					ntohl(notify_ptr->write_len),
					ntohl(notify_ptr->write_index) & IMM_INDEX_MASK,
					ntohl(notify_ptr->write_flags) & IMM_FLAG_MASK,
					&qhandle, &signo);
			break;

		case OUR_ACK:
#if defined(ALLOW_PARANOID_CHECKING)
			if (byte_len != sizeof(block->send_header)
					+ sizeof(block->ack_buf)) {
				/* should never happen */
				TRACE_ERR("got ack wc->byte_len %d expected %zu",
					byte_len,
					sizeof(block->send_header)
						+ sizeof(block->ack_buf));
			}
#endif
			STATISTICS_INCREMENT(recv_ack_completions);

			our_handle_recv_ack(cb, block->recv_header.hd_seqno,
					(struct our_stream_ack *)block->recv_buf,
					block, wc, errbuf,
					message, &do_eof_broadcast,
					&force_exit, &qhandle, &signo);
			goto after_common;

		case OUR_EOF:
			break;

#if defined(ALLOW_PARANOID_CHECKING)
		default:
			/* should never happen */
			TRACE_ERR("block %p has illegal type %d", block, rtype);
			force_exit = 1;
			goto early_exit;
#endif
		}

		ret = our_repost_recv_block(cb, block, wc, errbuf, message,
							&do_eof_broadcast);

		/* we have not processed any of this data yet */
		block->buf_offset = 0;

		/* Do post-processing of the received data.  Stuff that can be
		 * done after we copy the data out. */
		switch(rtype)
		{
		case OUR_SETUP_REQUEST:
			/* TO-DO -- deal with result */
			goto after_common;

		case OUR_SETUP_REPLY:
			/* TO-DO -- deal with result */
			goto after_common;

#if defined(ALLOW_PARANOID_CHECKING)
		default:
			/* should never happen */
			TRACE_ERR("block %p has illegal type %d", block, rtype);
			force_exit = 1;
			goto early_exit;
#endif


		case OUR_ADVERTISEMENT:
		case OUR_WRITE_NOTIFY:
		case OUR_ACK:
			break;

		case OUR_EOF:
			/* for EOF, byte_len already set to 0 */
			STATISTICS_INCREMENT(recv_eof_completions);
			break;
		}	/* switch(rtype) */

after_common:
		break;

	case IBV_WC_RECV_RDMA_WITH_IMM:
		STATISTICS_INCREMENT(recv_rdma_with_imm_completions);

		/* imm_data is in network byte order
		 * all other work completion fields are in host byte order */
		TRACE("\twc->opcode = IBV_WC_RECV_RDMA_WITH_IMM(%d)\n",
						IBV_WC_RECV_RDMA_WITH_IMM);
		TRACE("\twc->byte_len = %" PRIu32 "\n", byte_len);
		TRACE("\twc->imm_data = %" PRIu32 "\n",
							ntohl(wc->imm_data));
		TRACE("\twc->wc_flags = %" PRIx32 "\n", wc->wc_flags);

		if(wc->status == IBV_WC_WR_FLUSH_ERR) {
			/* synchronize with closing threads, don't post another
			 * recv block */
			block->recv_header.hd_type = rtype = OUR_EOF;

			/*wake up thread waiting to close down*/
			do_eof_broadcast = 1;

			/*don't do any more processing*/
			break;
		} else {
			cb->handle_recv_write_notify(cb,
					byte_len,
					ntohl(wc->imm_data) & IMM_INDEX_MASK,
					ntohl(wc->imm_data) & IMM_FLAG_MASK,
					&qhandle, &signo);

			ret = our_repost_recv_block(cb, block, wc, errbuf, message,
							&do_eof_broadcast);
		}

		break;

	case IBV_WC_RDMA_WRITE:		/* 1 */
		STATISTICS_INCREMENT(send_rdma_write_completions);
		TRACE("\twc->opcode = "
				"IBV_WC_RDMA_WRITE(%d)\n",
				IBV_WC_RDMA_WRITE);
		cb->handle_send_write(cb, block, message, &qhandle, &signo);
		break;

#if defined(ALLOW_PARANOID_CHECKING)
	case IBV_WC_RDMA_READ:		/* 2 */
		/* should never happen
		 * EXS currently never does rdma_read()
		 */
		TRACE_ERR("Got unexpected wc->opcode = IBV_WC_RDMA_READ");
		break;
#endif

#if defined(ALLOW_PARANOID_CHECKING)
	default:
		/* should never happen */
		TRACE_ERR("Got unknown wc->opcode = %d", wc->opcode);
		force_exit = 1;
		goto early_exit;
#endif

	}	/* switch expected_comp */

around_again:
	/* signo  > 0 means send & event and actually raise a
	 *		UNIX signal
	 * signo == 0 means send an event but raise no signal
	 * signo  < 0 means don't send an event or a signal
	 * qhandle == NULL means user doesn't want a signal
	 *		or event
	 */
	if (signo >= 0 && qhandle != NULL) {
		_evt_queue_do_notify(qhandle, signo);
	}

#if defined(ALLOW_PARANOID_CHECKING)
early_exit:
#endif
	/* end of critical section, allow cancellation again */
	err=pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	if (err) {
		OUR_PERROR("pthread_setcancelstate(PTHREAD_CANCEL_ENABLE)",
				err);
	}

	return force_exit;
}

/* thread to catch completion events on connection "arg"
 *
 * while waiting or polling for the next completion event on this connection,
 *	no mutexes are held
 *
 * this thread initially can be cancelled by our_cancel_completion_thread(),
 * which will first lock event_mutex to ensure no race with this thread

 *	The arg passed in at creation is a pointer to the cb of
 *	either a client connection or an agent connection.
 *	It is never the cb of a listener connection.
 *
 */
static void *
our_completion_thread(void *arg)
{
	int			ret, force_exit;
	int			our_disconnect_state;
	int			got_flush_err;
	struct exs_conn		*cb = arg;
	struct ibv_cq		*event_cq;
	struct ibv_wc		*volatile wc;
	struct ibv_wc		*wc_limit;
	void			*event_ctx;
	char			message[100];
	char			errbuf[_MAX_STRERROR_MALLOC];

	sprintf(message, "[%lu] %s for cb %p",
		pthread_self(), __func__, cb);

	TRACE("%s %s Enter, cb->fd_flags %04x\n",
		_time_prefix(), message, cb->fd_flags);

	/* don't let the completion thread start to reap completions
	 * until the connection manager thread has completely finished
	 * establishing this connection -- otherwise we get a race
	 * between these 2 threads
	 */
	pthread_cleanup_push(our_completion_cancel_handler,
					&cb->event_mutex);
	TRACE("\t\t%4d:     LOCK event_mutex %p\n", __LINE__,
			&cb->event_mutex);
	pthread_mutex_lock(&cb->event_mutex);
	TRACE("\t\t%4d:     PAST event_mutex %p\n", __LINE__,
			&cb->event_mutex);

	while( !(cb->conn_flags & EXS_NEED_DISCONNECT)) {
		pthread_cond_wait(&cb->transmit_ok_notify, &cb->event_mutex);
	}

	/* this thread is really up */
	cb->conn_flags |= EXS_NEED_COMP_CANCEL;

	TRACE("\t\t%4d:   UNLOCK event_mutex %p\n", __LINE__,
			&cb->event_mutex);
	pthread_mutex_unlock(&cb->event_mutex);
	pthread_cleanup_pop(0);		/* i.e., do NOT do the unlock */

	pthread_detach(pthread_self());

	TRACE("%s %s Active, cb->fd_flags %04x\n",
		_time_prefix(), message, cb->fd_flags);

	for(force_exit = 0; force_exit == 0; )
	{
		STATISTICS_INCREMENT(completion_loops);

		/* loop to wait/poll for at least 1 new work completion */
		for( ; ; )
		{
			/* don't do this loop again if already canceled */
			pthread_testcancel();

#if 0
			if(!(cb->conn_flags & EXS_NEED_DISCONNECT))
			{	/* rdma_disconnect has already been done,
				 * there should no longer be a source to
				 * generate cq_events, so exit now!
				 */
				TRACE("%s %s self exiting, "
					"cb %p, state %s(%d), conn_flags %0x,"
					" needs_joining %u\n",
					_time_prefix(), message, cb,
					state_labels[cb->conn_state],
					cb->conn_state, cb->conn_flags,
					cb->comp_thread_needs_joining);
				cb->conn_flags &= ~(EXS_NEED_COMP_CANCEL);
				goto out0;
			}
#endif

			/* wait for cq event */
			if (!(cb->fd_flags & EXS_FD_BUSYPOLL)) {
				if ((ret = ibv_get_cq_event(cb->comp_channel,
						&event_cq, &event_ctx))) {
					if (strerror_r(ret, errbuf,
							_MAX_STRERROR_MALLOC)) {
						snprintf(errbuf,
							_MAX_STRERROR_MALLOC,
							"unknown");
					}
					TRACE_ERR("%s: ibv_get_cq_event(%p) on cb %p failed, error %d: %s",
							message,
							cb->comp_channel, cb,
							ret, errbuf);
					goto out0;
				}
				STATISTICS_INCREMENT(good_ibv_get_cq_events);

				if(event_cq != cb->cq)
				{
					TRACE_ERR("%s got event_cq = %p on cb %p, but expected cb->cq = %p",
							message,
							event_cq, cb, cb->cq);
					goto out0;
				}
				TRACE("%s %s got event_cq = cb->cq = %p on channel %p\n",
						_time_prefix(), message, cb->cq,
						cb->comp_channel);

				/* accumulate number of cq events that need to
				 * be acked, and periodically ack them
				 */
				cb->n_cq_events_that_need_ack++;
				if (cb->n_cq_events_that_need_ack == UINT_MAX) {
					ibv_ack_cq_events(cb->cq, UINT_MAX);
					cb->n_cq_events_that_need_ack = 0;
				}
				/* immediately set up completion queue
				 * notificaton again so that when next entry of
				 * any type is added to the completion queue, an
				 * event will be triggered on the completion
				 * channel
				 */
				ret = ibv_req_notify_cq(cb->cq, 0);
				if(ret)
				{
					if (strerror_r(ret, errbuf,
							_MAX_STRERROR_MALLOC)) {
						snprintf(errbuf,
							_MAX_STRERROR_MALLOC,
							"unknown");
					}
					TRACE_ERR("%s failed ibv_req_notify_cq on cb %p, error %d: %s",
						message, cb,
						ret, errbuf);
					goto out0;
				}
				TRACE("%s %s set up ibv cq notify for cb->cq %p on channel %p\n",
					_time_prefix(), message,
					cb->cq, cb->comp_channel);
			}


			/* pull all work completions off completion queue */
			if((ret = ibv_poll_cq(cb->cq, cb->max_n_wc_slots,
						cb->wc_array)) > 0) {
				/* got at least 1 new work completion */
				break;
			} else if (ret < 0) {
				/* got an error */
				ret = -ret;
				if(strerror_r(ret,errbuf,_MAX_STRERROR_MALLOC)){
					snprintf(errbuf, _MAX_STRERROR_MALLOC,
						"unknown");
				}
				TRACE_ERR("%s failed ibv_poll_cq on cb %p, error %d: %s",
					message, cb,
					ret, errbuf);
				goto out0;
			}
			/* else bogus cq_event, ignore it and loop again */
			STATISTICS_INCREMENT(empty_ibv_poll_cqs);
		}	/* for ever */

		/* when get out of loop, ret is positive ibv_poll_cq() result */
		STATISTICS_INCREMENT(good_ibv_poll_cqs);

		/* now process all the work completions we just polled */
		wc_limit = &cb->wc_array[ret];

		/* once around this loop for each work completion just polled */
		got_flush_err = 0;
		for (wc = cb->wc_array; wc < wc_limit && force_exit == 0; wc++){
			force_exit = our_deal_with_completion(wc, cb, errbuf,
								message);
			if (wc->status == IBV_WC_WR_FLUSH_ERR) {
				got_flush_err = 1;
			}
		}	/* for each work completion */
		if ((cb->conn_flags & EXS_OK_TO_ISSUE_IO) == 0
				&& cb->pending_send_count == 0
				&& cb->pending_recv_count == 0
				&& cb->conn_state == EXS_CONN_RUNNING) {
			if (atomic_swap_if_match(3, 0, &cb->disconnect_flag) == 0) {
				rdma_disconnect(cb->cm_id);
			}
			got_flush_err = 1;
		} else if (force_exit) {
			if (atomic_swap_if_match(3, 0, &cb->disconnect_flag) == 0) {
				rdma_disconnect(cb->cm_id);
			}
		}
		if (got_flush_err) {
			break;
		}
	}	/* for ever */

	/* for ever loop exits only on a fatal error
	 * (should never happen unless we got a flush error)
	 */
	if (!got_flush_err) {
		TRACE_ERR("%s Leave on fatal error", message);
	} else {
		TRACE_DBG(EXS_LOG_UNSPECIFIED,
			"%s Leave on flush error", message);
	}

out0:
	OUR_MEMBER_MUTEX_LOCK(cb, event_mutex);
	our_disconnect_state = atomic_set_bits(0x2, &cb->disconnect_state);

	our_flush_stranded_send_blocks(cb, EPIPE);
	our_flush_all_stranded_recv_blocks(cb);

	if (our_disconnect_state == 0x7) {
		our_qhandle_t	qptr;
		int		signo;

		/* We're the last ones out */
		TRACE("%s %s doing final cleanup\n",
				_time_prefix(), message);
		our_free_qp(cb);
		our_undo_exs_socket(cb->sockno, cb, 5, message,
					&qptr, &signo);
		if(signo >= 0)
			_evt_queue_do_notify(qptr, signo);
	} else {
		OUR_MEMBER_MUTEX_UNLOCK(cb, event_mutex);
	}
	pthread_exit(NULL);
}	/* our_completion_thread */


/* create a new our_completion_thread
 * to collect ibv completion events on this new connection
 *
 * called by client from our_connect_resolving()
 *	when our_con_man thread gets RDMA_CM_EVENT_ROUTE_RESOLVED
 *
 * called by server from our_merge_pending_accept()
 *	when our_con_man thread gets RDMA_CM_EVENT_CONNECT_REQUEST
 *
 * this probably should be done later in the state machine
 *	by the client in our_connect_finishing()
 *	by the server in our_finish_pending_accept()
 * after they set EXS_NEED_DISCONNECT in cb->conn_flags
 *	(so our_completion_thread can start reaping completion events
 *	 immediately, without waiting for the transmit_ok_notify signal)
 *
 * When called, event_mutex is locked
 */
static int
our_create_completion_thread(struct exs_conn *cb, const char *message)
{
	int	ret;
	char	errbuf[_MAX_STRERROR_MALLOC];

	TRACE("%s %s for cb %p, signal_limit %d\n",
		_time_prefix(), __func__, cb, cb->signal_limit);

	ret = pthread_create(&cb->comp_thread, NULL, our_completion_thread, cb);
	if(ret)
	{
		if (strerror_r(ret, errbuf, _MAX_STRERROR_MALLOC)) {
			snprintf(errbuf, _MAX_STRERROR_MALLOC, "unknown");
		}
		TRACE_ERR("%s: pthread_create of our_completion_thread failed, error %d: %s",
			message, ret, errbuf);
		goto out;
	}
	TRACE("%s %s: pthread_create created new"
		" our_completion_thread with id %lu\n",
		_time_prefix(), message, cb->comp_thread);

	if (cb->comp_thread_cpu != INT_MAX) {
		if ((ret = _exs_set_thread_affinity(
				cb->comp_thread,
				cb->comp_thread_cpu, NULL)) < 0) {
			if (strerror_r(errno, errbuf, _MAX_STRERROR_MALLOC)) {
				snprintf(errbuf, _MAX_STRERROR_MALLOC, "unknown");
			}
			TRACE_ERR("%s: set completion thread affinity failed, error %d: %s",
					message, errno, errbuf);
		}
	}

	/* do not detach this new thread because the completion thread
	 * is ALWAYS cancelled, and whoever does the cancel will then
	 * join the completion thread to be sure it is terminated
	 */
out:
	return ret;
}	/* our_create_completion_thread */

/*
 * called from our_merge_pending_accept() and exs_socket()
 *
 * initializes mutexes, condition variables, lists needed by all sockets
 */
static int
our_init_completion_synch(struct exs_conn *this_conn, const char *message)
{
	int	ret;
	char	errbuf[_MAX_STRERROR_MALLOC];

	ret = pthread_mutex_init(&this_conn->event_mutex, NULL);
	if(ret)
	{
		if (strerror_r(ret, errbuf, _MAX_STRERROR_MALLOC)) {
			snprintf(errbuf, _MAX_STRERROR_MALLOC, "unknown");
		}
		TRACE_ERR("%s: pthread_mutex_init event_mutex failed, error %d: %s",
				message, ret, errbuf);
		goto err1;
	}

	ret = pthread_cond_init(&this_conn->transmit_ok_notify, NULL);
	if(ret)
	{
		if (strerror_r(ret, errbuf, _MAX_STRERROR_MALLOC)) {
			snprintf(errbuf, _MAX_STRERROR_MALLOC, "unknown");
		}
		TRACE_ERR("%s: pthread_cond_init transmit_ok_notify failed, error %d: %s",
			message, ret, errbuf);
		goto err5;
	}

	goto err;

err5:
	pthread_mutex_destroy(&this_conn->event_mutex);
err1:
err:
	return ret;
}	/* our_init_completion_synch */


/* destroy the stuff set up in our_init_completion_synch() above */
static void
our_destroy_completion_synch(struct exs_conn *this_conn)
{
	pthread_cond_destroy(&this_conn->transmit_ok_notify);
	pthread_mutex_destroy(&this_conn->event_mutex);
}	/* our_destroy_completion_synch */


static inline int
our_create_stream_recv_buf(struct exs_conn *cb)
{
	long page_size;

	errno = 0;
	page_size = sysconf(_SC_PAGE_SIZE);
	if (page_size < 0) {
		if (errno != 0 && errno != EINVAL) {
			/* Some other serious error happened */
			OUR_PERROR("Could not get system page size", errno);
			return -1;
		}
		/* Default the page size to a reasonable value */
		page_size = 4096;
	}

	cb->remote_buf_n_empty = cb->stream_buf_size;

	do {
		if ((errno = exs_buf_alloc(cb, &cb->recv_buf.buf,
					cb->stream_buf_size,
					"recv stream buffer")) == 0) {
			break;
		} else {
			cb->stream_buf_size /= 2;
		}
	} while (cb->stream_buf_size > (unsigned long)page_size);
	if (errno != 0) {
		OUR_PERROR("Could not allocate local stream recv buf",
				errno);
		return -1;
	}
	cb->recv_buf.n_empty_threshold
		= cb->stream_buf_size / 3;
	cb->n_unacked_wwi_threshold
		= 0;
	TRACE("%s %s local recv buf is %p rkey %#.8" PRIx32 " len %zu\n",
			_time_prefix(), __func__,
			cb->recv_buf.buf.ptr,
			cb->recv_buf.buf.mr->rkey,
			cb->recv_buf.buf.size);
	return 0;
}	/* our_create_stream_recv_buf */


/*
 * Called from either our_merge_pending_accept (in a con_man context)
 *		   or exs_accept (in a user context)
 *
 * When called, listen_conn's event_mutex MUST be locked
 *		(prevents accepted_conn from processing completions until
 *		 we are finished setting everything up here)
 *		and listen_conn->pending_accept_list MUST NOT be empty
 *
 * Returns 0 if all ok, -1 on error.
 */
static int
our_continue_merge_pending_accept(struct exs_conn *listen_conn,
			 const struct rdma_cm_event *cm_event,
			 const struct our_nego_mess *nego_recv)
{
	int				ret = 0;
	struct pending_accept		*new_pending;
	struct rdma_conn_param		conn_param;
	struct exs_conn			*accepted_conn;
	uint32_t			value;

	/* remove first item from pending_accept_list */
	new_pending = exs_p_list_del_head(&listen_conn->pending_accept_list);
	if (new_pending == NULL) {
		errno = EINVAL;
		return -1;
	}

	TRACE("%s %s: Got new_pending %p\n",
		_time_prefix(), __func__, new_pending);
	accepted_conn = new_pending->accepted_conn;

	TRACE("%s %s: About to fill in pending accept, "
		"accepted_conn %p\n", _time_prefix(), __func__, accepted_conn);

	/* accept_parent back pointer used by our_completion_thread()
	 * to prevent processing completions until setup is finished here
	 */
	accepted_conn->accept_parent = listen_conn;

	/* client's version matches ours, negotiate parameter values
	 * by keeping the minimum of client's and listener's values
	 */
	TRACE("%s %s: %21s %u.%u.%u\n",
		_time_prefix(), __func__, "client's exs version",
		nego_recv->version, nego_recv->subversion,
		nego_recv->subsubversion);

	/* agent's credits set to min(client's local, listener's remote)
	 */
	value = ntohl(nego_recv->client_credits);
	if (value > listen_conn->posted_credits) {
		value = listen_conn->posted_credits;
	}
	our_set_send_credit_limits(accepted_conn, value);
	TRACE("%s %s: %21s %u\n",
			_time_prefix(), __func__, "posted credits",
			accepted_conn->posted_credits);

	/* agent's sp_max_size set to min(client's, listener's) */
	value = ntohl(nego_recv->sp_max_size);
	if (value > listen_conn->small_packet_max_size) {
		value = listen_conn->small_packet_max_size;
	}
	accepted_conn->small_packet_max_size = value;
	TRACE("%s %s: %21s %u\n",
			_time_prefix(),__func__,"small packet max size",value);

	if (!accepted_conn->is_seqpacket
				&& !(accepted_conn->fd_flags & EXS_FD_NODELAY)) {
		/* grab remote recv buffer information */
		accepted_conn->remote_recv_buf.ad_buf_addr
			= ntohll(nego_recv->recv_buf_addr);
		accepted_conn->remote_recv_buf.ad_buf_len
			= ntohl(nego_recv->recv_buf_len);
		accepted_conn->remote_recv_buf.ad_buf_key
			= ntohl(nego_recv->recv_buf_rkey);
		accepted_conn->remote_recv_buf.ad_buf_index = 0;
		TRACE("%s %s remote_recv_buf is %#.16" PRIx64 " rkey %#.8" PRIx32 " len %" PRIu32 "\n",
				_time_prefix(), __func__,
				accepted_conn->remote_recv_buf.ad_buf_addr,
				accepted_conn->remote_recv_buf.ad_buf_key,
				accepted_conn->remote_recv_buf.ad_buf_len);
	}

	/* other accepted connection attributes are same as listener's */
	accepted_conn->signal_limit
				= listen_conn->signal_limit;
	accepted_conn->fd_flags = listen_conn->fd_flags;
	accepted_conn->stream_buf_size = listen_conn->stream_buf_size;
	accepted_conn->is_seqpacket = listen_conn->is_seqpacket;
	accepted_conn->xmit_if_match = listen_conn->xmit_if_match;
	accepted_conn->enqueue_send = listen_conn->enqueue_send;
	accepted_conn->handle_recv_advert = listen_conn->handle_recv_advert;
	accepted_conn->handle_send_write = listen_conn->handle_send_write;
	accepted_conn->handle_recv_write_notify
		= listen_conn->handle_recv_write_notify;

	/* initialize completion condition variables */
	ret = our_init_completion_synch(accepted_conn, __func__);
	if(ret)
	{
		new_pending->err_numb = ret;
		goto err1;
	}

	accepted_conn->event_channel = rdma_create_event_channel();
	if (accepted_conn->event_channel == NULL) {
		new_pending->err_numb = errno;
		OUR_PERROR("rdma_create_event_channel failed",
			new_pending->err_numb);
		goto err1_5;
	}
	/* save the user's fd in accepted_conn */
	accepted_conn->sockno = accepted_conn->event_channel->fd;

	/* redundant, but makes it easier to post exs_event_t events */
	accepted_conn->conn_pending->sockno = accepted_conn->sockno;

	/* agent is given a new cm_id by the system, so migrate it from the
	 * listener's event channel to the new event channel just created
	 */
	TRACE("%s %s: Merge cm_id %p (from con_man event) to "
		"accepted_conn %p as its cm_id\n",
		_time_prefix(), __func__, cm_event->id, accepted_conn);
	accepted_conn->cm_id = cm_event->id;
	if ((ret = rdma_migrate_id(cm_event->id, accepted_conn->event_channel)) != 0) {
		ret = errno;
		new_pending->err_numb = ret;
		OUR_PERROR("rdma_migrate_id failed", ret);
		/* need to unwind stuff we just did above */
		goto err2;
	}

	/* create our_con_man_thread to pick up agent's asynch cm events */
	ret = create_our_con_man_thread(accepted_conn, __func__);
	if(ret) {
		/* error message already given,
		 * but need to unwind stuff just done above
		 */
		goto err3;
	}
	ret = our_setup_qp(accepted_conn, accepted_conn->cm_id);
	if(ret)
	{
		new_pending->err_numb = ret;
		OUR_PERROR("our_setup_qp failed", ret);
		/* need to unwind stuff we just did above */
		goto err4;
	}
	TRACE("%s %s: Got queue pair %p for pending accept\n",
		_time_prefix(), __func__, accepted_conn->qp);

	TRACE("%s %s: Accepted on source port %u\n",
		_time_prefix(), __func__,
		ntohs(rdma_get_src_port(accepted_conn->cm_id)));
	TRACE("%s %s: Accepted destination port %u\n",
		_time_prefix(), __func__,
		ntohs(rdma_get_dst_port(accepted_conn->cm_id)));

	TRACE("%s %s: Dump of src_addr:\n", _time_prefix(), __func__);
	DUMP_BUF((void *)&accepted_conn->cm_id->route.addr.
		src_addr, sizeof(struct sockaddr_storage));

	TRACE("%s %s: Dump of dst_addr:\n", _time_prefix(), __func__);
	DUMP_BUF((void *)&accepted_conn->cm_id->route.addr.
		dst_addr, sizeof(struct sockaddr_storage));

	/* create a thread to collect ibv completion events on
	 * this newly accepted connection
	 *
	 * this probably should be done later in the state machine
	 * in our_finish_pending_accept()
	 * when EXS_NEED_DISCONNECT is set in conn_flags
	 */

	accepted_conn->conn_state = EXS_CONN_ACCEPTING;
	accepted_conn->event_expected = RDMA_CM_EVENT_ESTABLISHED;

	 /*** !!! listener's event_mutex must be locked !!! ***/
	if((ret = our_create_completion_thread(accepted_conn, __func__)) != 0) {
		/* need to unwind stuff we just did above */
		new_pending->err_numb = ret;
		goto err5;
	}

	if (!accepted_conn->is_seqpacket
				&& !(accepted_conn->fd_flags & EXS_FD_NODELAY)) {
		if (our_create_stream_recv_buf(accepted_conn) < 0) {
			new_pending->err_numb = ENOMEM;
			goto err6;
		}
	}

	our_setup_agent_nego_mess(accepted_conn, &new_pending->nego_send);

	/* create all the empty transaction blocks, post some of them */
	ret = our_create_and_post_blocks(accepted_conn, __func__);
	if(ret)
	{
		OUR_PERROR("failed create and post blocks", ret);
		new_pending->err_numb = ret;
		goto err7;
	}

	memset(&conn_param, 0, sizeof(conn_param));

	conn_param.private_data = &new_pending->nego_send;
	conn_param.private_data_len = sizeof(new_pending->nego_send);
	conn_param.responder_resources = RESPONDER_RESOURCES;
	conn_param.initiator_depth = INITIATOR_DEPTH;
	conn_param.retry_count = RETRY_COUNT;
	conn_param.rnr_retry_count = RNR_RETRY_COUNT;

	/* actually accept the next client connection
	 * the accepted_conn->cm_id was set with the newly created
	 * id taken from the listening event we waited for above
	 * to tell us about the connection request.
	 */
	TRACE("%s %s: Accepting the pending accept\n",
		_time_prefix(), __func__);

	ret = rdma_accept(accepted_conn->cm_id, &conn_param);
	if(ret)
	{
		ret = errno;
		new_pending->err_numb = ret;
		OUR_PERROR("rdma_accept failed", ret);
		/* need to unwind stuff we just did above */
		goto err6;
	}

	return 0;

err7:
	exs_buf_free(&accepted_conn->recv_buf.buf);
err6:
	our_cancel_completion_thread(accepted_conn, __func__);
err5:
	our_free_qp(accepted_conn);
err4:
	our_terminate_con_man_thread(accepted_conn, __func__);
err3:
	rdma_destroy_id(accepted_conn->cm_id);
err2:
	rdma_destroy_event_channel(accepted_conn->event_channel);
err1_5:
	/* destroy the stuff set up in our_init_completion_synch() above */
	our_destroy_completion_synch(accepted_conn);
err1:
	/* add item back to pending_accept_list */
	exs_p_list_add_tail(&listen_conn->pending_accept_list, new_pending);
	return -1;
}	/* our_continue_merge_pending_accept */


/* called only by exs_accept() for a listening connection
 * called only in a user thread context
 *
 * appends a new non-empty list of "struct pending_accept" items
 * to the existing (possibly empty) accept_list of listen_conn
 */
static inline void
append_list_to_pending_accept_list(struct our_list_head *new_list,
				struct exs_conn *listen_conn)
{
	exs_p_list_splice(&listen_conn->pending_accept_list, new_list);
}	/* append_list_to_pending_accept_list */


/* called by exs_accept() and our_free_all_pending_lists()
 * for a listening connection
 * called only in a user thread context
 *
 * frees up all "struct pending_accept" items on new_list
 */
static void
our_free_new_accept_list(struct our_list_head *new_list, const char *message)
{
	int			j;
	struct pending_accept	*new_pending;

	TRACE("\t\t Free all pending accepts on new_list %p\n",
		 new_list);

	j = 0;
	while ((new_pending = exs_p_list_del_head(new_list))) {
		/* free "struct pending_connect" in "struct pending_accept" */
		EXS_FREE(new_pending->accepted_conn->conn_pending);

		/* free "struct exs_conn" in "struct pending_accept" */
		our_unalloc_socket(new_pending->accepted_conn, message);

		/* free "struct pending_accept" itself */
		EXS_FREE(new_pending);
		j++;
	}
	exs_p_list_destroy(new_list);

	TRACE("%s %s: Freed %d pending accepts from new_list %p\n",
		_time_prefix(), message, j, new_list);
}	/* our_free_new_accept_list */


/*
 *	exs_accept - page 11 of Interconnect Software Consortium Issue 1.0
 *		asynchronously accept incoming connections on a socket
 *
 *	returns 0 if async op initiated ok,
 *		else -1 with errno set to error value
 *
 * If the EXS_BLOCK flag is set in flags, this function will block until the
 * operation completes (i.e., the operation will be synchronous) and the
 * supplied addrvec_cnt must be 1.  In this case, the qhandle argument is
 * ignored.  No other flags are supported.
 */
EXS_EXPORT int
exs_accept(int sockno, exs_acceptaddr_t *addrvec, int addrvec_cnt,
		int flags, exs_qhandle_t qhandle)
{
	int			i, ret;
	exs_qhandle_t		qn;
	our_qhandle_t		qptr;
	our_qhandle_t		local_qhandle;
	struct exs_conn		*listen_conn;
	struct pending_accept	*new_pending;
	struct our_list_head	new_list;
	exs_event_t		local_vector;
	struct conn_requested	*old_requested;
	char			errbuf[_MAX_STRERROR_MALLOC];

#if defined(ALLOW_PID_CHECKING)
	pid_t			current;

	current = getpid();
	TRACE("%s start %s, pid %d, qhandle %p\n",
		_time_prefix(), __func__, current, qhandle);

	if(global_owner != current)
#else
	if(global_owner == 0)
#endif
	{
		errno = EPERM;		/* Operation not permitted */
		return -1;
	}

	/* verify that the user's flags are supported */
	if (flags & ~(EXS_BLOCK)) {
		errno = EINVAL;
		return -1;
	}

	exs_p_list_init(&new_list, 0);

#if defined(ALLOW_FULL_TRACE)
	TRACE("%s %s: socket %d, addrvec %p, "
		"addrvec_cnt %d, flags 0x%x\n", _time_prefix(), __func__, sockno,
		addrvec, addrvec_cnt, flags);
#endif

	/* verify that this socket has been created */
	if((listen_conn = our_lookup_socket(sockno)) == NULL)
	{
		errno = EBADF;		/* Bad file number */
		return -1;
	}

	TRACE("%s %s: enter, socket %d, listen_conn %p, state %s(%d)\n",
		_time_prefix(), __func__, sockno, listen_conn,
		state_labels[listen_conn->conn_state], listen_conn->conn_state);

	/* verify that the caller is asking for at least 1 connection */
	if(addrvec_cnt <= 0)
	{
		errno = EINVAL;		/* Invalid argument */
		return -1;
	}

	/* see if user gave us an event queue or wants a blocking accept */
	if(flags & EXS_BLOCK)
	{	/* caller can ask for only 1 connection when blocking */
		if(addrvec_cnt != 1)
		{
			errno = EINVAL;		/* Invalid argument */
			return -1;
		}

		/* get qhandle from cache, dynamically creating one if needed */
		local_qhandle = our_get_from_queue_cache(listen_conn);
		if(local_qhandle == NULL)
		{	/* couldn't create it, errno already set */
			return -1;
		}
		qptr = local_qhandle;	/* use our local qhandle */
		qn = (exs_qhandle_t)qptr->qnumb;
	}
	else
		/* verify that caller is giving us a valid qhandle */
		if((qptr = _cnv_qnumb_to_ptr(qhandle)) == NULL)
	{	/* Invalid argument */
		return -1;
	}
	else
	{	/* user supplied his own event queue, use it, no local one */
		local_qhandle = NULL;
		qn = qhandle;
	}

	TRACE("%s %s: exs_accept on source port %u\n",
		_time_prefix(), __func__,
		ntohs(rdma_get_src_port(listen_conn->cm_id)));
	TRACE("%s %s: Listening destination port %u\n",
		_time_prefix(), __func__,
		ntohs(rdma_get_dst_port(listen_conn->cm_id)));

	/* verify that this socket is currently listening */
	if(listen_conn->conn_state != EXS_CONN_LISTENING)
	{
		if(local_qhandle != NULL)
		{	/* caller wanted a blocking (synchronous) connect,
			 * delete the locally created queue we were going to use
			 */
			our_put_to_queue_cache(listen_conn, local_qhandle);
		}
		errno = EINVAL;   /* Invalid argument */
		return -1;
	}

	/* we lock event_mutex.
	 * At first glance this seems excessive, but at this time
	 * nothing else should be happening on this listener socket,
	 * so there should be no contention for this lock --
	 * we lock it to explicitly prevent anything else from happening!
	 */

	/* need to lock event_mutex to protect listener's conn_state */
	TRACE("\t\t%4d:     LOCK cb %p event_mutex %p\n",
		__LINE__, listen_conn, &listen_conn->event_mutex);
	pthread_mutex_lock(&listen_conn->event_mutex);
	TRACE("\t\t%4d:     PAST cb %p event_mutex %p\n",
		__LINE__, listen_conn, &listen_conn->event_mutex);

	/* check again that this socket is currently listening */
	if(listen_conn->conn_state != EXS_CONN_LISTENING)
	{
		errno = EINVAL;   /* Invalid argument */
		ret = -1;
	}
	else
	{	/* allocate memory for new accepted connections and events.
		 * A failure here is returned synchronously.
		 */
		ret = 0;
		for(i = 0; i < addrvec_cnt; i++)
		{
			TRACE("%s %s: Allocating memory for pending "
				"accept %d\n", _time_prefix(), __func__, i);

			new_pending = (struct pending_accept*)
					EXS_CALLOC(1, sizeof(*new_pending));
			if(new_pending == NULL)
			{
				/* can't come up with enough memory */
				if (strerror_r(ENOMEM, errbuf,
							_MAX_STRERROR_MALLOC)) {
					snprintf(errbuf, _MAX_STRERROR_MALLOC,
						"unknown");
				}
				TRACE_ERR("calloc of %zu bytes for new_pending %d on listener socket %d failed: %s",
						sizeof(struct pending_accept),
						i, sockno, errbuf);
				goto errmalloc1;
			}

			TRACE("%s %s: Pending new_pending[%d] = %p\n",
				_time_prefix(), __func__, i, new_pending);

			/* allocates a new struct exs_conn, and initializes its
			 * state field to EXS_CONN_UNOPENED
			 */
			new_pending->accepted_conn = (struct exs_conn*)
					our_alloc_socket(__func__);
			if(new_pending->accepted_conn == NULL)
			{
				/* can't come up with enough memory */
				goto errmalloc2;
			}

			/* set pending back pointer */
			new_pending->accepted_conn->accept_pending=new_pending;

			/* copy exs_fcntl settings */
			new_pending->accepted_conn->comp_thread_cpu =
						listen_conn->comp_thread_cpu;
			new_pending->accepted_conn->inline_limit =
						listen_conn->inline_limit;

			if(addrvec != NULL)
				new_pending->addrvec = &addrvec[i];
			new_pending->evt_item
				= _get_empty_evt_queue_event(qptr, qn);
			if(new_pending->evt_item == NULL)
			{
				/* free connection in pending_accept */
				goto errmalloc4;
			}
			new_pending->qhandle = qptr;
			exs_p_list_add_tail(&new_list, new_pending);
		}
		TRACE("%s %s: Done allocating memory for all %d "
			"pending accepts\n", _time_prefix(), __func__, i);
		append_list_to_pending_accept_list(&new_list, listen_conn);
		/* new_list is invalidated after the last call to
		 * append_list_to_pending_accept_list */

		while(!exs_p_list_empty(&listen_conn->pending_accept_list)
			&& listen_conn->conn_requested_backlog > 0
			&& !exs_p_list_empty(&listen_conn->conn_requested_list)) {
			/* a connect request from remote client already arrived
			 * and was waiting for these accepts, satisfy it now.
			 */
			old_requested = exs_p_list_del_head(
					&listen_conn->conn_requested_list);
			if(our_continue_merge_pending_accept(listen_conn,
						&old_requested->cm_event,
						&old_requested->nego_recv)) {
				/* merge failed */
				break;
			}
			/* merge worked ok, recyle old_requested */
			exs_p_list_add_tail(
					&listen_conn->conn_requested_free_list,
					old_requested);
			listen_conn->conn_requested_backlog--;
		}	/* while */
		goto carry_on;

		/* cleanup from failed accept initialization */

	errmalloc4:/* free struct pending_connect in incomplete pending_accept*/
		EXS_FREE(new_pending->accepted_conn->conn_pending);
		    /* free connection in incomplete pending_accept */
		our_unalloc_socket(new_pending->accepted_conn, __func__);

	errmalloc2: /* free incomplete pending_accept */
		EXS_FREE(new_pending);

	errmalloc1: /* free any completed pending_accept structs */
		our_free_new_accept_list(&new_list, __func__);
		exs_p_list_destroy(&new_list);

		/* FIXME: What failed? */
		OUR_PERROR("failed", ENOMEM);
		errno = ENOMEM;			/* Out of memory */
		ret = -1;
	}
carry_on:
	TRACE("\t\t%4d:   UNLOCK cb %p event_mutex %p\n",
		__LINE__, listen_conn, &listen_conn->event_mutex);
	pthread_mutex_unlock(&listen_conn->event_mutex);

	if(ret)
	{	/* always return any locally created queue to conn's cache */
		if(local_qhandle != NULL)
			our_put_to_queue_cache(listen_conn, local_qhandle);
		goto out_now;
	}

	/* synchronous part of accept started ok, rest happens asynchly */
	if(local_qhandle == NULL)
		goto out_now;

	/* caller wants a blocking (synchronous) accept */

	TRACE("%s %s: blocking accept, local_qhandle %p\n",
		_time_prefix(), __func__, local_qhandle);

	/* accept was started ok, now wait for it to complete */
	ret = _evt_queue_dequeue_vector(local_qhandle,
					&local_vector, 1, NULL);

	TRACE("%s %s: back from dequeue, local_qhandle %p, "
		"ret %d, exs_evt_type %d, exs_evt_errno %d\n",
		_time_prefix(), __func__, local_qhandle, ret,
		local_vector.exs_evt_type,
		local_vector.exs_evt_errno);

	if(ret != -1)
	{	/* no error on dequeue itself check completion */
		if((errno = local_vector.exs_evt_errno) != 0) {
			ret = -1;
		}
#if defined(ALLOW_PARANOID_CHECKING)
		else if(local_vector.exs_evt_type != EXS_EVT_ACCEPT)
		{	/* should never happen */
			TRACE_ERR("_evt_queue_dequeue_vector got item type %d, expected %d (EXS_EVT_ACCEPT)",
				local_vector.exs_evt_type, EXS_EVT_ACCEPT);
			errno = EBADSLT;	/* Invalid slot !!! */
			ret = -1;
		}
#endif
		else
		{	/* no error on accept, return fd of new connection */
			ret = local_vector.exs_evt_union.exs_evt_accept
				.exs_evt_new_socket;
		}
	}

	/* always return the locally created queue to the connection's cache */
	our_put_to_queue_cache(listen_conn, local_qhandle);

out_now:
	TRACE("%s %s: finished, listen_conn %p, state %s(%d), ret "
		"%d\n", _time_prefix(), __func__, listen_conn,
		state_labels[listen_conn->conn_state],
		listen_conn->conn_state, ret);
	return ret;

}	/* exs_accept */


/*
 *	exs_blocking_accept()
 *			- not in Interconnect Software Consortium Issue 1.0
 *			      was added by us
 *
 *	returns >= 0  if accepted ok
 *		== -1 if error (with errno set to error value)
 */
EXS_EXPORT int
exs_blocking_accept(int sockno, struct sockaddr *addr, socklen_t *addrlen)
{
	int ret;

	exs_acceptaddr_t addrvec;
	addrvec.exs_addr = addr;
	addrvec.exs_addrlen = *addrlen;
	addrvec.exs_ahandle = NULL;

	ret = exs_accept(sockno, &addrvec, 1, EXS_BLOCK, NULL);
	*addrlen = addrvec.exs_addrlen;
	return ret;
}	/* exs_blocking_accept */


/*
 * Called only in our_con_man_thread() context for a listening connection
 *
 * called directly from our_con_man_thread()
 * when an RDMA_CM_EVENT_CONNECT_REQUEST has been received on listen_conn
 *
 * When called, event_mutex MUST be locked.
 *
 * Returns 0 if all ok, -1 on error.
 */
static int
our_merge_pending_accept(struct exs_conn *listen_conn,
			 struct rdma_cm_event *cm_event)
{
	int				ret = 0;
	struct conn_requested		*new_requested;
	const struct our_nego_mess	*nego_recv;
	char				errbuf[_MAX_STRERROR_MALLOC];

	/* note: client may send more than the length the client actually sent,
	 *	 since the amount transfered is transport dependent
	 */
	if ((nego_recv = cm_event->param.conn.private_data) == NULL
		|| cm_event->param.conn.private_data_len < sizeof(*nego_recv)){
		/* something wrong with client's private data */
		if (nego_recv == NULL) {
			TRACE_ERR("client sent no private_data");
		} else {
			TRACE_ERR("client sent private_data_len %u, expected %zu",
				cm_event->param.conn.private_data_len,
				sizeof(*nego_recv));
		}
		/* Operation not supported on transport endpoint */
		ret = EOPNOTSUPP;
		return -1;
	}

	/* client sent private data, check the version number and socktype */
	if (nego_recv->version != EXS_VERSION
		|| nego_recv->subversion != EXS_SUBVERSION) {
		TRACE_ERR("client sent version %u.%u, expected %u.%u",
			nego_recv->version, nego_recv->subversion,
			EXS_VERSION, EXS_SUBVERSION);
		/* Operation not supported on transport endpoint */
		ret = EOPNOTSUPP;
		return -1;
	}

	/* check the socktype */
	if (nego_recv->is_seqpacket != listen_conn->is_seqpacket) {
		TRACE_ERR("client requested %s, server expected %s",
			nego_recv->is_seqpacket ? "SOCK_SEQPACKET" : "SOCK_STREAM",
			listen_conn->is_seqpacket ? "SOCK_SEQPACKET" : "SOCK_STREAM");
		/* Operation not supported on transport endpoint */
		ret = EOPNOTSUPP;
		return -1;
	}

	/* check the fd_flags */
	if (!nego_recv->is_seqpacket &&
			(nego_recv->fd_flags & EXS_FD_NODELAY)
			!= (listen_conn->fd_flags & EXS_FD_NODELAY)) {
		ret = EOPNOTSUPP;
		return -1;
	}

	if(!exs_p_list_empty(&listen_conn->pending_accept_list))
	{	/* a remote client wants to connect and this listener
		 * has at least 1 exs_accept() pending to match it
		 */
		return our_continue_merge_pending_accept(listen_conn, cm_event,
				nego_recv);
	}

	/* a remote client wants to connect, but this listener
	 * has no exs_accept() pending to match it
	 */
	TRACE("\t\tcb %p got event type CONNECT_REQUEST "
		"but pending_accept_list is empty\n", listen_conn);
	if(!(new_requested = exs_p_list_del_head(
				&listen_conn->conn_requested_free_list))) {
		/* no free structures, try to create a new one */
		new_requested = EXS_CALLOC(1, sizeof(*new_requested));
		if(new_requested == NULL)
		{	/* not enough memory for a new structure */
			if (strerror_r(ENOMEM, errbuf,
					_MAX_STRERROR_MALLOC)) {
				snprintf(errbuf, _MAX_STRERROR_MALLOC,
					"unknown");
			}
			TRACE_ERR("calloc of %zu bytes for struct conn_requested failed: %s",
					sizeof(*new_requested), errbuf);
			ret = ENOMEM;
			return -1;
		}
	}
	listen_conn->conn_requested_backlog++;
	if(listen_conn->conn_requested_backlog
					> listen_conn->backlog_limit)
	{
		TRACE_ERR("current backlog %d exceeds limit %d on listening socket %d",
			listen_conn->conn_requested_backlog,
			listen_conn->backlog_limit,
			listen_conn->sockno);
	}
	TRACE("%s %s: Copy cm_id %p to "
		"new_requested %p as its cm_id, current backlog %d\n",
		_time_prefix(), __func__, cm_event->id,
		new_requested, listen_conn->conn_requested_backlog);
	memcpy(&new_requested->cm_event,cm_event,sizeof(struct rdma_cm_event));
	memcpy(&new_requested->nego_recv,nego_recv,sizeof(*nego_recv));
	exs_p_list_add_tail(&listen_conn->conn_requested_list, new_requested);
	return 0;
}	/* our_merge_pending_accept */


/*
 * Called only in our_con_man_thread() context for an agent connection
 * called only from our_con_man_thread()
 *
 * when an RDMA_CM_EVENT_ESTABLISHED has been received on listen_conn
 *
 * When called, event_mutex MUST be locked.
 *
 * Returns 0 if all ok, -1 on error.
 */
static int
our_finish_pending_accept(struct exs_conn *accepted_conn)
{
	/* turn on all possible conn_flags */
	accepted_conn->conn_flags |= (EXS_NEED_LOCAL_CLOSE
				| EXS_NEED_DISCONNECT | EXS_OK_TO_ISSUE_IO);

	accepted_conn->conn_state = EXS_CONN_AGENT_SETUP;
	accepted_conn->event_expected = RDMA_CM_EVENT_DISCONNECTED;

	/* add accepted_conn to end of conn_list */
	our_add_socket_to_list(accepted_conn);
	return 0;
}	/* our_finish_pending_accept */

static void
our_uninit_transaction_block(struct transaction_block *block)
{
	EXS_FREE(block->user_event_ptr);
	our_mderegister(block->recv_mhandle);
	if(block->send_buf_mr)
	{
		ibv_dereg_mr(block->send_buf_mr);
		EXS_FREE(block->send_buf);
	}

	if(block->recv_rdma_buf)
	{
		ibv_dereg_mr(block->recv_rdma_mr);
		EXS_FREE(block->recv_rdma_buf);
	}
}	/* our_uninit_transaction_block */

/*
 * called from our_setup_qp(),
 *	twice from our_free_qp(), our_undo_exs_socket(), our_connect_done()
 *	our_pending_accept_done()
 *
 * destroy all the things created in our_create_transaction_block()
 */
static void
our_free_transaction_block(struct transaction_block *block)
{
	our_uninit_transaction_block(block);
	ibv_dereg_mr(block->block_mr);
	EXS_FREE(block);
}	/* our_free_transaction_block */


/*	destroy the queue pair (QP) on connection cb
 *
 *	"Users must destroy any QP associated with an rdma_cm_id
 *	 before destroying the ID."
 */
static void
our_destroy_qp(struct exs_conn *cb, const char *message)
{
	(void)message;
	TRACE("%s %s: call rdma_destroy_qp(%p) on cb %p\n",
		_time_prefix(), message, cb->cm_id, cb);
	rdma_destroy_qp(cb->cm_id);
}	 /* our_destroy_qp */


/*	destroy the completion queue (CQ) on connection cb
 *
 *	Problematic:
 *	"To avoid races, destroying a CQ will wait for all
 *	 completion events to be acknowledged; this guarantees a
 *	 one-to_one correspondence between acks and successful gets."
 */
static void
our_destroy_cq(struct exs_conn *cb, const char *message)
{
	int	ret;
	char	errbuf[_MAX_STRERROR_MALLOC];

	/* acknowledge all accumulated successful ibv_get_cq_event() calls */
	if (cb->n_cq_events_that_need_ack != 0) {
		ibv_ack_cq_events(cb->cq, cb->n_cq_events_that_need_ack);
		cb->n_cq_events_that_need_ack = 0;
	}

	TRACE("%s %s: call ibv_destroy_cq(%p) on cb %p\n",
		_time_prefix(), message, cb->cq, cb);
	if((ret = ibv_destroy_cq(cb->cq)) != 0)
	{	/* fails if any queue pair is still associated with this CQ */
		if (strerror_r(ret, errbuf, _MAX_STRERROR_MALLOC)) {
			snprintf(errbuf, _MAX_STRERROR_MALLOC, "unknown");
		}
		TRACE_ERR("%s: ibv_destroy_cq(%p) on cb %p failed, error %d: %s",
				message, cb->cq, cb, ret, errbuf);
	}
	if (cb->wc_array != NULL)
		EXS_FREE(cb->wc_array);
}	 /* our_destroy_cq */


/*	destroy the completion event channel on connection cb */
static void
our_destroy_comp_channel(struct exs_conn *cb, const char *message)
{
	int	ret;
	char	errbuf[_MAX_STRERROR_MALLOC];

	if (cb->fd_flags & EXS_FD_BUSYPOLL) {
		return;
	}

	TRACE("%s %s: call ibv_destroy_comp_channel(%p) on cb %p\n",
		_time_prefix(), message, cb->comp_channel, cb);
	if((ret = ibv_destroy_comp_channel(cb->comp_channel)) != 0)
	{	/* this fails if any CQs are still associated with the
		 * completion event channel being destroyed
		 */
		if (strerror_r(ret, errbuf, _MAX_STRERROR_MALLOC)) {
			snprintf(errbuf, _MAX_STRERROR_MALLOC, "unknown");
		}
		TRACE_ERR("%s: ibv_destroy_comp_channel(%p) on cb %p failed, error %d: %s",
			message, cb->comp_channel, cb, ret, errbuf);
	}
}	/* our_destroy_comp_channel */

/* Attempt to create a queue pair with the given inline limit.  Returns the
 * value of errno on error and 0 on success. */
int
our_try_rdma_create_qp(struct exs_conn *cb, struct ibv_qp_init_attr *init_attr,
		int inline_limit)
{
	int ret;

	init_attr->cap.max_inline_data =
		cb->requested_cap.max_inline_data = inline_limit;
	if((ret = rdma_create_qp(cb->cm_id, cb->pd, init_attr))) {
		ret = errno;
		TRACE("%s %s rdma_create_qp failed requesting "
			"max_inline_data %u, errno %d\n",
			_time_prefix(), __func__, inline_limit, ret);
	} else {
		ret = 0;
		TRACE("%s %s rdma_create_qp ok requesting "
			"max_inline_data %u\n",
			_time_prefix(), __func__, inline_limit);
	}

	return ret;
}

/*
 * Called only in our_con_man_thread() context
 * called only from our_setup_qp()
 *
 * to create a queue pair
 *
 * no locks needed for this function
 *
 * Returns 0 if all ok, errno on error.
 */
static int
our_create_qp(struct exs_conn *cb)
{
	struct ibv_qp_init_attr	init_attr;
	struct ibv_qp_attr	attr;
	int			ret, i;

	TRACE("%s start %s\n", _time_prefix(), __func__);

	/* initialize the queue pair capabilities we will request */
	cb->requested_cap.max_send_wr = SQ_DEPTH(cb);
	cb->requested_cap.max_recv_wr = RQ_DEPTH(cb);
	cb->requested_cap.max_recv_sge = 4;
	cb->requested_cap.max_send_sge = 4;
	cb->requested_cap.max_inline_data = inline_data_size[0];

	memset(&init_attr, 0, sizeof(init_attr));
	memcpy(&init_attr.cap, &cb->requested_cap, sizeof(struct ibv_qp_cap));

	init_attr.qp_type = IBV_QPT_RC;
	init_attr.send_cq = cb->cq;
	init_attr.recv_cq = cb->cq;

	ret = -1;

	/* allocate a queue pair associated with the cm_id for this connection
	 * and transition it for sending and receiving.  On return, it is ready
	 * to handle posting of sends and receives.
	 *
	 * Try the user-specified inline limit first */
	if(cb->inline_limit != INT_MAX && our_try_rdma_create_qp(cb, &init_attr,
						cb->inline_limit) == 0) {
		goto got_qp;
	}

	/* Now try some known inline limits */
	for (i = 0; i < KNOWN_INLINE_DATA_SIZE_LIMIT; ++i) {
		if (our_try_rdma_create_qp(cb, &init_attr,
						inline_data_size[i]) == 0) {
			goto got_qp;
		}
	}

	/* If we get here, then we failed to create a queue pair! */
	goto err;

got_qp:
	ret = 0;
	/* queue pair created ok, save the new cm_id and the
	 * capabilities we obtained */
	cb->qp = cb->cm_id->qp;
	ibv_query_qp(cb->qp, &attr, -1, &init_attr);
	memcpy(&cb->obtained_cap, &init_attr.cap,
					sizeof(struct ibv_qp_cap));
	if (attr.path_mtu > 0 && attr.path_mtu < (sizeof(path_size)
							/ sizeof(*path_size))) {
		cb->obtained_path_mtu_size = path_size[attr.path_mtu];
	}
	if (cb->inline_limit > cb->obtained_cap.max_inline_data) {
		cb->inline_limit = cb->obtained_cap.max_inline_data;
	}
	TRACE("%s %s obtained max_inline_data %u\n",
			_time_prefix(), __func__, cb->inline_limit);

	switch (cb->qp->context->device->transport_type) {
	case IBV_TRANSPORT_IB:
		TRACE("InfiniBand/RoCE interface supports RDMA_WRITE_WITH_IMM\n");
		break;
	case IBV_TRANSPORT_IWARP:
		TRACE("iWARP does not support RDMA_WRITE_WITH_IMM\n");
		cb->fd_flags |= EXS_FD_NO_WRITE_WITH_IMM;
		break;
	default:
		TRACE("Unknown transport type\n");
		cb->fd_flags |= EXS_FD_NO_WRITE_WITH_IMM;
		break;
	}	/* switch */

err:
	TRACE("%s %s: finished, returning %d\n",
		_time_prefix(), __func__, ret);

	return ret;
}	/* our_create_qp */


/*
 * Called only in our_con_man_thread() context
 * called from our_merge_pending_accept() and our_connect_resolving()
 *
 * to set up a complete queue-pair
 *
 * no locks needed for this function
 *
 * returns 0 if all ok, else the errno if things go bad
 */
static int
our_setup_qp(struct exs_conn *cb, struct rdma_cm_id *cm_id)
{
	int			ret;
	char			errbuf[_MAX_STRERROR_MALLOC];

	TRACE("%s start %s, cb %p, cm_id %p\n",
		_time_prefix(), __func__, cb, cm_id);

	/* get a global protection domain for this cm_id's device */
	if ((cb->pd = our_get_pd(cm_id->verbs)) == NULL) {
		ret = errno;
		goto err0;
	}

	/* create a completion channel for this cm_id */
	if (cb->fd_flags & EXS_FD_BUSYPOLL) {
		cb->comp_channel = NULL;
	} else {
		cb->comp_channel = ibv_create_comp_channel(cm_id->verbs);
		if(!cb->comp_channel)
		{
			ret = errno;
			if (strerror_r(ret, errbuf, _MAX_STRERROR_MALLOC)) {
				snprintf(errbuf, _MAX_STRERROR_MALLOC, "unknown");
			}
			TRACE_ERR("ibv_create_comp_channel(%p) on cb %p failed, error %d: %s",
				cm_id->verbs, cb, ret, errbuf);
			goto err1;
		}
		TRACE("%s \tcreated ibv completion channel %p, fd %d\n",
			_time_prefix(), cb->comp_channel, cb->comp_channel->fd);
	}

	/* create a completion queue */
	cb->cq = ibv_create_cq(cm_id->verbs, CQ_DEPTH(cb), cb,
				cb->comp_channel, 0);
	if(!cb->cq)
	{
		ret = errno;
		if (strerror_r(ret, errbuf, _MAX_STRERROR_MALLOC)) {
			snprintf(errbuf, _MAX_STRERROR_MALLOC, "unknown");
		}
		TRACE_ERR("ibv_create_cq(%p, %d, %p, %p) failed, error %d: %s",
			cm_id->verbs, CQ_DEPTH(cb), cb, cb->comp_channel,
			ret, errbuf);
		goto err2;
	}
	TRACE("%s \tcreated ibv cq %p, cqe %d\n",
		_time_prefix(), cb->cq, cb->cq->cqe);

	cb->requested_cq_nelements = CQ_DEPTH(cb);
	cb->obtained_cq_nelements = cb->cq->cqe;

	/* allocate an array to hold polled work completions that is
	 * big enough to never fill up, even if cq is full when polled
	 */
	cb->max_n_wc_slots = cb->cq->cqe + 1;
	if ((cb->wc_array = (struct ibv_wc*)EXS_CALLOC(cb->max_n_wc_slots,
					sizeof(struct ibv_wc))) == NULL) {
		/* not enough memory for the work completion array */
		ret = ENOMEM;
		if (strerror_r(ret, errbuf, _MAX_STRERROR_MALLOC)) {
			snprintf(errbuf, _MAX_STRERROR_MALLOC, "unknown");
		}
		TRACE_ERR("calloc of %zu bytes for wc_array failed, error %d: %s",
			sizeof(struct ibv_wc) * cb->max_n_wc_slots,
			ret, errbuf);
		goto err3;
	}

	TRACE("%s \tallocated wc_array %p, max_n_wc_slots %d\n",
		_time_prefix(), cb->wc_array, cb->max_n_wc_slots);

	/* create a queue pair */
	ret = our_create_qp(cb);
	if(ret)
	{
		OUR_PERROR("our_create_qp failed", ret);
		goto err3;
	}
	TRACE("%s \tcreated qp %p\n", _time_prefix(), cb->qp);

	if ((ret = pthread_mutex_init(&cb->advert_sequence_mutex, NULL)) != 0) {
		goto err4;
	}
	if ((ret = pthread_mutex_init(&cb->data_sequence_mutex, NULL)) != 0) {
		goto err5;
	}

	if (!cb->is_seqpacket) {
		if ((ret = pthread_mutex_init(&cb->stream_sync_mutex,
								NULL)) != 0) {
			goto err6;
		}
	}

	cb->signal_limit = SQ_DEPTH(cb) / 4;

	cb->pending_send_count = 0;		/* number currently posted */
	TRACE_VAR_ASSIGN(cb->max_posted_send_blocks, 0);
	cb->pending_recv_count = 0;		/* number currently posted */
	TRACE_VAR_ASSIGN(cb->max_posted_recv_blocks, 0);

	ret = our_nego_init(cb, &cb->nego_out_ptr);
	if(ret)
	{
		goto err9;
		/*goto err10;*/
	}

	/* to avoid races, prime the cq notification mechanism BEFORE
	 * we connect or accept the connection
	 */
	ret = ibv_req_notify_cq(cb->cq, 0);
	if (ret) {
		if(strerror_r(ret, errbuf, _MAX_STRERROR_MALLOC)){
			snprintf(errbuf, _MAX_STRERROR_MALLOC, "unknown");
		}
		TRACE_ERR("failed ibv_req_notify_cq on cb %p, error %d: %s",
			cb, ret, errbuf);
		goto err10;
	}
	TRACE("%s %s: finished ok\n", _time_prefix(), __func__);
	goto err0;

err10:
	cb->nego_in_ptr = NULL;
err9:
	if (!cb->is_seqpacket) {
		pthread_mutex_destroy(&cb->stream_sync_mutex);
	}
err6:
	pthread_mutex_destroy(&cb->data_sequence_mutex);
err5:
	pthread_mutex_destroy(&cb->advert_sequence_mutex);
err4:
	our_destroy_qp(cb, __func__);
err3:
	our_destroy_cq(cb, __func__);
err2:
	our_destroy_comp_channel(cb, __func__);
err1:
	/* FIXME: What failed? */
	OUR_PERROR("failed", ret);
err0:
	return ret;
}	/* our_setup_qp */


/*
 * called from our_merge_pending_accept(), our_connect_returning()
 *	our_close_accepting(), our_close_connecting, our_start_close()
 *
 * to destroy all the things created in our_setup_qp()
 *
 * when called, our_completion_thread MUST be canceled or this must be called
 * from our_completion_thread AFTER the event thread has exited and exs_close()
 * has been called
 *	(so it doesn't try to mess with cb's blocks and block lists)
 */
static void
our_free_qp(struct exs_conn *cb)
{
	struct ibv_wc		wc;
	int			ncompletes;
	unsigned int		i;

	TRACE("%s start %s, cb %p, cm_id %p\n",
		_time_prefix(), __func__, cb, cb->cm_id);

	/* first recycle all the blocks we created */
	if (cb->send_wwi_array_mr) {
		for (i = 0; i < cb->posted_credits; ++i) {
			our_uninit_transaction_block(&cb->send_wwi_array[i]);
			sem_destroy(&cb->send_wwi_array[i].sem);
		}
		ibv_dereg_mr(cb->send_wwi_array_mr);
		EXS_FREE(cb->send_wwi_array);
	}
	if (cb->send_advert_array_mr) {
		for (i = 0; i < cb->posted_credits; ++i) {
			our_uninit_transaction_block(&cb->send_advert_array[i]);
			sem_destroy(&cb->send_advert_array[i].sem);
		}
		ibv_dereg_mr(cb->send_advert_array_mr);
		EXS_FREE(cb->send_advert_array);
	}

	if (cb->recv_array_mr) {
		for (i = 0; i < RQ_DEPTH(cb); ++i) {
			our_uninit_transaction_block(&cb->recv_array[i]);
		}
		ibv_dereg_mr(cb->recv_array_mr);
		EXS_FREE(cb->recv_array);
	}

	if (cb->recved_advert_array) {
		EXS_FREE(cb->recved_advert_array);
	}

	if (!cb->is_seqpacket && !(cb->fd_flags & EXS_FD_NODELAY)) {
		exs_buf_free(&cb->recv_buf.buf);
	}

	if(cb->nego_out_ptr != NULL)
	{	/* need to free nego block */
		TRACE("%s %s: free nego out transaction block %p\n",
			_time_prefix(), __func__, cb->nego_out_ptr);
		our_free_transaction_block(cb->nego_out_ptr);
		cb->nego_out_ptr = NULL;	/* so we don't free it again */
	}

	/* try to pull "final" work completion(s) off completion queue */
	while ((ncompletes = ibv_poll_cq(cb->cq, 1, &wc)) > 0)
		{
		TRACE("%s %s: ibv_poll_cq() returned %d entries\n",
			_time_prefix(), __func__, ncompletes);
		PRINT_STATUS(__func__, wc.status);
		}
	if (ncompletes < 0)
		{
		OUR_PERROR("ibv_poll_cq() returned error", -ncompletes);
		}


	if (!cb->is_seqpacket) {
		pthread_mutex_destroy(&cb->stream_sync_mutex);
	}
	pthread_mutex_destroy(&cb->data_sequence_mutex);
	pthread_mutex_destroy(&cb->advert_sequence_mutex);

	TRACE("%s %s: call our_destroy_qp\n",
		_time_prefix(), __func__);
	our_destroy_qp(cb, __func__);

	TRACE("%s %s: call our_destroy_cq\n",
		_time_prefix(), __func__);
	our_destroy_cq(cb, __func__);

	TRACE("%s %s: call our_destroy_comp_channel\n",
		_time_prefix(), __func__);
	our_destroy_comp_channel(cb, __func__);

	TRACE("%s %s: finished ok\n", _time_prefix(), __func__);
}	/* our_free_qp */


/*
 * called from our_connect_routing(), our_connect_done(),
 *	our_client_finishing(), our_deal_with_bad_event()
 *	and three times by our_connect_resolving()
 *
 * when called, event_mutex MAY or MAY NOT be locked
 *
 * asynchronous error exits for various stages of exs_connect()
 *	- to unwind the partial series of allocations
 *
 * Returns whatever value is in ret (should be 0 if all ok, else errno)
 */
static int
our_connect_returning(struct exs_conn *new_conn, int err_level, int ret,
			our_qhandle_t *qhandle, int *signo)
{
	char			errbuf[_MAX_STRERROR_MALLOC];

	if (strerror_r(ret, errbuf, _MAX_STRERROR_MALLOC)) {
		snprintf(errbuf, _MAX_STRERROR_MALLOC, "unknown");
	}
	TRACE("%s %s: enter level %d, state %s(%d), "
		"qhandle %p, error %d: %s\n",
		_time_prefix(), __func__, err_level,
		state_labels[new_conn->conn_state],
		new_conn->conn_state,
		new_conn->conn_pending->qhandle,
		ret, errbuf);

	switch(err_level)
	{
	case 5:
		TRACE("%s %s: calling rdma_disconnect for cb %p,"
			" cm_id %p\n", _time_prefix(), __func__,
			new_conn, new_conn->cm_id);

		new_conn->conn_flags &= ~(EXS_NEED_DISCONNECT
							| EXS_OK_TO_ISSUE_IO);
		rdma_disconnect(new_conn->cm_id);
	case 45:
		exs_buf_free(&new_conn->recv_buf.buf);
	case 4:
		our_cancel_completion_thread(new_conn, __func__);
	case 3:
	case 2:
		/* destroy all the things created in our_setup_qp() */
		our_free_qp(new_conn);
	case 1:
		new_conn->conn_state = new_conn->conn_pending->old_state;
	case 0:
		break;
#if defined(ALLOW_PARANOID_CHECKING)
	default:
		/* should never happen */
		TRACE_ERR("unknown level %d", err_level);
		break;
#endif
	}	/* switch */

	new_conn->conn_pending->err_numb = ret;
	if(new_conn->conn_pending->qhandle != NULL)
	{
		_post_event_to_full_evt_queue(new_conn->conn_pending, qhandle,
									signo);
	}
	else
		*signo = -1;

	/* restore default value for error handling in con_man_thread() */
	new_conn->event_expected = RDMA_CM_EVENT_MULTICAST_ERROR;

	if (strerror_r(ret, errbuf, _MAX_STRERROR_MALLOC)) {
		snprintf(errbuf, _MAX_STRERROR_MALLOC, "unknown");
	}
	TRACE("%s %s: leave, state %s(%d), qhandle %p, "
		"error %d: %s\n",
		_time_prefix(), __func__,
		state_labels[new_conn->conn_state],
		new_conn->conn_state,
		qhandle,
		ret, errbuf);

	return ret;
}	/* our_connect_returning */


/*
 * Called only in our_con_man_thread() context for a client connection
 * called only from our_con_man_thread()
 *	when it receives an RDMA_CM_EVENT_ESTABLISHED
 *	and new_conn->conn_state == EXS_CONN_CONNECTING
 *
 * Starts the negotiation procedure.
 *
 * When called, event_mutex MUST be locked.
 *
 * Returns 0 if all ok, else the error number
 */
static int
our_connect_finishing(struct exs_conn *new_conn, struct rdma_cm_event *cm_event,
			our_qhandle_t *qhandle, int *signo)
{
	const struct our_nego_mess	*nego_recv;
	uint32_t			value;
	(void)qhandle;
	(void)signo;

	/* we have connected, so a future rdma_disconnect is needed */
	new_conn->conn_flags |= EXS_NEED_DISCONNECT;
	new_conn->conn_state = EXS_CONN_CLIENT_SETUP;
	new_conn->event_expected = RDMA_CM_EVENT_DISCONNECTED;

	TRACE("%s %s: Connected on source port %u\n",
		_time_prefix(), __func__,
		ntohs(rdma_get_src_port(new_conn->cm_id)));
	TRACE("%s %s: Connected destination port %u\n",
		_time_prefix(), __func__,
		ntohs(rdma_get_dst_port(new_conn->cm_id)));
	TRACE("%s %s: Dump of src_addr:\n",
		_time_prefix(), __func__);
	DUMP_BUF((void *)&new_conn->cm_id->route.addr.src_addr,
		sizeof(struct sockaddr_storage));
	TRACE("%s %s: Dump of dst_addr:\n",
		_time_prefix(), __func__);
	DUMP_BUF((void *)&new_conn->cm_id->route.addr.dst_addr,
		sizeof(struct sockaddr_storage));

	/* agent's version matched ours, so it returned parameter values
	 * by keeping the minimum of client's and agent's values
	 */
	nego_recv = cm_event->param.conn.private_data;

	TRACE("%s %s: %21s %u.%u.%u\n",
		_time_prefix(), __func__, "agent's exs version",
		 nego_recv->version, nego_recv->subversion,
		 nego_recv->subsubversion);

	value = ntohl(nego_recv->client_credits);
	if (value > new_conn->posted_credits) {
		value = new_conn->posted_credits;
	}
	our_set_send_credit_limits(new_conn, value);
	TRACE("%s %s: %21s %u\n",
			_time_prefix(), __func__, "posted credits",
			new_conn->posted_credits);

	value = ntohl(nego_recv->sp_max_size);
	if (value > new_conn->small_packet_max_size) {
		value = new_conn->small_packet_max_size;
	}
	new_conn->small_packet_max_size = value;
	TRACE("%s %s: %21s %u\n",
			_time_prefix(),__func__,"small packet max size",value);

	if (!new_conn->is_seqpacket && !(new_conn->fd_flags & EXS_FD_NODELAY)) {
		new_conn->remote_recv_buf.ad_buf_addr
			= ntohll(nego_recv->recv_buf_addr);
		new_conn->remote_recv_buf.ad_buf_len
			= ntohl(nego_recv->recv_buf_len);
		new_conn->remote_recv_buf.ad_buf_key
			= ntohl(nego_recv->recv_buf_rkey);
		new_conn->remote_recv_buf.ad_buf_index = 0;
		TRACE("%s %s remote_recv_buf is %#.16" PRIx64 " rkey %#.8" PRIx32 " len %" PRIu32 "\n",
				_time_prefix(), __func__,
				new_conn->remote_recv_buf.ad_buf_addr,
				new_conn->remote_recv_buf.ad_buf_key,
				new_conn->remote_recv_buf.ad_buf_len);
	}

	/* Rather than finishing up at this point, we now
	 * build a SETUP_REQUEST message and send it to the server.
	 * This message contains a standard header and 4 fields:
	 *		a - the current version
	 *		b - the maximum number of buffers we want to be able
	 *		    to send to the server.  This is our current
	 *		    this_conn->posted_remote_credits, and is
	 *		    "What I want to be able to send to you".
	 *		c - the number of buffers we are reserving
	 *		    for the server to send to us.  This is
	 *		    our current cb->posted_local_credits
	 *		    and is "What I am able to receive from you".
	 *		d - the small_packet_max_size we want to use
	 */

	/* now send the SETUP_REQUEST to the agent */
	return our_client_nego_launch(new_conn);
}	/* our_connect_finishing */


/*
 * called only from our_client_finishing()
 */
static void
our_connect_done(struct exs_conn *new_conn,
			our_qhandle_t *qhandle, int *signo)
{

	new_conn->conn_flags |= EXS_OK_TO_ISSUE_IO;
	new_conn->conn_state = EXS_CONN_RUNNING;
	new_conn->event_expected = RDMA_CM_EVENT_DISCONNECTED;

	TRACE("%s %s: Connected on source port %u\n",
		_time_prefix(), __func__,
		ntohs(rdma_get_src_port(new_conn->cm_id)));
	TRACE("%s %s: Connected destination port %u\n",
		_time_prefix(), __func__,
		ntohs(rdma_get_dst_port(new_conn->cm_id)));
	TRACE("%s %s: Dump of src_addr:\n",
		_time_prefix(), __func__);
	DUMP_BUF((void *)&new_conn->cm_id->route.addr.src_addr,
		sizeof(struct sockaddr_storage));
	TRACE("%s %s: Dump of dst_addr:\n",
		_time_prefix(), __func__);
	DUMP_BUF((void *)&new_conn->cm_id->route.addr.dst_addr,
		sizeof(struct sockaddr_storage));

	TRACE("%s done %s ok\n", _time_prefix(), __func__);

	if(new_conn->nego_out_ptr != NULL)
	{	/* need to free nego block */
		TRACE("%s %s: free nego out transaction block %p\n",
			_time_prefix(), __func__, new_conn->nego_out_ptr);
		our_free_transaction_block(new_conn->nego_out_ptr);
		new_conn->nego_out_ptr = NULL;	/* so we don't free it again */
	}

	/* post a "connect complete" event to user's "qhandle" on the return */
	our_connect_returning(new_conn, 0, 0, qhandle, signo);
	new_conn->event_expected = RDMA_CM_EVENT_DISCONNECTED;
}	/* our_connect_done */


/*
 * Called only in our_con_man_thread() context for a client connection
 *
 * called only from our_con_man_thread()
 *	when it receives an RDMA_CM_EVENT_ROUTE_RESOLVED
 *	and new_conn->conn_state == EXS_CONN_ROUTING
 *
 * the networking part of the connection has been completed,
 * so we call our_setup_qp() to set up our queue-pair,
 * mark the state EXS_CONN_SETUP,
 * post some receive buffers,
 * start up the completion thread for this connection,
 * then mark the state EXS_CONN_CONNECTING
 * and call rdma_connect() before departing asynchronously.
 *
 * When rdma_connect() completes,
 * we should get control back in our_connect_finishing().
 *
 * When called, event_mutex MUST be locked.
 *
 * Returns 0 if all ok, else errno if something goes wrong
 */
static int
our_connect_resolving(struct exs_conn *new_conn, our_qhandle_t *qhandle,
			int *signo)
{
	int			ret;
	struct rdma_conn_param	conn_param;

	TRACE("%s %s: calling our_setup_qp\n",
		_time_prefix(), __func__);

	ret = our_setup_qp(new_conn, new_conn->cm_id);
	if(ret)
	{
		TRACE("%s %s: our_setup_qp failed, "
			"ret = %d\n", _time_prefix(), __func__, ret);
		return our_connect_returning(new_conn,1,ret,qhandle,signo);
	}
	TRACE("%s %s: back from our_setup_qp, new qp %p\n",
		_time_prefix(), __func__, new_conn->cm_id->qp);

	new_conn->conn_state = EXS_CONN_SETUP;

	/* create a thread to collect ibv completion events on this connection
	 *
	 * this probably should be done later in the state machine
	 * in our_connect_finishing()
	 * when EXS_NEED_DISCONNECT is set in conn_flags
	 */
	if((ret = our_create_completion_thread(new_conn, __func__)) != 0) {
		/* need to unwind stuff we just did above */
		return our_connect_returning(new_conn,3,ret,qhandle,signo);
	}


	if (!new_conn->is_seqpacket && !(new_conn->fd_flags & EXS_FD_NODELAY)) {
		if ((ret = our_create_stream_recv_buf(new_conn)) < 0) {
			return our_connect_returning(new_conn,4,ret,qhandle,signo);
		}
	}

	our_fill_client_nego_mess(new_conn, &new_conn->nego_send);

	/* see "man rdma_connect" for values of these parameter fields */
	memset(&conn_param, 0, sizeof(conn_param));

	conn_param.private_data = &new_conn->nego_send;
	conn_param.private_data_len = sizeof(new_conn->nego_send);
	conn_param.responder_resources = RESPONDER_RESOURCES;
	conn_param.initiator_depth = INITIATOR_DEPTH;
	conn_param.retry_count = RETRY_COUNT;
	conn_param.rnr_retry_count = RNR_RETRY_COUNT;

	TRACE("%s %s: calling rdma_connect\n",
		_time_prefix(), __func__);

	new_conn->conn_state = EXS_CONN_CONNECTING;
	new_conn->event_expected = RDMA_CM_EVENT_ESTABLISHED;

	if((ret = rdma_connect(new_conn->cm_id, &conn_param)))
	{
		OUR_PERROR("rdma_connect failed", errno);
		return our_connect_returning(new_conn,45,ret,qhandle,signo);
	}
	TRACE("%s %s: rdma_connect started ok\n",
		_time_prefix(), __func__);

	return ret;
}	/* our_connect_resolving */


/*
 * Called only in our_con_man_thread() context for a client connection
 *
 * called only from our_con_man_thread()
 *	when it receives an RDMA_CM_EVENT_ADDR_RESOLVED
 *	and new_conn->conn_state == EXS_CONN_ADDRESSING
 *
 * The addressing part of the connection has been completed,
 * so we mark the state EXS_CONN_ROUTING,
 * and call rdma_resolve_route() before departing asynchronously.
 *
 * When rdma_resolve_route() completes,
 * we should get control back in our_connect_resolving().
 *
 * When called, event_mutex MUST be locked.
 *
 * Returns 0 if all ok, else the error number
 */
static int
our_connect_routing(struct exs_conn *new_conn, our_qhandle_t *qhandle,
			int *signo)
{
	int			ret;

	TRACE("%s %s: calling rdma_resolve_route\n",
		_time_prefix(), __func__);

	new_conn->conn_state = EXS_CONN_ROUTING;
	new_conn->event_expected = RDMA_CM_EVENT_ROUTE_RESOLVED;

	/*---=-=--==--=--=Resolve route time-=--==-=-=-=-=-*/
	if((ret = rdma_resolve_route(new_conn->cm_id,
						2000/*timeout->tv_sec*1000*/)))
	{
		ret = errno;
		OUR_PERROR("rdma resolve route failed", ret);
		return our_connect_returning(new_conn,1,ret,qhandle,signo);
	}
	TRACE("%s %s: rdma_resolve_route started ok\n",
		_time_prefix(), __func__);
	return ret;

}	/* our_connect_routing */


/*
 *	exs_connect - page 17 of Interconnect Software Consortium Issue 1.0
 *		asynchronously connect a socket
 *
 *	returns 0 if async op initiated ok,
 *		else -1 with errno set to error value
 *
 *	This routine starts a sequence of asynchronous actions that
 *	require the following structures to have already been created:
 *		new_conn
 *		new_conn->event_channel
 *		new_conn->cm_id
 *		new_conn->pd
 *		new_conn->comp_channel
 *		new_conn->cq
 *		new_conn->cm_id->qp == new_conn->qp
 *
 * In this routine, we first verify that these have been created
 * and that we have not already gone beyond the state EXS_CONN_BOUND.
 * If this is all in order, we create
 *		new_conn->conn_pending
 * to hold the working parameters of this exs_connect.
 *
 * We then set the conn_state to EXS_CONN_ADDRESSING,
 * and call rdma_resolve_addr() before departing asynchronously.
 *
 * When rdma_resolve_addr() completes,
 *	our_con_man_thread() should get an RDMA_CM_EVENT_ADDR_RESOLVED event
 *	and transfer control in our_connect_routing().
 *
 * If the EXS_BLOCK flag is set in flags, this function will block until the
 * operation completes (i.e., the operation will be synchronous).  In this case,
 * the qhandle and ahandle arguments are ignored.  No other flags are supported.
 */
EXS_EXPORT int
exs_connect(int sockno, const struct sockaddr *address,
		socklen_t address_len, int flags, struct timeval *timeout,
		exs_qhandle_t qhandle, exs_ahandle_t ahandle)
{
	int			ret = 0;
	our_qhandle_t		qptr;
	our_qhandle_t		local_qhandle;
	struct exs_conn		*new_conn;
	exs_event_t		local_vector;
	char			errbuf[_MAX_STRERROR_MALLOC];

#if defined(ALLOW_PID_CHECKING)
	pid_t			current;

	current = getpid();
	TRACE("%s pid %d start %s, fd %d, qhandle %p\n",
		_time_prefix(), current, __func__, sockno, qhandle);

	if(global_owner != current)
#else
	if(global_owner == 0)
#endif
	{
		errno = EPERM;		/* Operation not permitted */
		return -1;
	}

	/* verify that this socket has been created */
	if((new_conn = our_lookup_socket(sockno)) == NULL)
	{
		errno = EBADF;		/* Bad file number */
		return -1;
	}

	/* verify that the user's flags are supported */
	if (flags & ~(EXS_BLOCK)) {
		errno = EINVAL;
		return -1;
	}

	TRACE("%s %s: socket %d, new_conn %p, state %s(%d)\n",
		_time_prefix(), __func__, sockno, new_conn,
		state_labels[new_conn->conn_state], new_conn->conn_state);

	/* see if user gave us an event queue or wants a blocking connect */
	if(flags & EXS_BLOCK)
	{	/* get qhandle from cache, dynamically creating one if needed */
		local_qhandle = our_get_from_queue_cache(new_conn);
		if(local_qhandle == NULL)
		{	/* couldn't create it, errno already set */
			return -1;
		}
		qptr = local_qhandle;	/* use our local qhandle */
	}
	else
	/* verify that caller is giving us a valid qhandle */
	if((qptr = _cnv_qnumb_to_ptr(qhandle)) == NULL)
	{	/* Invalid argument */
		return -1;
	}
	else
	{	/* user supplied his own event queue, use it, no local one */
		local_qhandle = NULL;
	}

	/* verify that this socket has not already been connected or listened */
	if(new_conn->conn_state != EXS_CONN_UNBOUND
				&& new_conn->conn_state != EXS_CONN_BOUND)
	{
		if(local_qhandle != NULL)
		{	/* caller wanted a blocking (synchronous) connect,
			 * delete the locally created queue we were going to use
			 */
			our_put_to_queue_cache(new_conn, local_qhandle);
		}
		errno = EINVAL;   /* Invalid argument */
		return -1;
	}

	/* we lock event_mutex.
	 * At first glance this seems excessive, but at this time
	 * nothing else should be happening on this client socket,
	 * so there should be no contention for either of these locks --
	 * we lock them to explicitly prevent anything else from happening!
	 */

	/* need to lock event_mutex to protect conn_state */
	TRACE("\t\t%4d:     LOCK cb %p event_mutex %p\n",
		__LINE__, new_conn, &new_conn->event_mutex);
	pthread_mutex_lock(&new_conn->event_mutex);
	TRACE("\t\t%4d:     PAST cb %p event_mutex %p\n",
		__LINE__, new_conn, &new_conn->event_mutex);

	/* check again that this socket has not already been connected */
	if(new_conn->conn_state != EXS_CONN_UNBOUND
				&& new_conn->conn_state != EXS_CONN_BOUND)
	{
		errno = EINVAL;   /* Invalid argument */
		ret = -1;
	}
	else
	{
		TRACE("%s %s: Begin connection process\n",
			_time_prefix(), __func__);

		/* fill in ALL fields of struct pending_connect for this conn */
		new_conn->conn_pending->qhandle = qptr;
		new_conn->conn_pending->ahandle = ahandle;
		new_conn->conn_pending->err_numb = 0;
		new_conn->conn_pending->sockno = sockno;
		new_conn->conn_pending->evt_type = EXS_EVT_CONNECT;
		new_conn->conn_pending->old_state = new_conn->conn_state;

		new_conn->conn_state = EXS_CONN_ADDRESSING;
		new_conn->event_expected = RDMA_CM_EVENT_ADDR_RESOLVED;

		/*-=-=-=-=--=--=-=Resolve addr-==-=-=--=-==-=-*/
		(void)address_len; /* not required by rdma_bind_addr */
		ret = rdma_resolve_addr(new_conn->cm_id, NULL,
			(struct sockaddr *)address, 2000);
		/* FIXME: timeout parameter not respected because there must be
		 * multiple RDMA operations in order to do an exs_connect(). */
		if(ret)
		{	/* rdma_resolve_addr() itself failed */
			OUR_PERROR("rdma_resolve_addr failed", errno);
			errno = ret;
			new_conn->conn_state=new_conn->conn_pending->old_state;
		}
	}

	TRACE("\t\t%4d:   UNLOCK cb %p event_mutex %p\n",
		__LINE__, new_conn, &new_conn->event_mutex);
	pthread_mutex_unlock(&new_conn->event_mutex);

	if(ret)
		goto err0;

	TRACE("%s %s: rdma_resolve_addr started ok\n",
		_time_prefix(), __func__);

	if(local_qhandle == NULL)
		goto outok;

	/* caller wants a blocking (synchronous) connect */

	TRACE("%s %s: blocking connect, local_qhandle %p, "
		"ahandle %p\n", _time_prefix(),
		__func__, local_qhandle, ahandle);

	/* connect was started ok, now wait for it to complete */
	ret = _evt_queue_dequeue_vector(local_qhandle, &local_vector, 1, NULL);

	TRACE("%s %s: back from dequeue, local_qhandle %p, "
		"ahandle %p, ret %d, "
		"exs_evt_type %d, exs_evt_errno %d, "
		"exs_evt_ahandle %p\n",
		_time_prefix(), __func__,
		local_qhandle, ahandle, ret,
		local_vector.exs_evt_type,
		local_vector.exs_evt_errno,
		local_vector.exs_evt_ahandle);

	if(ret == -1 || (errno = local_vector.exs_evt_errno) != 0)
		goto err0;

	/* always return the locally created queue to the connection's cache */
	our_put_to_queue_cache(new_conn, local_qhandle);

outok:
	TRACE("%s %s: finished ok, new_conn %p, state %s(%d)\n",
		_time_prefix(), __func__, new_conn,
		state_labels[new_conn->conn_state], new_conn->conn_state);

	return 0;

err0:
	ret = errno;
	if(local_qhandle != NULL)
	{	/* caller wanted a blocking (synchronous) connect,
		 * delete the locally created queue
		 */
		our_put_to_queue_cache(new_conn, local_qhandle);
	}
	if (strerror_r(ret, errbuf, _MAX_STRERROR_MALLOC)) {
		snprintf(errbuf, _MAX_STRERROR_MALLOC, "unknown");
	}
	/* FIXME: What failed? */
	TRACE_ERR("failed, new_conn %p, state %s(%d) error %d: %s", new_conn,
		state_labels[new_conn->conn_state], new_conn->conn_state,
		ret, errbuf);
	return -1;
}	/* exs_connect */


/*
 *	exs_blocking_connect()
 *			- not in Interconnect Software Consortium Issue 1.0
 *			      was added by us
 *
 *	returns == 0  if connected ok
 *		== -1 if error (with errno set to error value)
 */
EXS_EXPORT int
exs_blocking_connect(int sockno, const struct sockaddr *address,
		socklen_t address_len)
{
	return exs_connect(sockno, address, address_len, EXS_BLOCK,
					NULL, NULL, NULL);
}	/* exs_blocking_connect */


/*
 * called only from our_common_send()
 *
 * to send an RDMA_WRITE with remote notification
 * local side will write this buffer as part of the send
 *
 * This posts an IBV_WR_RDMA_WRITE_WITH_IMM if the operation is supported by the
 * underlying transport (all except for iWARP). On iWARP, this function posts an
 * IBV_WR_RDMA_WRITE followed by an IBV_WR_SEND.
 *
 * send_unreg_mr will be set to a ibv_mr that will need to be deregistered after
 * the WRITE operation completes, or NULL if nothing needs to be deregistered.
 *
 * On input, if IBV_SEND_SIGNALED is set in send_flags, a signal will be forced
 * for this WRITE whether or not it is needed.  On output, send_flags will
 * contain all flags that were set in the work request for the RDMA WRITE
 * request (plus the IBV_SEND_SIGNALED flag if the transport is iWARP and the
 * IBV_WR_SEND request needed to be signaled).
 *
 * Neither send_unreg_mr nor send_flags should be NULL.
 *
 * returns number of bytes actually sent if our_post_send() worked
 * successfully, -1 on any error
 */
static ssize_t
our_send_rdma_write_notify(struct exs_conn *this_conn,
		struct transaction_block *block,
		const void *buffer, size_t length, exs_mhandle_t mhandle,
		struct our_advert *ad_ptr, const char *message,
		int *send_flags, struct ibv_mr **send_unreg_mr)
{
	int			ret;
	struct ibv_mr		*use_this_mr;
	int			force_signaled;
	uint64_t		remote_addr;
	uint32_t		remote_key;
	uint32_t		imm_data;
	uint32_t		imm_flags;
	unsigned int		new_data_signal_count;

	TRACE("%s %s started this_conn %p block %p buffer %p length %zd ad_ptr %p send_flags %8x\n",
				_time_prefix(), __func__, (void *)this_conn,
				(void *)block, (void *)buffer, length,
				(void *)ad_ptr, (unsigned)*send_flags);

	/* Assume this will fail */
	ret = -1;

	if (ad_ptr) {
		remote_addr = ad_ptr->ad_buf_addr;
		remote_key = ad_ptr->ad_buf_key;
		TRACE_DBG(EXS_LOG_UNSPECIFIED,
			"%s addr %#" PRIx64 " rkey %#" PRIx32,
			message, remote_addr, remote_key);
	} else {
		remote_addr = this_conn->remote_recv_buf.ad_buf_addr
			+ this_conn->remote_buf_offset;
		remote_key = this_conn->remote_recv_buf.ad_buf_key;

		/* and don't send more than receiver can handle */
		if (length > this_conn->remote_buf_n_empty) {
			TRACE_DBG(EXS_LOG_UNSPECIFIED,
				"length %zu < remote_buf_n_empty %u\n",
				length, this_conn->remote_buf_n_empty);
			length = this_conn->remote_buf_n_empty;
		}
		/* and don't go over the edge of the buffer */
		if (length > (this_conn->remote_recv_buf.ad_buf_len
					- this_conn->remote_buf_offset)) {
			TRACE_DBG(EXS_LOG_UNSPECIFIED,
				"length %zu < remaining buffer space %u\n",
				length, this_conn->remote_recv_buf.ad_buf_len
					- this_conn->remote_buf_offset);
			length = this_conn->remote_recv_buf.ad_buf_len
				- this_conn->remote_buf_offset;
		}
	}

	/* output arguments must not be NULL */
	if (send_unreg_mr == NULL) {
		TRACE("%s %s %s send_unreg_mr output parameter is NULL\n",
				_time_prefix(), message, __func__);
		errno = EINVAL;
		return -1;
	}
	if (send_flags == NULL) {
		TRACE("%s %s %s send_flags output parameter is NULL\n",
				_time_prefix(), message, __func__);
		errno = EINVAL;
		return -1;
	}

	force_signaled = (*send_flags & IBV_SEND_SIGNALED);

	if (length < this_conn->inline_limit) {
		/* Send WRITE inline */
		if ((new_data_signal_count = atomic_add_reset(1,
					this_conn->signal_limit,
					&this_conn->data_signal_count)) != 0
							&& !force_signaled) {
			block->writereq_wr.send_flags = IBV_SEND_INLINE;
			STATISTICS_INCREMENT(unsignaled_rdma_write_blocks_sent);
		} else {
			/* time to force a periodic signal */
			TRACE("%s %s force a signal, data_signal_count %d, signal_limit %d\n",
				_time_prefix(), __func__,
				new_data_signal_count, this_conn->signal_limit);
			block->writereq_wr.send_flags
					= IBV_SEND_INLINE|IBV_SEND_SIGNALED;
			STATISTICS_INCREMENT(signaled_rdma_write_blocks_sent);
		}
		use_this_mr = NULL;
	} else {
		/* Cannot send WRITE inline; must send signaled */
		block->writereq_wr.send_flags = IBV_SEND_SIGNALED;
		if (block->send_unreg_mr != NULL) {
			/* We already dynamically registered this buffer, so use
			 * the buffer that it was so expensive to create */
			use_this_mr = *send_unreg_mr;
		} else if(mhandle == EXS_MHANDLE_UNREGISTERED) {
			/* this data was not registered by the user */
			if ((this_conn->fd_flags & EXS_FD_SP_COPY_UNREG_ON_SEND)
					&& (length
					<= this_conn->small_packet_max_size)) {
				/* copy the data */
				TRACE("%s %s this_conn %p block %p copying %zu bytes\n",
						_time_prefix(), message,
						this_conn, block, length);
				memcpy(block->send_buf, buffer, length);
				buffer = block->send_buf;
				use_this_mr = block->send_buf_mr;
			} else {
				/* dynamically register user's data */
				*send_unreg_mr = our_call_ibv_reg_mr(
							this_conn->pd,
							(void *)buffer, length,
							IBV_ACCESS_REMOTE_READ,
							"user send buffer");
				if(*send_unreg_mr == NULL)
				{
					goto err1;
				}
				use_this_mr = *send_unreg_mr;
			}
		} else {
			/* this data was already registered by the user */
			TRACE("%s %s: mhandle %p is registered\n",
					_time_prefix(), message, mhandle);
			use_this_mr = our_ensure_buffer_registered(this_conn,
					mhandle, message);
			if (use_this_mr == NULL) {
				/* should be able to register user buffer
				 * but couldn't */
				goto err1;
			}
		}
	}

	if (this_conn->is_seqpacket) {
		imm_data = ad_ptr->ad_buf_index;
		imm_flags = 0;
	} else {
		imm_data = this_conn->send_seqno;
		this_conn->send_seqno += length;
		imm_flags = (ad_ptr ? IMM_DIRECT_BIT : 0);
	}

	if (!(this_conn->fd_flags & EXS_FD_NO_WRITE_WITH_IMM)) {
		/* set expected completion code to be RDMA_WRITE */
		block->expected_comp = IBV_WC_RDMA_WRITE;
		block->writereq_wr.imm_data = htonl((imm_data & IMM_INDEX_MASK)
						| (imm_flags & IMM_FLAG_MASK));
	} else {
		block->writereq_wr.send_flags &= ~IBV_SEND_SIGNALED;
		block->writeack_wr.send_flags |= IBV_SEND_SIGNALED;
		block->expected_comp = IBV_WC_SEND;
		block->recv_header.hd_type = OUR_WRITE_NOTIFY;
		block->writeack_header.hd_type
					= htonl(block->recv_header.hd_type);
		block->writeack_buf.write_len = htonl(length);
		block->writeack_buf.write_index = htonl(imm_data);
		block->writeack_buf.write_flags = htonl(imm_flags);
	}

	block->buf_offset += length;
	block->send_total_bytes += length;

	/* send an RDMA_WRITE_WITH_IMM with the data */
	block->writereq_wr.wr.rdma.rkey = remote_key;
	block->writereq_wr.wr.rdma.remote_addr = remote_addr;
	block->writereq_wr.sg_list->length = length;
	block->writereq_wr.sg_list->addr
			= (uint64_t)(unsigned long)buffer;
	if (use_this_mr) {
		block->writereq_wr.sg_list->lkey = use_this_mr->lkey;
	}

	if (our_post_send(this_conn, &block->writereq_wr)) {
		goto err2;
	}
	if (block->writereq_wr.send_flags & IBV_SEND_SIGNALED) {
		this_conn->data_signal_count = 0;
	}

	STATISTICS_INCREMENT(rdma_write_blocks_sent);
	STATISTICS_ADD(rdma_write_bytes_sent, length);

	*send_flags = block->writereq_wr.send_flags;
	if ((this_conn->fd_flags & EXS_FD_NO_WRITE_WITH_IMM)
			&& (block->writeack_wr.send_flags & IBV_SEND_SIGNALED)){
		*send_flags |= IBV_SEND_SIGNALED;
	}

	ret = length;
	goto err1;
err2:
	if(*send_unreg_mr != NULL) {
		ibv_dereg_mr(*send_unreg_mr);
		*send_unreg_mr = NULL;
	}
err1:

	TRACE("%s %s returning %d this_conn %p block %p buffer %p length %zd ad_ptr %p send_flags %8x send_unreg_mr %p\n",
				_time_prefix(), __func__, ret,
				(void *)this_conn, (void *)block,
				(void *)buffer, length,
				(void *)ad_ptr, (unsigned)*send_flags,
				(void *)*send_unreg_mr);
	return ret;
}	/* our_send_rdma_write_notify */

/*
 * called only from exs_recv()
 *
 * to set up and send a buffer advertisement so the
 * remote side can write directly to this buffer
 *
 * returns 0 if advertisement was successfully posted; else -1 on any error
 */
static int
our_send_advert(struct exs_conn *this_conn, struct transaction_block *block,
		const void *buffer, size_t length, struct ibv_mr *mr,
		unsigned int index)
{
	TRACE("%s %s: conn %p, block %p, advert_buf %p,"
		" advert_wr %p, advert_sg_list %p, mr %p index %"PRIu32" flags %"PRIu32"\n",
		_time_prefix(),
		__func__, this_conn, block, &block->advert_buf,
		&block->send_wr, block->send_wr.sg_list, mr,
		index & IMM_INDEX_MASK,
		index & IMM_FLAG_MASK);

	/* Do some setup for the header for this message */
	block->expected_comp = IBV_WC_SEND;
	block->recv_header.hd_type = OUR_ADVERTISEMENT;
	block->recv_header.hd_seqno = this_conn->source_seqno;
	TRACE_DBG(EXS_LOG_UNSPECIFIED,
		"assigned next advert hd_seqno %u\n",
		block->recv_header.hd_seqno);

	/* Increment source_seqno.  For SOCK_SEQPACKET, we increment it per
	 * message.  For SOCK_STREAM, we increment it per byte. */
	block->advert_buf.ad_buf_seqno = htonll(this_conn->source_seqno);
	if (this_conn->is_seqpacket) {
		++this_conn->source_seqno;
	} else {
		this_conn->source_seqno += length;
	}

	block->send_header.hd_type = htonl(block->recv_header.hd_type);
	block->send_header.hd_seqno = htonl(block->recv_header.hd_seqno);

	/* send_sg_list[0] permanently points to send_header
	 * send_sg_list[1] permanently points to advert_buf
	 * so fill in advert_buf with description of user's buffer
	 */
	if (block->recv_flags & RECV_COPY_UNREG_BIT) {
		block->advert_buf.ad_buf_addr
			= htonll((uint64_t)(unsigned long)block->recv_rdma_buf);
		block->advert_buf.ad_buf_key
			= htonl(block->recv_rdma_mr->rkey);
	} else {
		block->advert_buf.ad_buf_addr
				= htonll((uint64_t)(unsigned long)buffer);
		block->advert_buf.ad_buf_key
				= htonl(mr->rkey);
	}
	block->advert_buf.ad_buf_len
			= htonl(length);
	if (this_conn->is_seqpacket) {
		block->advert_buf.ad_buf_index = htonl(index);
		block->advert_buf.ad_buf_flags = 0;
	} else {
		block->advert_buf.ad_buf_index = htonl(this_conn->source_seqno);
		block->advert_buf.ad_buf_flags = htonl(
			block->buffer_flags & MSG_WAITALL);
	}

	TRACE("%s %s: remote key 0x%08x, "
				"remote addr 0x%016llx, remote length %d\n",
				_time_prefix(), __func__,
				ntohl(block->advert_buf.ad_buf_key),
				(unsigned long long)(unsigned long)buffer,
				(int)length);

	block->send_wr.send_flags = 0;

	/* point 2nd send s/g list element to point to advert_buf message */
	block->send_sg_list[1].addr = (uint64_t)(uintptr_t)&block->advert_buf;
	block->send_sg_list[1].length = sizeof(block->advert_buf);
	block->send_sg_list[1].lkey = block->block_mr->lkey;

	if ((sizeof(block->send_header) + sizeof(block->advert_buf))
						<= this_conn->inline_limit) {
		block->send_wr.send_flags |= IBV_SEND_INLINE;
		if (atomic_add_reset(1, this_conn->signal_limit,
					&this_conn->advert_signal_count) == 0) {
			/* time to force a periodic signal */
			TRACE_DBG(EXS_LOG_UNSPECIFIED,
				"force a signal, advert_signal_count %d, signal_limit %d\n",
				this_conn->advert_signal_count,
				this_conn->signal_limit);
			atomic_set_bits(ADVERT_SENDING_BIT, &block->send_state);
			block->send_wr.send_flags |= IBV_SEND_SIGNALED;
			STATISTICS_INCREMENT(signaled_advert_blocks_sent);
		} else {
			STATISTICS_INCREMENT(unsignaled_advert_blocks_sent);
		}
	} else {
		atomic_set_bits(ADVERT_SENDING_BIT, &block->send_state);
		block->send_wr.send_flags |= IBV_SEND_SIGNALED;
		this_conn->advert_signal_count = 0;
		STATISTICS_INCREMENT(signaled_advert_blocks_sent);
	}

	/* now post send of this advert, then leave */
	TRACE_DBG(EXS_LOG_UNSPECIFIED,
			"send advert: seqno %" PRIu64 " n_empty %u",
			this_conn->source_seqno,
			this_conn->recv_buf.n_empty);
	if (our_post_send(this_conn, &block->send_wr)) {
		/* got an input/output error, so advert was NOT SENT */
		atomic_clear_bits(ADVERT_SENDING_BIT, &block->send_state);
		return -1;
	}
	STATISTICS_INCREMENT(advert_blocks_sent);

	TRACE("%s %s: finished ok\n", _time_prefix(), __func__);

	return 0;
}	/* our_send_advert */


/*
 * called from exs_send() and exs_sendfile(), and our_deal_with_completion()
 *
 * sets up and posts the send of
 *	the WWI message for 1 registered or large unregistered data buffer
 *
 * If the WWI can be sent inline and unsignalled, the user event to post will be
 * placed into qhandle and a signal to push will be placed into signo.
 * Otherwise, qhandle will be set to NULL and signo will be set to -1.
 *
 * returns 0 if all ok, else -1 with errno set on any error
 */
static int
our_xmit_if_match_seqpacket(struct exs_conn *this_conn,
		struct transaction_block *this,
		our_qhandle_t *qhandle, int *signo,
		const char *message)
{
	int		err;
	int		send_flags;
	uint32_t	wwi_length;
	uint32_t	advert_length;
	char		errbuf[_MAX_STRERROR_MALLOC];
	exs_mhandle_t	mhandle;
	const void *buffer;

	*qhandle = NULL;	/* assume we will not send any event */
	*signo = -1;		/* assume we will not send any signal */

	/* Set length */
	wwi_length = this->user_event_ptr->exs_evt_union
						.exs_evt_xfer.exs_evt_length;
	advert_length = this->advert_buf.ad_buf_len;
	if (wwi_length > advert_length) {
		wwi_length = advert_length;
	}

	/* Clear the bits */
	atomic_clear_bits(ADVERT_READY_BIT | SEND_READY_BIT,
						&this->send_state);

	buffer = this->user_event_ptr->exs_evt_union.exs_evt_xfer.exs_evt_buffer;
	mhandle = this->user_event_ptr->exs_evt_union.exs_evt_xfer.exs_evt_mhandle;

	send_flags = IBV_SEND_SIGNALED;
	if (our_send_rdma_write_notify(this_conn, this, buffer, wwi_length,
					mhandle, &this->advert_buf, message,
					&send_flags, &this->send_unreg_mr) < 0) {
		err = errno;
		goto err1;
	}

	if (!(send_flags & IBV_SEND_SIGNALED)) {
		TRACE("unsignalled write and last\n");
		STATISTICS_INCREMENT(send_done_after_sending);
		*signo = our_all_done_with_block(this_conn, this,
							qhandle, 0, message);
	}

	TRACE("%s %s: finished ok\n",
			_time_prefix(), message);
	return 0;

err1:
	err = errno;
	if (strerror_r(err, errbuf, _MAX_STRERROR_MALLOC)) {
		snprintf(errbuf, _MAX_STRERROR_MALLOC, "unknown");
	}
	TRACE_ERR("%s: failed, this_conn %p, state %s(%d) error %d: %s",
		message, this_conn,
		state_labels[this_conn->conn_state], this_conn->conn_state,
		err, errbuf);
	errno = err;
	return -1;
}	/* our_xmit_if_match_seqpacket */


/* Returns 1 if the send was successful and completes the exs_send() operation,
 * 0 if the send was successful but we still must wait to send more data or for
 * a completion, and -1 if there was an error (check errno). */
static int
our_handle_match_stream(struct exs_conn *this_conn,
		struct transaction_block *this,
		struct stream_advert_block *advert,
		const char *message)
{
	uint32_t	wwi_length;
	uint32_t	advert_length;
	exs_mhandle_t	mhandle;
	int done_with_send, recv_flags, send_flags;
	struct exs_evt_xfer *xfer;
	const void *buffer;

	xfer = &this->user_event_ptr->exs_evt_union.exs_evt_xfer;

	memcpy(&this->advert_buf, &advert->advert_buf,
						sizeof(this->advert_buf));
	recv_flags = this->advert_buf.ad_buf_flags;
	if (advert->offset > 0) {
		this->advert_buf.ad_buf_addr += advert->offset;
		this->advert_buf.ad_buf_len -= advert->offset;
	}

	/* Set length */
	wwi_length = xfer->exs_evt_length - this->buf_offset;
	advert_length = this->advert_buf.ad_buf_len;

	if ((done_with_send = (wwi_length <= advert_length))) {
		atomic_clear_bits(SEND_READY_BIT, &this->send_state);
		this_conn->next_send_primed = this->next;
	} else {
		wwi_length = advert_length;
	}

	advert->offset += wwi_length;
	if (advert_length <= wwi_length || !(recv_flags & MSG_WAITALL)) {
		atomic_clear_bits(ADVERT_READY_BIT, &advert->advert_state);
		this_conn->next_advert_primed = advert->next;
	}

	TRACE_DBG(EXS_LOG_STREAM,
		"[via %s] seqno %"PRIu64" matched send_block %p offset %zu length %u advert_block %p offset %u length %u",
		message,
		this_conn->send_seqno,
		(void *)this, this->buf_offset, wwi_length,
		(void *)advert, advert->offset - wwi_length,
		advert_length);

	buffer = xfer->exs_evt_buffer + this->buf_offset;
	mhandle = xfer->exs_evt_mhandle;

	atomic_add_reset(1, UINT_MAX, &this->outstanding_wwi_count);
	send_flags = IBV_SEND_SIGNALED;
	if (our_send_rdma_write_notify(this_conn, this, buffer,
					wwi_length, mhandle,
					&this->advert_buf,
					message, &send_flags,
					&this->send_unreg_mr) < 0) {
		atomic_add_reset(-1, UINT_MAX, &this->outstanding_wwi_count);
		return -1;
	}

	return (done_with_send && !(send_flags & IBV_SEND_SIGNALED)) ? 1 : 0;
}


static int
our_xmit_if_match_stream(struct exs_conn *this_conn,
		struct transaction_block *this,
		our_qhandle_t *qhandle, int *signo,
		const char *message)
{
	int		err, ret;
	char		errbuf[_MAX_STRERROR_MALLOC];

	*qhandle = NULL;	/* assume we will not send any event */
	*signo = -1;		/* assume we will not send any signal */

	while (1) {
		OUR_MEMBER_MUTEX_LOCK(this_conn, stream_sync_mutex);
		if ((this_conn->next_send_primed->send_state & SEND_READY_BIT)
				&& (this_conn->next_advert_primed->advert_state
							& ADVERT_READY_BIT)) {
			this = this_conn->next_send_primed;
			ret = our_handle_match_stream(this_conn, this,
					this_conn->next_advert_primed,
					message);
			OUR_MEMBER_MUTEX_UNLOCK(this_conn, stream_sync_mutex);

			if (ret < 0) {
				goto err1;
			} else if (ret > 0) {
				TRACE_DBG(EXS_LOG_STREAM,
					"[via %s] completed write operation for block %p\n",
					message, this);
				STATISTICS_INCREMENT(send_done_after_sending);
				*signo = our_all_done_with_block(this_conn,
						this, qhandle, 0, message);
			}
		} else {
			OUR_MEMBER_MUTEX_UNLOCK(this_conn, stream_sync_mutex);
			break;
		}
	}

	TRACE("%s %s: finished ok\n",
			_time_prefix(), __func__);
	return 0;

err1:
	err = errno;
	if (strerror_r(err, errbuf, _MAX_STRERROR_MALLOC)) {
		snprintf(errbuf, _MAX_STRERROR_MALLOC, "unknown");
	}
	TRACE_ERR("%s: failed, this_conn %p, state %s(%d) error %d: %s",
		message, this_conn,
		state_labels[this_conn->conn_state], this_conn->conn_state,
		err, errbuf);
	errno = err;
	return -1;
}	/* our_xmit_if_match_stream */

static int
our_xmit_if_match_streamcopy(struct exs_conn *this_conn,
		struct transaction_block *this,
		our_qhandle_t *qhandle, int *signo,
		const char *message)
{
	struct stream_advert_block *advert;
	int done_with_send;
	int		ret;

	*qhandle = NULL;	/* assume we will not send any event */
	*signo = -1;		/* assume we will not send any signal */

	OUR_MEMBER_MUTEX_LOCK(this_conn, stream_sync_mutex);
	while (1) {
		TRACE_DBG(EXS_LOG_STREAM,
				"remote_buf_n_empty %u",
				this_conn->remote_buf_n_empty);
		this = this_conn->next_send_primed;
		if (this->send_state & SEND_READY_BIT) {
			advert = this_conn->next_advert_primed;
			if (advert->advert_state & ADVERT_READY_BIT) {
				if (advert->advert_buf.ad_buf_seqno
						!= this_conn->send_seqno) {
					/* We already sent at least part of this
					 * data, so throw away the advert */
					atomic_clear_bits(ADVERT_READY_BIT,
							&advert->advert_state);
					this_conn->next_advert_primed
								= advert->next;
					continue;
				}
				ret = our_handle_match_stream(this_conn, this,
					this_conn->next_advert_primed,
					message);
				if (ret < 0) {
					OUR_PERROR("could not send WWI", errno);
				} else if (ret > 0) {
					TRACE_DBG(EXS_LOG_STREAM,
						"[via %s] completed write operation for block %p",
						message, this);
					STATISTICS_INCREMENT(send_done_after_sending);
					if (atomic_set_bits(SEND_DONE_BIT,
							&this->send_state)
							& ACK_DONE_BIT) {
						*signo = our_all_done_with_block(this_conn,
							this, qhandle, 0,
							message);
					}
				}
			} else if (this_conn->remote_buf_n_empty > 0) {
				ret = our_send_direct(this_conn, this);
				done_with_send = (ret != 0);
				if (this->buf_offset >= this->user_event_ptr
						->exs_evt_union.exs_evt_xfer
						.exs_evt_length) {
					atomic_clear_bits(SEND_READY_BIT,
							&this->send_state);
					this_conn->next_send_primed
						= this->next;
				}
			} else {
				break;
			}
		} else {
			break;
		}
	}
	OUR_MEMBER_MUTEX_UNLOCK(this_conn, stream_sync_mutex);

	TRACE("%s %s: finished ok\n",
			_time_prefix(), message);
	return 0;
}	/* our_xmit_if_match_streamcopy */

/*
 *	exs_send - page 51 of Interconnect Software Consortium Issue 1.0
 *		asynchronously send a message on a socket
 *
 *	returns == 0  if async op initiated ok,
 *		>= 0  if sync op finished ok (value is no. of bytes written),
 *		== -1 if error ( errno set to error value)
 *
 *	only flags values accepted:
 *
 *		EXS_DONTWAIT		(nonstandard)
 *			without this flag, this function may block until a
 *				posted_send_slot becomes available (happens when
 *				user has too many outstanding exs_send calls)
 *			with this flag, this function will return EAGAIN error
 *				if no posted_send_slot is available
 *			with or without this flag, if this function returns 0
 *				then the data has not been sent until the
 *				completion event has been posted to qhandle.
 *		EXS_BLOCK		(nonstandard)
 *			force this send to block until it completes.  The
 *			qhandle and ahandle parameters are ignored.
 *		EXS_UNSIGNALED		(nonstandard)
 *			don't cause a completion event when this send completes.
 *			The qhandle and ahandle parameters are ignored.
 *
 *	qhandle values accepted:
 *		NULL
 *			don't cause a completion event when this send completes
 *		EXS_QHANDLE_INVALID
 *			left over from a previous error, cause a new error
 *		anything else
 *			assume it is a user-created queue
 *
 *	this may cause the completion event to be posted before returning
 */
EXS_EXPORT ssize_t
exs_send(int sockno, const void *buffer, size_t length, int flags,
		exs_qhandle_t qhandle, exs_ahandle_t ahandle,
		exs_mhandle_t mhandle)
{
	ssize_t				ret;
	int				signo;
	our_qhandle_t			qptr;
	our_qhandle_t			local_qhandle;
	struct exs_conn			*this_conn;
	struct transaction_block	*this;
	exs_event_t			local_vector;

#if defined(ALLOW_PID_CHECKING)
	pid_t				current;

	current = getpid();
	TRACE("%s pid %d start %s, qhandle %p, "
		"length %d, flags 0x%x\n",
		_time_prefix(), current, __func__, qhandle, (int)length, flags);

	if(global_owner != current)
#else
	if(global_owner == 0)
#endif
	{
		errno = EPERM;		/* Operation not permitted */
		return -1;
	}

	/* verify that this socket has been created */
	if((this_conn = our_lookup_socket(sockno)) == NULL)
	{
		errno = EBADF;		/* Bad file number */
		return -1;
	}

	local_qhandle = NULL;		/* assume this is not created locally */

	if(this_conn->fd_flags & EXS_FD_BYPASS_PARAM_CHECKING)
		goto no_parameter_checks;

	/* verify that this socket has already been connected */
	if(this_conn->conn_state != EXS_CONN_RUNNING)
	{
		/* Transport endpoint is not connected */
		errno = ENOTCONN;
		return -1;
	}

	TRACE("%s %s: socket %d this_conn %p state %s(%d) "
		"qhandle %p flags %#08x length %d\n", _time_prefix(), __func__,
		sockno, this_conn,
		state_labels[this_conn->conn_state], this_conn->conn_state,
		qhandle, flags, (int)length);

	/* verify that caller is either giving us a valid qhandle or NULL */
	if(qhandle == EXS_QHANDLE_INVALID)
	{
		errno = EINVAL;		/* Invalid argument */
		return -1;
	}

	/* verify that flags are supported for this socket type or protocol */
	if((flags & ~(EXS_DONTWAIT|EXS_BLOCK|EXS_UNSIGNALED))!=0)
	{
		errno = EOPNOTSUPP;	/* Operation not supported */
		return -1;
	}

	/* if user specified a NULL qhandle, then ensure that either EXS_BLOCK
	 * or EXS_UNSIGNALED flag was specified. */
	if (qhandle == NULL && !(flags & (EXS_BLOCK|EXS_UNSIGNALED))) {
		errno = EINVAL;		/* Invalid argument */
		return -1;
	}

	/* verify that length is at least 1 and less than 1 Gigabyte*/
	if (length <= 0 || length > 1000000000L) {
		errno = EINVAL;		/* Invalid argument */
		return -1;
	}

	/* ensure that buffer is not NULL,
	 * TO-DO -- and is not an outstanding recycle buffer,
	 * since that leads to a race condition
	 */
	if(buffer == NULL)
	{
		errno = EINVAL;		/* Invalid argument */
		return -1;
	}

no_parameter_checks:

	/* see if user gave us an event queue or wants a blocking send.  After
	 * this, qptr will be set or this function will have returned an error
	 * condition. */
	if (flags & EXS_BLOCK) {
		/* get qhandle from cache, dynamically creating one if needed */
		local_qhandle = our_get_from_queue_cache(this_conn);
		if(local_qhandle == NULL)
		{	/* couldn't create it, errno already set */
			return -1;
		}
		qptr = local_qhandle;
	} else if (qhandle == NULL) {
		/* caller gave us no qhandle, so don't use one */
		qptr = NULL;
	} else if ((qptr = _cnv_qnumb_to_ptr(qhandle)) == NULL) {
		/* Invalid argument */
		errno = EINVAL;
		return -1;
	}

	if (flags & EXS_DONTWAIT) {
		OUR_MEMBER_MUTEX_TRYLOCK(this_conn, data_sequence_mutex, &ret);
		if (ret != 0) {
			/* Returned errno is most likely EAGAIN, but catch other
			 * errors as well. */
			errno = ret;
			return -1;
		}
	} else {
		OUR_MEMBER_MUTEX_LOCK(this_conn, data_sequence_mutex);
	}

	if (!(this_conn->conn_flags & EXS_OK_TO_ISSUE_IO))
	{	/* this side called exs_close() or other side disconnected */
		errno = ENOTCONN;
		ret = -1;
		goto err0;
	}

	/* get a flow-control send slot so we don't overrun the receiver */
	if((flags & EXS_DONTWAIT)) {
		/* don't block if all empty send slots are currently used up */
		TRACE("TRYWAIT next_send %p sem %p\n", this_conn->next_send,
				&this_conn->next_send->sem);
		if(sem_trywait(&this_conn->next_send->sem) < 0) {

			TRACE("%s %s: no free send slots on cb %p\n",
				_time_prefix(), __func__, this_conn);

			errno = EAGAIN;
			ret = -1;
			goto err0;
		}
		TRACE("PAST next_send %p sem %p\n", this_conn->next_send,
				&this_conn->next_send->sem);
	}
	else
	{	/* may block if all empty send slots are currently used up */
		OUR_MEMBER_SEM_WAIT(this_conn->next_send, sem);
	}

	atomic_add_reset(1, UINT_MAX, &this_conn->pending_send_count);

	/* Verify again that we are not in the middle of disconnecting */
	if ((this_conn->conn_flags & (EXS_OK_TO_ISSUE_IO|EXS_NEED_DISCONNECT))
				!= (EXS_OK_TO_ISSUE_IO|EXS_NEED_DISCONNECT))
	{	/* this side has closed or other side has disconnected */
		OUR_MEMBER_SEM_POST(this_conn->next_send, sem);

		errno = ENOTCONN;
		ret = -1;
		goto err0;
	}

	if (!this_conn->next_send->user_event_ptr) {
		/* Buffer for completion event is not available */
		OUR_MEMBER_SEM_POST(this_conn->next_send, sem);

		errno = ENOBUFS;
		ret = -1;
		goto err0;
	}

	/* get an unused transaction block for sending */
	this = this_conn->next_send;
	this_conn->next_send = this->next;

	/* send transactions keep extra state information */
	our_setup_send_state(this);

	/* set block's buffer_flags field with flags from the exs_send */
	this->buffer_flags = flags;

	/* fill in transaction's fields with fixed parameters from exs_send */
	this->in_progress = 1;
	this->buf_offset = 0;
	this->mhandle = mhandle;
	this->qhandle = qptr;
	this->outstanding_completion_count = 0;
	this->outstanding_wwi_count = 0;
	this->user_event_ptr->exs_evt_type = EXS_EVT_SEND;
	this->user_event_ptr->exs_evt_errno = 0;
	this->user_event_ptr->exs_evt_ahandle = ahandle;
	this->user_event_ptr->exs_evt_socket = sockno;

	this->user_event_ptr->exs_evt_union.exs_evt_xfer.exs_evt_buffer
							= (void *)buffer;
	this->user_event_ptr->exs_evt_union.exs_evt_xfer.exs_evt_length
							= length;
	this->user_event_ptr->exs_evt_union.exs_evt_xfer.exs_evt_mhandle
							= mhandle;

	/* This will reset qptr to NULL or set it to a meaningful value. */
	signo = -1;
	ret = this_conn->enqueue_send(this_conn, this, buffer, length,
						&qptr, &signo);

	if (signo >= 0 && qptr != NULL) {
		_evt_queue_do_notify(qptr, signo);
	}

err0:
	OUR_MEMBER_MUTEX_UNLOCK(this_conn, data_sequence_mutex);

	if (ret < 0) {
		return ret;
	}

	if(local_qhandle == NULL) {
		return 0;
	}
	else
	{	/* caller wants a blocking (synchronous) send */

		TRACE("%s %s: back from common_send, local_qhandle %p, "
			"ahandle %p, buffer %p, ret %ld\n", _time_prefix(),
			__func__, local_qhandle, ahandle, buffer, ret);

		if(ret == 0)
		{	/* send was started ok, now wait for it to complete */
			ret = _evt_queue_dequeue_vector(local_qhandle,
					&local_vector, 1, NULL);

			TRACE("%s %s: back from dequeue, "
				"local qhandle %p, ahandle %p, buffer %p, "
				"ret %ld, exs_evt_type %d, exs_evt_errno %d, "
				"exs_evt_ahandle %p, exs_evt_length %ld, "
				"exs_evt_buffer %p\n", _time_prefix(), __func__,
				local_qhandle, ahandle, buffer, ret,
				local_vector.exs_evt_type,
				local_vector.exs_evt_errno,
				local_vector.exs_evt_ahandle,
				local_vector.exs_evt_union.exs_evt_xfer
							.exs_evt_length,
				local_vector.exs_evt_union.exs_evt_xfer
							.exs_evt_buffer);

			if(ret != -1)
			{	/* no error on dequeue itself check completion*/
				if((errno = local_vector.exs_evt_errno) != 0)
					ret = -1;
				else
				{	/* no error on send, get bytes written*/
					ret = local_vector.exs_evt_union
						.exs_evt_xfer.exs_evt_length;
				}
			}
		}

		/* always return locally created queue to connection's cache */
		our_put_to_queue_cache(this_conn, local_qhandle);
	}

	/* return < 0 on error, >= 0 if started (or completed) ok */
	return ret;
}	/* exs_send */

static int
our_enqueue_send_seqpacket(struct exs_conn *this_conn, struct transaction_block *this,
		const void *buffer, size_t length, our_qhandle_t *qptr,
		int *signo)
{
	int ret;
	(void)buffer;
	(void)length;

	if (atomic_set_bits(SEND_READY_BIT, &this->send_state)
							& ADVERT_READY_BIT) {
		TRACE("send and advert ready on exs_send\n");
		STATISTICS_INCREMENT(matched_recv_await);

		ret = this_conn->xmit_if_match(this_conn, this,
					qptr, signo, __func__);
	} else {
		TRACE("send ready only on exs_send\n");
		STATISTICS_INCREMENT(unmatched_recv_await);
		ret = 0;
	}

	return ret;
}	/* our_enqueue_send_seqpacket */


/* when called, we know that:
 *	nbytes > 0
 *	length >= options->direct_minimum
 *	client_conn->send_offset == client_conn->take_offset
 *	client_conn->credit_count > 0
 *	client_conn->n_remote > 0
 *	stream_sync_mutex is locked
 *
 * returns ==0 if anything fails
 *	   > 0 number of bytes successfully sent directly
 */
static int
our_send_direct(struct exs_conn *cb, struct transaction_block *block)
{
	void			*source;
	size_t			length;
	int			offset;
	ssize_t			nbytes;
	struct ibv_mr		*local_mr;
	int			local_send_flags;

	source=block->user_event_ptr->exs_evt_union.exs_evt_xfer.exs_evt_buffer;
	length=block->user_event_ptr->exs_evt_union.exs_evt_xfer.exs_evt_length;
	offset = block->buf_offset;
	nbytes = length - offset;

	local_send_flags = IBV_SEND_SIGNALED;

	/* Must do this ahead of time */
	atomic_add_reset(1, UINT_MAX,
				&block->outstanding_completion_count);
	atomic_add_reset(1, UINT_MAX,
				&block->outstanding_wwi_count);

	TRACE_DBG(EXS_LOG_STREAM,
		"rdma_write %zd bytes, length %zu, local offset %u, remote offset %u, n_remote %u\n",
		nbytes, length, offset, cb->remote_buf_offset, cb->remote_buf_n_empty);
	local_send_flags = IBV_SEND_SIGNALED;
	if ((nbytes = our_send_rdma_write_notify(cb, block,
				source + offset, nbytes,
				block->mhandle, NULL,
				__func__,
				&local_send_flags, &local_mr)) < 0) {
		atomic_add_reset(-1, UINT_MAX,
				&block->outstanding_completion_count);
		atomic_add_reset(-1, UINT_MAX,
				&block->outstanding_wwi_count);
		return 0;
	}
	if (!(local_send_flags & IBV_SEND_SIGNALED)) {
		atomic_set_bits(SEND_DONE_BIT, &block->send_state);
		atomic_add_reset(-1, UINT_MAX,
				&block->outstanding_completion_count);
	}

	/* send directly from user source to remote buffer worked ok */
	TRACE("direct post_send %zd bytes\n", nbytes);

	cb->remote_buf_offset += nbytes;
	if (cb->remote_buf_offset >= cb->remote_recv_buf.ad_buf_len) {
		cb->remote_buf_offset = 0;
	}
	cb->remote_buf_n_empty -= nbytes;

	return nbytes;
}	/* our_send_direct */


static int
our_enqueue_send_stream(struct exs_conn *this_conn, struct transaction_block *this,
		const void *buffer, size_t length, our_qhandle_t *qptr,
		int *signo)
{
	int ret;
	(void)buffer;
	(void)length;
	ret = 0;
	atomic_set_bits(SEND_READY_BIT, &this->send_state);
	if (this_conn->next_advert_primed->advert_state & ADVERT_READY_BIT) {
		STATISTICS_INCREMENT(matched_recv_await);
		TRACE("%s %s send %p advert %p send and advert ready\n",
				_time_prefix(), __func__, this,
				this_conn->next_advert_primed);

		errno = 0;
		ret = this_conn->xmit_if_match(this_conn, this,
					qptr, signo, __func__);
	} else {
		TRACE("%s %s send %p advert %p send ready only\n",
				_time_prefix(), __func__, this,
				this_conn->next_advert_primed);
		STATISTICS_INCREMENT(unmatched_recv_await);
	}

	return ret;
}	/* our_enqueue_send_stream */

static int
our_enqueue_send_streamcopy(struct exs_conn *cb, struct transaction_block *this,
		const void *buffer, size_t length, our_qhandle_t *qptr,
		int *signo)
{
	(void)buffer;
	(void)length;
	atomic_set_bits(SEND_READY_BIT, &this->send_state);
	return cb->xmit_if_match(cb, this, qptr, signo, __func__);
}	/* our_enqueue_send_streamcopy */


/*
 *	exs_sendmsg - page 60 of Interconnect Software Consortium Issue 1.0
 *		asynchronously send a message on a socket
 *
 *	returns 0 if async op initiated ok,
 *		else -1 with errno set to error value
 */
EXS_EXPORT ssize_t
exs_sendmsg(int sockno, const exs_msghdr_t *message, int flags,
		exs_qhandle_t qhandle, exs_ahandle_t ahandle)
{
	ssize_t		ret;
	our_qhandle_t	qptr;
	struct exs_conn	*this_conn;

#if defined(ALLOW_PID_CHECKING)
	pid_t		current;

	current = getpid();
	TRACE("%s pid %d start %s, qhandle %p, flags 0x%x\n",
		_time_prefix(), current, __func__, qhandle, flags);

	if(global_owner != current)
#else
	if(global_owner == 0)
#endif
	{
		errno = EPERM;		/* Operation not permitted */
		return -1;
	}

	/* verify that this socket has been created */
	if((this_conn = our_lookup_socket(sockno)) == NULL)
	{
		errno = EBADF;		/* Bad file number */
		return -1;
	}

	if(this_conn->fd_flags & EXS_FD_BYPASS_PARAM_CHECKING)
		goto no_parameter_checks;

	/* verify that this socket has already been connected */
	if(this_conn->conn_state != EXS_CONN_RUNNING)
	{
		/* Transport endpoint is not connected */
		errno = ENOTCONN;
		return -1;
	}

	TRACE("%s %s: socket %d, this_conn %p, state %s(%d)\n",
		_time_prefix(), __func__, sockno, this_conn,
		state_labels[this_conn->conn_state], this_conn->conn_state);

	/* verify that caller is giving us a valid qhandle */
	if((qptr = _cnv_qnumb_to_ptr(qhandle)) == NULL)
	{	/* Invalid argument */
		return -1;
	}

	/* verify that flags are supported for this socket type or protocol */
	if(flags != 0)
	{
		errno = EOPNOTSUPP;	/* Operation not supported */
		return -1;
	}

no_parameter_checks:

	ret = -1;
	errno = ENOSYS;		/* Function not implemented */
	(void)message;
	(void)ahandle;

#if defined(ALLOW_FULL_TRACE)
	TRACE("%s %s: finished, returning %d\n",
		_time_prefix(), __func__, (int)ret);
#endif

	return ret;
}	/* exs_sendmsg */


/* A predicate function that returns true if and only if we can start an
 * exs_recv() operation.  For SOCK_SEQPACKET connections and SOCK_STREAM
 * connections with the EXS_FD_NODELAY flag, the connection must be still
 * alive.  For SOCK_STREAM connections without the EXS_FD_NODELAY flag, either
 * the connection must be still alive or there must be data remaining in the
 * stream recv buffer.
 *
 * Returns non-zero if we can do the exs_recv() and 0 if we cannot. */
int
our_can_exs_recv(struct exs_conn *this_conn)
{
	/* verify that this socket has already been connected */
	if(this_conn->conn_state != EXS_CONN_RUNNING
			|| !(this_conn->conn_flags & EXS_OK_TO_ISSUE_IO)) {
		if (!this_conn->is_seqpacket
				&& !(this_conn->fd_flags & EXS_FD_NODELAY)) {
			unsigned int nfull;
			OUR_MEMBER_MUTEX_LOCK(this_conn, stream_sync_mutex);
			nfull = this_conn->recv_buf.n_full;
			OUR_MEMBER_MUTEX_UNLOCK(this_conn, stream_sync_mutex);
			if (nfull > 0) {
				/* Can do exs_recv() on data we currently
				 * have */
				TRACE_DBG(EXS_LOG_STREAM,
						"still %u bytes in buffer", nfull);
				return 1;
			}
		}

		if (this_conn->conn_state == EXS_CONN_TERMINATING) {
			/* Connection reset by peer */
			errno = ECONNRESET;
		} else {
			/* Transport endpoint is not connected */
			errno = ENOTCONN;
		}
		return 0;
	}

	return 1;
}


/*
 *	exs_recv - page 42 of Interconnect Software Consortium Issue 1.0
 *		asynchronously receive a message from a connected socket
 *
 *	returns	== 0  if async op initiated ok,
 *		>= 0  if sync op finished ok (value is no. of bytes read),
 *		== -1 if error (errno set to error value)
 *
 *	only flags values accepted:
 *
 *		EXS_DONTWAIT		(nonstandard)
 *			without this flag, this function may block until a
 *				posted_recv_slot becomes available (happens when
 *				user has too many outstanding exs_recv calls)
 *			with this flag, this function will return EAGAIN error
 *				if no posted_recv_slot is available
 *			with or without this flag, if this function returns 0
 *				then the data has not been received until the
 *				completion event has been posted to qhandle.
 *		EXS_BLOCK		(nonstandard)
 *			force this recv to block until it completes.  The
 *			qhandle and ahandle parameters are ignored.
 *		EXS_UNSIGNALED		(nonstandard)
 *			don't cause a completion event when this send completes.
 *			The qhandle and ahandle parameters are ignored.
 *
 *	qhandle values accepted:
 *		NULL
 *			cause an error
 *		EXS_QHANDLE_INVALID
 *			left over from a previous error, cause a new error
 *		anything else
 *			assume it is a user-created queue
 *
 *	this may cause the completion event to be posted before returning
 */
EXS_EXPORT ssize_t
exs_recv(int sockno, void *buffer, size_t length, int flags,
		exs_qhandle_t qhandle, exs_ahandle_t ahandle,
		exs_mhandle_t mhandle)
{
	ssize_t				ret;
	our_qhandle_t			qptr;
	our_qhandle_t			local_qhandle;
	struct exs_conn			*this_conn;
	exs_event_t			local_vector;
	struct transaction_block	*this;
	struct ibv_mr			*use_this_mr;
	ssize_t				bytes_recved;
	int				need_ack;
	int				send_flags, signo;


#if defined(ALLOW_PID_CHECKING)
	pid_t				current;

	current = getpid();
	TRACE("%s pid %d start %s, qhandle %p, "
		"length %d, flags 0x%x\n",
		_time_prefix(), current, __func__, qhandle, (int)length, flags);

	if(global_owner != current)
#else
	if(global_owner == 0)
#endif
	{
		errno = EPERM;		/* Operation not permitted */
		return -1;
	}

	/* verify that this socket has been created */
	if((this_conn = our_lookup_socket(sockno)) == NULL)
	{
		errno = EBADF;		/* Bad file number */
		return -1;
	}

	local_qhandle = NULL;		/* assume this is not created locally */

	if(this_conn->fd_flags & EXS_FD_BYPASS_PARAM_CHECKING)
		goto no_parameter_checks;

	TRACE("%s %s: socket %d this_conn %p state %s(%d) "
		"qhandle %p flags %#08x length %d\n", _time_prefix(), __func__,
		sockno, this_conn,
		state_labels[this_conn->conn_state], this_conn->conn_state,
		qhandle, flags, (int)length);

	/* verify that caller is giving us either a valid qhandle or NULL */
	if(qhandle == EXS_QHANDLE_INVALID)
	{
		errno = EINVAL;		/* Invalid argument */
		return -1;
	}

	/* verify that flags are supported for this socket type or protocol */
	if((flags & ~(EXS_DONTWAIT|EXS_BLOCK|EXS_UNSIGNALED|MSG_WAITALL)) != 0)
	{
		errno = EOPNOTSUPP;	/* Operation not supported */
		return -1;
	}

	if ((flags & MSG_WAITALL) && this_conn->is_seqpacket) {
		/* This flag does not apply to a message-based socket, so we are
		 * allowed to ignore it. */
		flags &= ~MSG_WAITALL;
	}

	/* if user specified a NULL qhandle, then ensure that either EXS_BLOCK
	 * or EXS_UNSIGNALED flag was specified. */
	if (qhandle == NULL && !(flags & (EXS_BLOCK|EXS_UNSIGNALED))) {
		errno = EINVAL;		/* Invalid argument */
		return -1;
	}

	/* verify that length is at least 1 */
	if(length <= 0)
	{
		errno = EINVAL;		/* Invalid argument */
		return -1;
	}

	/* ensure that buffer is not NULL,
	 * TO-DO -- and is not an outstanding recycle buffer,
	 * since that leads to a race condition
	 */
	if(buffer == NULL)
	{
		errno = EINVAL;		/* Invalid argument */
		return -1;
	}

no_parameter_checks:
	if (!our_can_exs_recv(this_conn)) {
		return -1;
	}

	need_ack = 0;
	bytes_recved = 0;
	if ((flags & EXS_BLOCK) && !this_conn->is_seqpacket
			&& !(this_conn->fd_flags & EXS_FD_NODELAY)
			&& this_conn->next_recv_primed == this_conn->next_recv){
		/* For stream copy mode, we may be able to avoid creating a
		 * transaction block altogether. */
		OUR_MEMBER_MUTEX_LOCK(this_conn, stream_sync_mutex);
		ret = our_stream_copy_recv(this_conn, buffer, length,
				flags, __func__);
		if ((this_conn->recv_buf.n_empty
					> this_conn->recv_buf.n_empty_threshold)
				|| (this_conn->n_unacked_wwi
					> this_conn->n_unacked_wwi_threshold)) {
			need_ack = 1;
			bytes_recved = ret;
		}
		OUR_MEMBER_MUTEX_UNLOCK(this_conn, stream_sync_mutex);

		if (!need_ack && (ret > 0)) {
			return ret;
		}
	}

	/* see if user gave us an event queue or wants a blocking send.  After
	 * this, qptr will be set or this function will have returned an error
	 * condition. */
	if (flags & EXS_BLOCK) {
		/* get qhandle from cache, dynamically creating one if needed */
		local_qhandle = our_get_from_queue_cache(this_conn);
		if(local_qhandle == NULL)
		{	/* couldn't create it, errno already set */
			return -1;
		}
		qptr = local_qhandle;
	} else if (qhandle == NULL) {
		/* caller gave us no qhandle, so don't use one */
		qptr = NULL;
	} else if ((qptr = _cnv_qnumb_to_ptr(qhandle)) == NULL) {
		/* Invalid argument */
		errno = EINVAL;
		return -1;
	}

	if (flags & EXS_DONTWAIT) {
		OUR_MEMBER_MUTEX_TRYLOCK(this_conn, advert_sequence_mutex, &ret);
		if (ret != 0) {
			/* Returned errno is most likely EAGAIN, but catch other
			 * errors as well. */
			errno = ret;
			return -1;
		}
	} else {
		OUR_MEMBER_MUTEX_LOCK(this_conn, advert_sequence_mutex);
	}

	ret = -1;

	if ((flags & EXS_DONTWAIT)) {
		TRACE("sem_trywait next_recv %p sem %p\n", this_conn->next_recv,
				&this_conn->next_recv->sem);
		if (sem_trywait(&this_conn->next_recv->sem) < 0) {
			TRACE("%s %s: no free recv slots on this_conn %p\n",
				_time_prefix(), __func__, this_conn);

			/* errno should already set to EAGAIN, otherwise there's
			 * a serious error that we should probably pass along
			 * anyways */
			goto exit0;
		}
		TRACE("PAST next_recv %p sem %p\n", this_conn->next_recv,
				&this_conn->next_recv->sem);
	} else {
		OUR_MEMBER_SEM_WAIT(this_conn->next_recv, sem);
	}

	atomic_add_reset(1, UINT_MAX, &this_conn->pending_recv_count);

	if (!our_can_exs_recv(this_conn)) {
		/* this side has closed or other side has disconnected */
		OUR_MEMBER_SEM_POST(this_conn->next_recv, sem);

		errno = ENOTCONN;
		goto exit0;
	}

	if (!this_conn->next_recv->user_event_ptr) {
		/* Buffer for completion event is not available */
		OUR_MEMBER_SEM_POST(this_conn->next_recv, sem);

		errno = ENOBUFS;
		goto exit0;
	}

	this = this_conn->next_recv;

	/* set block's buffer_flags field with flags from the exs_recv */
	this->buffer_flags = flags;

	if(mhandle == EXS_MHANDLE_UNREGISTERED) {
		/* this data was not registered by the user */
		if ((this_conn->fd_flags & EXS_FD_SP_COPY_UNREG_ON_RECV)
				&& length <= this_conn->small_packet_max_size) {
			this->recv_flags = RECV_COPY_UNREG_BIT;
			this->recv_unreg_mr = NULL;
			use_this_mr = NULL;
		} else {
			/* dynamically register user's data */
			this->recv_unreg_mr = our_call_ibv_reg_mr(this_conn->pd,
						(void *)buffer, length,
						IBV_ACCESS_LOCAL_WRITE
						|IBV_ACCESS_REMOTE_WRITE,
						"user send buffer");
			if(this->recv_unreg_mr == NULL)
			{
				OUR_MEMBER_SEM_POST(this_conn->next_send, sem);
				ret = -1;
				goto exit1;
			}
			this->recv_flags = 0;
			use_this_mr = this->recv_unreg_mr;
		}
	} else {
		/* this data was already registered by the user */
		this->recv_flags = 0;
		this->recv_unreg_mr = NULL;
		TRACE("%s %s: mhandle %p is registered\n",
			_time_prefix(), __func__, mhandle);
		use_this_mr = our_ensure_buffer_registered(this_conn,
						mhandle, __func__);
		if (use_this_mr == NULL) {
			/* should be able to register user buffer
			 * but couldn't */
			OUR_MEMBER_SEM_POST(this_conn->next_send, sem);
			errno = EINVAL;
			ret = -1;
			goto exit1;
		}
	}

	/* fill in transaction's fields with parameters from the exs_recv */
	this->in_progress = 1;
	this->qhandle = qptr;
	this->recv_byte_len = bytes_recved;
	this->user_event_ptr->exs_evt_type = EXS_EVT_RECV;
	this->user_event_ptr->exs_evt_errno = 0;
	this->user_event_ptr->exs_evt_ahandle = ahandle;
	this->user_event_ptr->exs_evt_socket = sockno;

	this->user_event_ptr->exs_evt_union.exs_evt_xfer.exs_evt_buffer
							= (void *)buffer;
	this->user_event_ptr->exs_evt_union.exs_evt_xfer.exs_evt_length
							= length;
	this->user_event_ptr->exs_evt_union.exs_evt_xfer.exs_evt_mhandle
							= mhandle;
	this->user_event_ptr->
			exs_evt_union.exs_evt_xfer.exs_evt_buffer_length = 0;
	this->user_event_ptr->
			exs_evt_union.exs_evt_xfer.__exs_recv_transaction=NULL;

	this_conn->next_recv = this->next;
	if (this_conn->is_seqpacket || (this_conn->fd_flags & EXS_FD_NODELAY)
			|| (bytes_recved == 0 && this_conn->recv_buf.n_full == 0
						&& (flags & MSG_WAITALL))) {
		/* RECV_READY_BIT used by exs_stream_recv_loop to determine if
		 * it should put data from the recv buffer here.  We don't want
		 * to let this happen until we have had a chance to send the
		 * ADVERT.  Otherwise, data might get placed there and then an
		 * ACK might be sent from this block, leading to a possible
		 * corrupted message sent across the wire. */
		OUR_MEMBER_MUTEX_LOCK(this_conn, stream_sync_mutex);
		ret = our_send_advert(this_conn, this, buffer, length, use_this_mr,
								this->index);
		OUR_MEMBER_MUTEX_UNLOCK(this_conn, stream_sync_mutex);
		atomic_set_bits(RECV_READY_BIT, &this->send_state);
	} else {
		if (need_ack) {
			OUR_MEMBER_MUTEX_LOCK(this_conn, stream_sync_mutex);
			/* We only got this block to send an ACK, so advance
			 * next_recv_primed since this is not a real data recv
			 * block */
			this_conn->next_recv_primed = this->next;
			ret = 0;
			send_flags = IBV_SEND_SIGNALED;
			if (our_send_ack(this_conn, this, &send_flags, __func__) < 0) {
				if (errno != ENOTCONN) {
					ret = -1;
					errno = EIO;
				}
			} else if (send_flags & IBV_SEND_SIGNALED) {
				/* We've recorded the bytes received into the
				 * transaction block, but now we have to wait
				 * for a send ACK completion.  Reset this so
				 * that the completion frees the block. */
				bytes_recved = 0;
			} else {
				our_exs_recv_done(this_conn, this, &qptr, 0);
			}
			OUR_MEMBER_MUTEX_UNLOCK(this_conn, stream_sync_mutex);
		} else {
			atomic_set_bits(RECV_READY_BIT, &this->send_state);
			this_conn->source_seqno += length;
			signo = 0;
			ret = our_stream_recv_loop(this_conn, &qptr, &signo, __func__);
		}
	}

	if (ret < 0) {
		goto exit0;
	}

	TRACE("%s %s: finished ok, ret %zd\n",
		_time_prefix(), __func__, ret);
	ret = 0;
	goto exit0;

exit1:
	this_conn->next_recv = this->next;
exit0:
	OUR_MEMBER_MUTEX_UNLOCK(this_conn, advert_sequence_mutex);

	if (ret < 0) {
		return -1;
	}

	if(local_qhandle == NULL)
		return 0;

	/* caller wants a blocking (synchronous) recv */

	if (bytes_recved > 0) {
		/* but we already got data for them */
		return bytes_recved;
	}

	TRACE("%s %s: blocking recv, local_qhandle %p, "
		"ahandle %p, buffer %p\n", _time_prefix(),
		__func__, local_qhandle, ahandle, buffer);

	/* recv was started ok, now wait for it to complete */
	ret = _evt_queue_dequeue_vector(local_qhandle, &local_vector, 1, NULL);

	TRACE("%s %s: back from dequeue, local_qhandle %p, "
		"ahandle %p, buffer %p, ret %ld, "
		"exs_evt_type %d, exs_evt_errno %d, "
		"exs_evt_ahandle %p, exs_evt_length %ld, "
		"exs_evt_buffer %p\n", _time_prefix(), __func__,
		local_qhandle, ahandle, buffer, ret,
		local_vector.exs_evt_type,
		local_vector.exs_evt_errno,
		local_vector.exs_evt_ahandle,
		local_vector.exs_evt_union.exs_evt_xfer
					.exs_evt_length,
		local_vector.exs_evt_union.exs_evt_xfer
					.exs_evt_buffer);

	if(ret != -1)
	{	/* no error on dequeue itself check completion */
		if((errno = local_vector.exs_evt_errno) != 0)
			ret = -1;
		else
		{	/* no error on recv, get bytes read */
			ret = local_vector.exs_evt_union
				.exs_evt_xfer.exs_evt_length;
		}
	}

	/* always return the locally created queue to the connection's cache */
	our_put_to_queue_cache(this_conn, local_qhandle);

	/* return < 0 on error, >= 0 if completed ok */
	return ret;
}	/* exs_recv */


/*
 *	exs_recvmsg - page 46 of Interconnect Software Consortium Issue 1.0
 *		asynchronously receive a message from a socket
 *
 *	returns 0 if async op initiated ok,
 *		else -1 with errno set to error value
 */
EXS_EXPORT ssize_t
exs_recvmsg(int sockno, exs_msghdr_t *message, int flags,
		exs_qhandle_t qhandle, exs_ahandle_t ahandle)
{
	ssize_t		ret;
	our_qhandle_t	qptr;
	struct exs_conn	*this_conn;

#if defined(ALLOW_PID_CHECKING)
	pid_t		current;

	current = getpid();
	TRACE("%s pid %d start %s, qhandle %p, "
		"flags 0x%x\n",
		_time_prefix(), current, __func__, qhandle, flags);

	if(global_owner != current)
#else
	if(global_owner == 0)
#endif
	{
		errno = EPERM;		/* Operation not permitted */
		return -1;
	}

	/* verify that this socket has been created */
	if((this_conn = our_lookup_socket(sockno)) == NULL)
	{
		errno = EBADF;		/* Bad file number */
		return -1;
	}

	if(this_conn->fd_flags & EXS_FD_BYPASS_PARAM_CHECKING)
		goto no_parameter_checks;

	/* verify that this socket has already been connected */
	if(this_conn->conn_state != EXS_CONN_RUNNING)
	{
		if (this_conn->conn_state == EXS_CONN_TERMINATING)
		{	/* Connection reset by peer */
			errno = ECONNRESET;
		}
		else
		{	/* Transport endpoint is not connected */
			errno = ENOTCONN;
		}
		return -1;
	}

	TRACE("%s %s: socket %d, this_conn %p, state %s(%d)\n",
		_time_prefix(), __func__, sockno, this_conn,
		state_labels[this_conn->conn_state], this_conn->conn_state);

	/* verify that caller is giving us a valid qhandle */
	if((qptr = _cnv_qnumb_to_ptr(qhandle)) == NULL)
	{	/* Invalid argument */
		return -1;
	}

	/* verify that flags are supported for this socket type or protocol */
	if(flags != 0)
	{
		errno = EOPNOTSUPP;	/* Operation not supported */
		return -1;
	}

no_parameter_checks:

	/*****
	ret = our_try_to_post_recv_block(cb, block, __func__);
	*****/

	ret = -1;
	errno = ENOSYS;		/* Function not implemented */
	(void)message;
	(void)ahandle;

#if defined(ALLOW_FULL_TRACE)
	TRACE("%s %s: finished, returning %d\n",
		_time_prefix(), __func__, (int)ret);
#endif

	return ret;
}	/* exs_recvmsg */


/*
 *	exs_blocking_recv() - not in Interconnect Software Consortium Issue 1.0
 *			      was added by us
 *
 *	returns >=0 as amount read ok
 *		else -1 with errno set to error value
 */
EXS_EXPORT ssize_t
exs_blocking_recv(int sockno, void *buffer, size_t max_bytes, int flags,
			exs_mhandle_t mhandle)
{

	return exs_recv(sockno, buffer, max_bytes, flags | EXS_BLOCK, NULL,
							NULL, mhandle);
}	/* exs_blocking_recv */


/*
 *	exs_read() - not in Interconnect Software Consortium Issue 1.0
 *			      was added by us
 *
 *	returns >=0 as amount read ok
 *		else -1 with errno set to error value
 */
EXS_EXPORT ssize_t
exs_read(int sockno, void *buffer, size_t max_bytes)
{
	return exs_recv(sockno, buffer, max_bytes, EXS_BLOCK, NULL,
					NULL, EXS_MHANDLE_UNREGISTERED);
}	/* exs_read */


/*
 *	exs_blocking_send() - not in Interconnect Software Consortium Issue 1.0
 *			      was added by us
 *
 *	returns >=0 as amount written ok
 *		else -1 with errno set to error value
 */
EXS_EXPORT ssize_t
exs_blocking_send(int sockno, void *buffer, size_t max_bytes, int flags,
			exs_mhandle_t mhandle)
{
	return exs_send(sockno, buffer, max_bytes, flags | EXS_BLOCK, NULL,
							NULL, mhandle);
}	/* exs_blocking_send */


/*
 *	exs_write() - not in Interconnect Software Consortium Issue 1.0
 *			      was added by us
 *
 *	returns >=0 as amount written ok
 *		else -1 with errno set to error value
 */
EXS_EXPORT ssize_t
exs_write(int sockno, void *buffer, size_t max_bytes)
{
	return exs_send(sockno, buffer, max_bytes, EXS_BLOCK, NULL,
					NULL, EXS_MHANDLE_UNREGISTERED);
}	/* exs_write */


/*
 *	exs_sendfile - page 55 of Interconnect Software Consortium Issue 1.0
 *		initiate transmission of the contents of a file from a socket
 *
 *	returns 0 if async op initiated ok,
 *		else -1 with errno set to error value
 */
EXS_EXPORT ssize_t
exs_sendfile(int sockno, exs_xferfile_t *sendvec, int sendvec_cnt,
		int flags, exs_qhandle_t qhandle, exs_ahandle_t ahandle)
{
	int				i;
	struct exs_conn			*this_conn;
	char				errbuf[_MAX_STRERROR_MALLOC];

#if defined(ALLOW_PID_CHECKING)
	pid_t				current;

	current = getpid();
	TRACE("%s pid %d start %s, qhandle %p, "
		"sendvec_cnt %d, flags 0x%x\n",
		_time_prefix(), current, __func__, qhandle, sendvec_cnt, flags);

	if(global_owner != current)
#else
	if(global_owner == 0)
#endif
	{
		errno = EPERM;		/* Operation not permitted */
		return -1;
	}

	/* verify that this socket has been created */
	if((this_conn = our_lookup_socket(sockno)) == NULL)
	{
		errno = EBADF;		/* Bad file number */
		return -1;
	}

	if(this_conn->fd_flags & EXS_FD_BYPASS_PARAM_CHECKING)
		goto no_parameter_checks;

	/* verify that this socket has already been connected */
	if(this_conn->conn_state != EXS_CONN_RUNNING)
	{
		/* Transport endpoint is not connected */
		errno = ENOTCONN;
		goto err0;
	}

	TRACE("%s %s: socket %d, this_conn %p, state %s(%d)\n",
		_time_prefix(), __func__, sockno, this_conn,
		state_labels[this_conn->conn_state], this_conn->conn_state);

	/* verify that flags are supported for this socket type or protocol */
	if(flags != 0)
	{
		errno = EOPNOTSUPP;	/* Operation not supported */
		goto err0;
	}

	/* verify that sendvec_cnt does not exceed the dynamic number of
	 * slots currently available
	 */
	if(sendvec_cnt <= 0)
	{
		errno = EINVAL;		/* Invalid argument */
		goto err0;
	}

no_parameter_checks:

	errno = ENOSYS;
	goto err0;

err0:
	i = errno;
	if (strerror_r(i, errbuf, _MAX_STRERROR_MALLOC)) {
		snprintf(errbuf, _MAX_STRERROR_MALLOC, "unknown");
	}
	TRACE("%s %s: finished, this_conn %p, state %s(%d), "
		"error %d: %s\n", _time_prefix(), __func__, this_conn,
		state_labels[this_conn->conn_state], this_conn->conn_state,
		i, errbuf);
	errno = i;
	return -1;
}	/* exs_sendfile */


static void
our_flush_stranded_send_blocks(struct exs_conn *this_conn, int errcode)
{
	struct transaction_block *first, *block;
	our_qhandle_t		qhandle;
	int			signo;

	signo = -1;
	block = first = this_conn->next_send;
	do {
#ifdef ALLOW_FULL_TRACE
		int sval;
		sem_getvalue(&block->sem, &sval);
		TRACE("%s block %p in_progress %d sem_value %d\n",
			__func__, block, block->in_progress, sval);
#endif

		qhandle = NULL;
		if (! block->in_progress) {
			continue;
		}

		/* report back to user "Software caused connection abort" */
		block->user_event_ptr->exs_evt_errno = errcode;

		/* report back to user that no bytes were transfered */
		block->user_event_ptr->exs_evt_union.exs_evt_xfer.exs_evt_length
								= 0;

		signo = our_all_done_with_block(this_conn, block,
				&qhandle, 0, __func__);

		if (signo >= 0 && qhandle != NULL) {
			_evt_queue_do_notify(qhandle, signo);
		}
	} while ((block = block->next) != first);
}	/* our_flush_stranded_send_blocks */


static void
our_flush_all_stranded_recv_blocks(struct exs_conn *this_conn)
{
	struct transaction_block *last, *block;
	our_qhandle_t		qhandle;
	int	signo;

	TRACE("%s [%lu] %s: conn %p, conn_state %s(%d), conn_flags %0x\n",
			_time_prefix(), pthread_self(), __func__, this_conn,
			state_labels[this_conn->conn_state],
			this_conn->conn_state, this_conn->conn_flags);

	block = this_conn->next_recv_primed;
	last = this_conn->next_recv;
	if (block->in_progress) {
		do {
#ifdef ALLOW_FULL_TRACE
			int sval;
			sem_getvalue(&block->sem, &sval);
			TRACE("%s %s block %p in_progress %d sem_value %d\n",
					_time_prefix(), __func__, block,
					block->in_progress,
					sval);
#endif

			if (!this_conn->is_seqpacket
					&& !(this_conn->fd_flags&EXS_FD_NODELAY)
					&& this_conn->recv_buf.n_full > 0) {
				our_stream_recv_loop(this_conn,
						&qhandle, &signo, __func__);
			} else {
				/* report back to user "Software caused
				 * connection abort" */
				block->user_event_ptr->exs_evt_errno
					= ECONNABORTED;

				signo = our_exs_recv_done(this_conn,
						block, &qhandle, 0);
			}

			if (signo >= 0 && qhandle != NULL) {
				_evt_queue_do_notify(qhandle, signo);
			}

		} while ((block = block->next) != last);
	}
}	/* our_flush_all_stranded_recv_blocks */


/*
 *	exs_poll - page 27 of Interconnect Software Consortium Issue 1.0
 *		modify the set of monitored conditions for multiple sockets
 *		to trigger asynchronous notification events on a specified event
 *
 *	returns nfds if all ok,
 *	else number of entries processed ok with errno set to error value
 */
EXS_EXPORT nfds_t
exs_poll(exs_pollfd_t fds[], nfds_t nfds, int flags, exs_qhandle_t qhandle)
{
	our_qhandle_t		qptr;

#if defined(ALLOW_PID_CHECKING)
	pid_t			current;

	current = getpid();
	TRACE("%s pid %d start %s, qhandle %p, "
		"nfds %lu, flags 0x%x\n",
		_time_prefix(), current, __func__, qhandle, nfds, flags);

	if(global_owner != current)
#else
	if(global_owner == 0)
#endif
	{
		errno = EPERM;		/* Operation not permitted */
		return -1;
	}

	if(flags != 0)
	{
		errno = ENOTSUP;	/* Not supported */
		return -1;
	}

	/* verify that caller is giving us a valid qhandle */
	if((qptr = _cnv_qnumb_to_ptr(qhandle)) == NULL)
	{	/* Invalid argument */
		return -1;
	}

	/* all user-supplied parameters checked ok */

	(void)fds;
	(void)nfds;
	errno = ENOSYS;
	return -1;
}	/* exs_poll */


static void
our_free_conn_requested_list(struct our_list_head *this_list)
{
	struct conn_requested	*requested;

	while ((requested = exs_p_list_del_head(this_list))) {
		EXS_FREE(requested);
	}
	exs_p_list_destroy(this_list);
}	/* our_free_conn_requested_list */


/* called only from our_unalloc_socket()
 *
 * this is the counterpart of our_init_pending()
 *
 * these lists should be non-empty only for listening_sockets
 * that did not do everything to completion
 */
static void
our_free_all_pending_lists(struct exs_conn *this_conn, const char *message)
{
	/* free list that holds unused "struct pending_accept"
	 * for this listening socket
	 */
	our_free_new_accept_list(&this_conn->pending_accept_list, message);

	/* init 2 lists to track connect_request events (handled in
	 * our_con_man_thread()) with no pending accepts to match
	 */
	our_free_conn_requested_list(&this_conn->conn_requested_list);
	our_free_conn_requested_list(&this_conn->conn_requested_free_list);
}	/* our_free_pending */


/*
 * called from	exs_socket() at levels 1, 15, 2, 3 and 4 holds no locks
 *		our_con_man_thread() at level 4 for listener holds no locks
 *		2 times from our_start_close() at level 7 for client or listener
 *								holds both locks
 *
 * when called, this_conn must be a client or listener, NOT a server agent
 *		because a server agent's socket was not created by exs_socket()
 *
 * peels back the allocations and creations performed by exs_socket() at
 * the specified level
 *
 * when called, this_conn->event_mutex MAY be locked, MAY NOT be!
 *	for levels 6 and 7 it is locked
 *	for levels 1, 15, 2, 3, 4, and 5 it is not locked
 */
static void
our_undo_exs_socket(int fd, struct exs_conn *this_conn, int level,
			const char *message, our_qhandle_t *qhandle,
			int *signo)
{
	(void)fd;
	TRACE("%s %s: enter, fd %d this_conn %p, level %d, "
		"cm_id %p, event_thread %ld\n", message, __func__, fd,
		this_conn, level, this_conn->cm_id, this_conn->event_thread);
	*signo = -1;
	switch (level) {
	case 7:	/* only called from our_start_close() when exs_socket() on this
		 * conn has completed ok and
		 * either it has only progressed through bind before the close
		 * or it is a listener, not a client or an agent.
		 * when called, caller has locked event_mutex
		 */
		if (this_conn->event_thread)
			our_terminate_con_man_thread(this_conn, message);
	case 6:
		/* unlock this_conn's event mutex before we destroy it */
		TRACE("\t\t%4d:   UNLOCK cb %p event_mutex %p\n",
			__LINE__, this_conn, &this_conn->event_mutex);
		pthread_mutex_unlock(&this_conn->event_mutex);

	case 5:
		/* make sure user can't find this listener again */
		our_rem_socket_from_list(this_conn);

	case 4:

	case 3:
		/* there are now no locks held on this_conn,
		 * and it is not in the open_socket_table/conn_list
		 */
		rdma_destroy_id(this_conn->cm_id);
		TRACE("%s %s: FREEd cm_id %p on socket %d\n",
			_time_prefix(), message, this_conn->cm_id, fd);
	case 2:
		rdma_destroy_event_channel(this_conn->event_channel);
		TRACE("%s %s: FREEd event_channel %p on socket %d\n",
			_time_prefix(), message, this_conn->event_channel, fd);
	case 15:
		/* destroy the stuff set up in our_init_completion_synch() */
		our_destroy_completion_synch(this_conn);
		TRACE("%s %s: destroyed our_init_completion_synch stuff"
			" on conn %p, pending->qhandle %p, qhandle %p\n",
			_time_prefix(), message, this_conn,
			this_conn->conn_pending->qhandle, qhandle);

		/* post the "final event" if one is set up to be posted */
		if(this_conn->conn_pending->qhandle != NULL)
		{
			_post_event_to_full_evt_queue(this_conn->conn_pending,
								qhandle, signo);
		}

	case 1:
		if(this_conn->nego_out_ptr != NULL)
		{	/* need to free nego block */
			TRACE("%s %s: free nego out transaction block "
			"%p\n",_time_prefix(),message,this_conn->nego_out_ptr);
			our_free_transaction_block(this_conn->nego_out_ptr);
		}


		/* destroy the stuff set up in our_alloc_socket() */
		our_unalloc_socket(this_conn, message);
		TRACE("%s %s: FREEd conn %p on socket %d\n",
			_time_prefix(), message, this_conn, fd);
	case 0:
	default:
		break;
	}	/* switch */
}	/* our_undo_exs_socket */


/*
 *	exs_socket() - not in Interconnect Software Consortium Issue 1.0
 *		       was added by us
 *
 *	returns >=0 fd of newly opened socket
 *			whose conn_state is EXS_CONN_UNBOUND
 *		else -1 with errno set to error value
 *
 *	what it does:
 *		checks for acceptable parameters
 *		creates conventional socket to give us an fd
 *		allocates and initializes new struct exs_conn for this socket
 *		creates new rdma event channel for this socket
 *		creates new "connection manager" (cm) id for this event channel
 *		add new struct exs_conn to master global list "conn_list"
 *			of all known connections of any type in any state
 *		creates instance of our_con_man_thread to deal with
 *			asynch cm events on this struct exs_conn
 */
EXS_EXPORT int
exs_socket(int family, int socktype, int protocol)
{
	int			fd = -1, ret, level = 0, signo = -1;
	struct exs_conn		*new_conn = NULL;
	char			errbuf[_MAX_STRERROR_MALLOC];

#if defined(ALLOW_PID_CHECKING)
	pid_t		current;

	current = getpid();
	TRACE("%s pid %d start %s, family %d, socktype %d, "
		"protocol %d\n", _time_prefix(), current, __func__,
		family, socktype, protocol);

	if(global_owner != current)
#else
	if(global_owner == 0)
#endif
	{
		ret = EPERM;		/* Operation not permitted */
		goto err0;
	}

	if (family != AF_INET && family != AF_INET6) {
		ret = EAFNOSUPPORT;
		goto err0;
	}

	if (socktype != SOCK_SEQPACKET && socktype != SOCK_STREAM) {
		ret = ESOCKTNOSUPPORT;
		goto err0;
	}

	if (protocol != 0 && protocol != IPPROTO_TCP) {
		ret = EPROTONOSUPPORT;
		goto err0;
	}

	/* allocates a new struct exs_conn, clears most fields to 0 or NULL,
	 * and initializes its state field to EXS_CONN_UNOPENED
	 */
	if((new_conn = our_alloc_socket(__func__)) == NULL)
	{
		ret = errno;
		goto err0;
	}

	/* set up the defaults for client to use in connection negotiation
	 * listeners don't use these directly, but every
	 * server agent inherits these defaults from its listener parent
	 */
	our_set_send_credit_limits(new_conn, DEFAULT_FLOW_CONTROL_CREDITS);
	new_conn->small_packet_max_size = SMALL_PACKET_MAX_SIZE;

	/* initialize completion condition variables */
	if((ret = our_init_completion_synch(new_conn, __func__)))
	{
		level = 1;
		goto err;
	}

	new_conn->event_channel = rdma_create_event_channel();
	if(new_conn->event_channel == NULL)
	{
		ret = errno;
		OUR_PERROR("rdma_create_event_channel failed", ret);
		level = 15;
		goto err;
	}
	/* save the user's fd in this newly allocated structure */
	new_conn->sockno = fd = new_conn->event_channel->fd;

	/* redundant, but makes it easier to post exs_event_t events */
	new_conn->conn_pending->sockno = fd;

	TRACE("%s %s: ALLOCATE %lu bytes for event_channel %p "
		"on socket %d\n", _time_prefix(), __func__,
		(long unsigned int)sizeof(*new_conn->event_channel),
		new_conn->event_channel, fd);

	/* Allocate a "communication identifier" pointer, which is the rdma
	 * equivalent of a socket fd.
	 *	This returns the new pointer in "cm_id".
	 *	Events associated with this "cm_id" will be reported on
	 *		the previously (above) created "event_channel".
	 *	Our unique tag associated with this "cm_id" will be "new_conn".
	 *	The "port space" to be used with this "cm_id" is TCP because
	 *		we want to use it for TCP connections.
	 *			Port space possibilities are:
	 *				RDMA_PS_TCP	TCP
	 *				RDMA_PS_UDP	UDP
	 */
	ret = rdma_create_id(new_conn->event_channel, &new_conn->cm_id,
						new_conn, RDMA_PS_TCP);
	if(ret)
	{
		OUR_PERROR("rdma_create_id failed", errno);
		level = 2;
		goto err;
	}
	TRACE("%s %s: ALLOCATE %lu bytes for cm_id %p "
		"on socket %d\n", _time_prefix(), __func__,
		(long unsigned int)sizeof(*new_conn->cm_id), new_conn->cm_id,
		fd);

	new_conn->conn_state = EXS_CONN_UNBOUND;

	if(socktype == SOCK_SEQPACKET)
	{	/* for SOCK_SEQPACKET connections, message boundaries must be
	`	 * preserved.  Therefore, no nagling on sending side, and on
		 * receiving side short receives truncate long sends
		 */
		new_conn->fd_flags = EXS_FD_SP_COPY_UNREG|EXS_FD_NODELAY;
		new_conn->is_seqpacket = 0xff;

		new_conn->xmit_if_match = &our_xmit_if_match_seqpacket;
		new_conn->enqueue_send = &our_enqueue_send_seqpacket;
		new_conn->handle_recv_advert=&our_handle_recv_advert_seqpacket;
		new_conn->handle_send_write=&our_handle_send_write_seqpacket;
		new_conn->handle_recv_write_notify
			= &our_handle_recv_write_notify_seqpacket;
	} else {
		new_conn->fd_flags = 0;

		new_conn->xmit_if_match = &our_xmit_if_match_streamcopy;
		new_conn->enqueue_send = &our_enqueue_send_streamcopy;
		new_conn->handle_recv_advert = &our_handle_recv_advert_streamcopy;
		new_conn->handle_send_write=&our_handle_send_write_streamcopy;
		new_conn->handle_recv_write_notify
			= &our_handle_recv_write_notify_streamcopy;
	}

	/* a default value for error handling in con_man_thread() */
	new_conn->event_expected = RDMA_CM_EVENT_MULTICAST_ERROR;

	new_conn->comp_thread_cpu = INT_MAX;

	/* create our_con_man_thread to pick up asynch cm events */
	ret = create_our_con_man_thread(new_conn, __func__);
	if(ret)
	{
		level = 3;
		goto err;
	}

	/* socket is completely set up, make it available to users */
	our_add_socket_to_list(new_conn);

#if !defined(ALLOW_EXS_TRACE)
	return fd;
#endif
	/* else fall thru to err0 for possible tracing */

err0:
	if (strerror_r(ret, errbuf, _MAX_STRERROR_MALLOC)) {
		snprintf(errbuf, _MAX_STRERROR_MALLOC, "unknown");
	}
	TRACE("%s %s: finished, fd %d, new_conn %p, state %s(%d), "
		"ret %d: %s\n", _time_prefix(), __func__, fd, new_conn,
		state_labels[new_conn->conn_state], new_conn->conn_state,
		ret, errbuf);

	if (ret)
		errno = ret;
	return fd;

err:
	our_undo_exs_socket(fd, new_conn, level, __func__, NULL, &signo);
	fd = -1;
	new_conn = NULL;
	goto err0;
}	/* exs_socket */


/*
 *	exs_bind() - not in Interconnect Software Consortium Issue 1.0
 *		       was added by us
 *
 *	returns 0 on success
 *		else -1 with errno set to error value
 */
EXS_EXPORT int
exs_bind(int fd, struct sockaddr *address, socklen_t addrlen)
{
	int		ret;
	struct exs_conn	*this_conn;

#if defined(ALLOW_PID_CHECKING)
	pid_t		current;

	current = getpid();
	TRACE("%s pid %d start %s, fd %d\n",
		_time_prefix(), current, __func__, fd);

	if(global_owner != current)
#else
	if(global_owner == 0)
#endif
	{
		errno = EPERM;		/* Operation not permitted */
		return -1;
	}

	/* lookup fd in table */
	if((this_conn = our_lookup_socket(fd)) == NULL)
	{
		errno = EBADF;		/* Bad file number */
		return -1;
	}

	TRACE("%s %s: socket %d, this_conn %p, state %s(%d)\n",
		_time_prefix(), __func__, fd, this_conn,
		state_labels[this_conn->conn_state], this_conn->conn_state);

	/* verify that this socket has not already been bound */
	if(this_conn->conn_state != EXS_CONN_UNBOUND)
	{
		errno = EINVAL;   /* Invalid argument */
		ret = -1;
	}

	/* We lock event_mutex to protect conn_state.
	 * At first glance this seems excessive, but at this time
	 * nothing else should be happening on this listener or client socket,
	 * so there should be no contention for this lock --
	 * we lock it to explicitly prevent anything else from happening!
	 */

	TRACE("\t\t%4d:     LOCK cb %p event_mutex %p\n",
		__LINE__, this_conn, &this_conn->event_mutex);
	pthread_mutex_lock(&this_conn->event_mutex);
	TRACE("\t\t%4d:     PAST cb %p event_mutex %p\n",
		__LINE__, this_conn, &this_conn->event_mutex);

	/* check again that this socket has not already been bound */
	if(this_conn->conn_state != EXS_CONN_UNBOUND)
	{
		errno = EINVAL;   /* Invalid argument */
		ret = -1;
	}
	else
	{
		(void)addrlen; /* not required by rdma_bind_addr */
		ret = rdma_bind_addr(this_conn->cm_id, address);
		if(ret)
		{
			ret = errno;
			OUR_PERROR("rdma_bind_addr failed", ret);
			errno = ret;
			ret = -1;
		}
		else
		{	/* the bind worked ok, advance this socket's state */
			this_conn->conn_state = EXS_CONN_BOUND;
		}
	}

	TRACE("\t\t%4d:   UNLOCK cb %p event_mutex %p\n",
		__LINE__, this_conn, &this_conn->event_mutex);
	pthread_mutex_unlock(&this_conn->event_mutex);

#if defined(ALLOW_FULL_TRACE)
	TRACE("%s %s: finished, this_conn %p, state %s(%d), ret %d\n",
		_time_prefix(), __func__, this_conn,
		state_labels[this_conn->conn_state], this_conn->conn_state,
		ret);
#endif

	return ret;
}	/* exs_bind */


/*
 *	exs_listen() - not in Interconnect Software Consortium Issue 1.0
 *		       was added by us
 *
 *	returns 0 on success
 *		else -1 with errno set to error value
 */
EXS_EXPORT int
exs_listen(int fd, int backlog)
{
	int		ret;
	struct exs_conn	*this_conn;

#if defined(ALLOW_PID_CHECKING)
	pid_t		current;

	current = getpid();
	TRACE("%s pid %d start %s, fd %d, backlog %d\n",
		_time_prefix(), current, __func__, fd, backlog);

	if(global_owner != current)
#else
	if(global_owner == 0)
#endif
	{
		errno = EPERM;		/* Operation not permitted */
		return -1;
	}

	/* lookup fd in table and
	 * verify that this socket has been created and optionally
	 * has been bound to the listening address
	 */
	if((this_conn = our_lookup_socket(fd)) == NULL)
	{
		errno = EBADF;		/* Bad file number */
		return -1;
	}

	TRACE("%s %s: socket %d, this_conn %p, state %s(%d)\n",
		_time_prefix(), __func__, fd, this_conn,
		state_labels[this_conn->conn_state], this_conn->conn_state);

	/* verify that this socket has not already been connected or listened */
	if(this_conn->conn_state != EXS_CONN_UNBOUND
				&& this_conn->conn_state != EXS_CONN_BOUND)
	{
		errno = EINVAL;   /* Invalid argument */
		return -1;
	}

	/* verify that backlog is not negative */
	if(backlog < 0) {
		errno = EINVAL;	  /* Invalid argument */
		return -1;
	}

	/* We lock event_mutex to protect conn_state.
	 * At first glance this seems excessive, but at this time
	 * nothing else should be happening on this listener socket,
	 * so there should be no contention for this lock --
	 * we lock it to explicitly prevent anything else from happening!
	 */

	TRACE("\t\t%4d:     LOCK cb %p event_mutex %p\n",
		__LINE__, this_conn, &this_conn->event_mutex);
	pthread_mutex_lock(&this_conn->event_mutex);
	TRACE("\t\t%4d:     PAST cb %p event_mutex %p\n",
		__LINE__, this_conn, &this_conn->event_mutex);

	/* check again that this socket has not gone beyond bound */
	if(this_conn->conn_state != EXS_CONN_UNBOUND
				&& this_conn->conn_state != EXS_CONN_BOUND)
	{
		errno = EINVAL;   /* Invalid argument */
		ret = -1;
	}
	else
	{
		/* the pending and conn_requested lists were already inited
		 * when this_conn (and every struct exs_conn) was allocated.
		 */

		ret = rdma_listen(this_conn->cm_id, backlog);
		if(ret)
		{
			ret = errno;
			OUR_PERROR("rdma_listen failed", ret);
			errno = ret;
			ret = -1;
		}
		else
		{	/* the listen worked ok, advance the conn_state */
			this_conn->conn_state = EXS_CONN_LISTENING;
			this_conn->backlog_limit = backlog;
		}
	}

	TRACE("\t\t%4d:   UNLOCK cb %p event_mutex %p\n",
		__LINE__, this_conn, &this_conn->event_mutex);
	pthread_mutex_unlock(&this_conn->event_mutex);

#if defined(ALLOW_FULL_TRACE)
	TRACE("%s %s: finished, fd %d, this_conn %p, state "
		"%s(%d), ret %d\n", _time_prefix(), __func__, fd, this_conn,
		state_labels[this_conn->conn_state], this_conn->conn_state,
		ret);
#endif

	return ret;
}	/* exs_listen */


/*
 * Called only in our_con_man_thread() context
 * called only from our_con_man_thread()
 *	when it receives an RDMA_CM_EVENT_DISCONNECTED
 *	and this_conn->conn_state == EXS_CONN_CLOSING
 * after this returns, our_con_man_thread calls pthread_exit()
 *
 * the connection was completely set up, and may have been
 * running, but we have started to close it down via a call to exs_close()
 * and an rdma_disconnect() has just completed successfully,
 * so we now finish the tear-down before departing asynchronously.
 *
 * when called, no locks should be locked.
 *
 * on return, this_conn has been freed -- don't reference it again!
 *	and the caller (con_man thread) will self-terminate
 */
static void
our_close_connecting(struct exs_conn *this_conn, const char *message,
		our_qhandle_t *qhandle, int *signo)
{
	TRACE("%s %s: enter, sockno %d, cb %p, "
		"conn_state %s(%d)\n", _time_prefix(), __func__,
		this_conn->sockno, this_conn,
		state_labels[this_conn->conn_state],
		this_conn->conn_state);

	/* destroy all the stuff set up in our_setup_qp() */
	TRACE("%s %s: calling our_free_qp on conn %p\n",
		_time_prefix(), __func__, this_conn);
	our_free_qp(this_conn);

	/* post close event and undo everything done in exs_socket()
	 * except canceling event thread (since we are that thread currently,
	 * and we'll do a pthread_exit() on return from this function)
	 */
	TRACE("%s %s: calling our_undo_exs_socket on conn %p, "
		"level 4\n", _time_prefix(), __func__, this_conn);
	our_undo_exs_socket(this_conn->sockno, this_conn, 4, message, qhandle,
									signo);

	/* on return from our_undo_exs_socket(), this_conn has been freed
	 * and qhandle and signo have been set */
}	/* our_close_connecting */


/*
 * called twice from our_start_close()
 *	once for a client connection, once for an agent connection
 * called only in a user thread context
 *
 * when called, parent_conn->event_mutex MUST be locked
 *	to avoid a race for it with connection manager thread
 *	this mutex protects conn_state
 *
 * Disconnects a connection and transitions any associated QP to the
 * error state, which will flush any posted work requests to the
 * completion queue.  This routine may be called by both the client
 * and agent side of a connection.
 * An RDMA_CM_EVENT_DISCONNECTED event may be generated on both
 * sides of the connection if no more send blocks are posted.
 *
 * returns 0 and changes state if rdma_disconnect started ok,
 * returns 1 and leaves state unchanged if our_con_man thread got a disconnect
 * returns -1, sets errno and leaves state unchanged if rdma_disconnect failed
 *
 * BE CAREFUL -- the event_mutex is on the control_conn, NOT this_conn
 *			because the critical section in our_con_man_thread
 *			is where the BROADCAST is done.
 */
static int
our_try_to_disconnect(struct exs_conn *this_conn,
		unsigned int disconnect_value, const char *message)
{
	int	ret = 0;


	/* rdma_accept() or rdma_connect()
	 * was still last successful con_man operation on this conn
	 * (remote side has not yet caused RDMA_CM_EVENT_DISCONNECTED)
	 * so want to try to call rdma_disconnect()here
	 *
	 * on entry,
	 *	state == EXS_CONN_RUNNING
	 * if no send or ack blocks are posted, issue the rdma_disconnect
	 *	and set state == EXS_CONN_CLOSING
	 * else
	 *	set state == EXS_CONN_DELAY_CLOSING
	 */

	this_conn->conn_flags &= ~(EXS_NEED_LOCAL_CLOSE
						| EXS_OK_TO_ISSUE_IO);

	if(this_conn->pending_send_count <= 0
				&& this_conn->pending_recv_count <= 0)
	{	/* no send or incomplete ack blocks are posted,
		 * ok to trigger local disconnect now
		 */

		if (atomic_swap_if_match(disconnect_value,
					DISCONNECT_NONE,
					&this_conn->disconnect_flag)
							== DISCONNECT_NONE) {
			TRACE("%s %s: calling our_rdma_disconnect for "
				"cb %p, cm_id %p\n", _time_prefix(), message,
				this_conn, this_conn->cm_id);

			ret = our_rdma_disconnect(this_conn, message);

			TRACE("%s %s: back from our_rdma_disconnect, "
				"ret %d, conn_state %s(%d), conn_flags %06x\n",
				_time_prefix(), message, ret,
				state_labels[this_conn->conn_state],
				this_conn->conn_state, this_conn->conn_flags);
		}

		/* this connection has to be finished in the con-man-thread */
		this_conn->conn_state = EXS_CONN_CLOSING;
	}
	else
	{
		TRACE("%s [%ld] %s conn %p socket %d: avoid calling "
			"rdma_disconnect because pending_send_count(%d) > 0 "
			" or pending_recv_count(%d) > 0, conn_state %s(%d), "
			"conn_flags %06x\n",
			_time_prefix(), pthread_self(), message,
			this_conn, this_conn->sockno,
			this_conn->pending_send_count,
			this_conn->pending_recv_count,
			state_labels[this_conn->conn_state],
			this_conn->conn_state, this_conn->conn_flags);

		/* don't issue local rdma_disconnect until all posted sends complete */
		this_conn->conn_state = EXS_CONN_DELAY_CLOSING;
	}

	return ret;
}	/* our_try_to_disconnect */


static int
our_start_close(int sockno, int flags, exs_qhandle_t qhandle,
		exs_ahandle_t ahandle, const char *message)
{
	int			ret = 0, signo;
	int			our_disconnect_state;
	our_qhandle_t		qptr;
	struct exs_conn		*this_conn;

	/* only supported flag is EXS_BLOCK for now, which this function
	 * doesn't care about */
	(void) flags;

#if defined(ALLOW_PID_CHECKING)
	pid_t			current;

	current = getpid();
	TRACE("%s pid %d start %s, socket %d\n",
		_time_prefix(), current, message, sockno);

	if(global_owner != current)
#else
	if(global_owner == 0)
#endif
	{
		errno = EPERM;		/* Operation not permitted */
		return -1;
	}

	/* verify that this socket has been created */
	if((this_conn = our_lookup_socket(sockno)) == NULL)
	{
		errno = EBADF;		/* Bad file number */
		return -1;
	}

	TRACE("%s %s: socket %d, this_conn %p, "
		"pending_recv_count %d, pending_send_count %d, "
		"conn_state %s(%d), conn_flags %0x\n",
		_time_prefix(), message, sockno, this_conn,
		this_conn->pending_recv_count,
		this_conn->pending_send_count,
		state_labels[this_conn->conn_state],
		this_conn->conn_state,
		this_conn->conn_flags);

	/* verify that caller is giving us a valid qhandle */
	if((qptr = _cnv_qnumb_to_ptr(qhandle)) == NULL)
	{	/* Invalid argument */
		return -1;
	}

	/* fill in ALL fields of struct pending_connect for this connection */
	this_conn->conn_pending->qhandle = qptr;
	this_conn->conn_pending->ahandle = ahandle;
	this_conn->conn_pending->err_numb = 0;
	this_conn->conn_pending->sockno = sockno;
	this_conn->conn_pending->evt_type = EXS_EVT_CLOSE;
	this_conn->conn_pending->old_state = EXS_CONN_CLOSED;

	TRACE("\t%4d: final max_posted_send_blocks %d, "
			"limit %d\n", __LINE__,
			this_conn->max_posted_send_blocks,
			this_conn->posted_credits);

	TRACE("\t%4d: final max_posted_recv_blocks %d, "
			"limit %d\n", __LINE__,
			this_conn->max_posted_recv_blocks,
			this_conn->posted_credits);

	fflush(stdout);

	/* need to lock event_mutex to protect conn_state */
	TRACE("\t\t%4d:     LOCK cb %p event_mutex %p\n",
		__LINE__, this_conn, &this_conn->event_mutex);
	pthread_mutex_lock(&this_conn->event_mutex);
	TRACE("\t\t%4d:     PAST cb %p event_mutex %p\n",
		__LINE__, this_conn, &this_conn->event_mutex);

	signo = -1;
	switch(this_conn->conn_state) {

	case EXS_CONN_CLOSING:
	case EXS_CONN_DELAY_CLOSING:
	case EXS_CONN_UNLISTENING:
	case EXS_CONN_ACCEPTING:
		/* impossible because sockno is not in the table for this state */
		/* or user previously called exs_close() on this fd/conn --
		 * it must be in the process of closing down asynchronously
		 */
		errno = EBUSY;		/* Device or resource busy */
		ret = -1;
		break;

	case EXS_CONN_FINISHED:
	case EXS_CONN_UNOPENED:
	case EXS_CONN_CLOSED:
		errno = ENOTCONN;	/* Transport endpoint is not connected*/
		ret = -1;
		break;

	case EXS_CONN_CLIENT_SETUP:
		/* CHANGE THIS -- work it out later */
	case EXS_CONN_RUNNING:
		/* make sure no one finds this_conn again! (invalidate fd) */
		our_rem_socket_from_list(this_conn);

		/* rdma_connect() was last successful con_man operation */
		ret = our_try_to_disconnect(this_conn,
						DISCONNECT_CLOSE_CONN, message);
		break;
	case EXS_CONN_CONNECTING:
		/* fall thru */
	case EXS_CONN_TERMINATING:
		/* remotely triggered EVENT_DISCONNECT was last operation
		 * prior to this explicit call to exs_close()
		 */
		this_conn->conn_state = EXS_CONN_FINISHED;
		this_conn->conn_pending->old_state = EXS_CONN_FINISHED;
		this_conn->conn_flags &=
				~(EXS_NEED_LOCAL_CLOSE | EXS_OK_TO_ISSUE_IO);

		/* make sure user can't find this client again */
		our_rem_socket_from_list(this_conn);

		break;

	case EXS_CONN_SETUP:
		/* our_setup_qp() during exs_connect was last operation */
		our_free_qp(this_conn);
	case EXS_CONN_ROUTING:
		/* rdma_resolve_route() during exs_connect was last operation
		 * so asynchronous activity on connection still expected!
		 * TO DO -- deal with this
		 */
	case EXS_CONN_ADDRESSING:
		/* rdma_resolve_addr() during exs_connect was last operation
		 * so asynchronous activity on connection still expected!
		 * TO DO -- deal with this
		 */
		this_conn->conn_state = this_conn->conn_pending->old_state;
	case EXS_CONN_BOUND:
		/* exs_bind() was last successful operation */
	case EXS_CONN_UNBOUND:
		/* exs_socket() was last successful operation */
		our_undo_exs_socket(sockno, this_conn, 7, message,
					&qptr, &signo);
		/* this_conn no longer valid (the conn struct has been freed) */
		this_conn = NULL;
		break;

	case EXS_CONN_AGENT_SETUP:
		/* CHANGE THIS -- work it out later */
		/* make sure no one finds this_conn again! (invalidate fd) */
		our_rem_socket_from_list(this_conn);

		/* rdma_accept() was last successful con_man operation
		 * (remote side has not yet caused RDMA_CM_EVENT_DISCONNECTED)
		 * so rdma_disconnect() must be called by exs_close() here
		 */

		ret = our_try_to_disconnect(this_conn, DISCONNECT_CLOSE_CONN, message);
		break;

	case EXS_CONN_LISTENING:
		/* exs_listen() was last successful operation */

		this_conn->conn_state = EXS_CONN_UNLISTENING;

		if (atomic_swap_if_match(DISCONNECT_LISTEN,
					DISCONNECT_NONE,
					&this_conn->disconnect_flag)
							== DISCONNECT_NONE) {
			pthread_cancel(this_conn->event_thread);
		}

		our_undo_exs_socket(sockno, this_conn, 7, message,
					&qptr, &signo);
		/* this_conn is now invalid */
		this_conn = NULL;
		break;

	}	/* switch */

	if (this_conn) {
		our_disconnect_state
			= atomic_set_bits(0x4, &this_conn->disconnect_state);

		/* parent conn still valid, its event_mutex is still locked */
		TRACE("\t\t%4d:   UNLOCK cb %p event_mutex %p\n",
			__LINE__, this_conn, &this_conn->event_mutex);
		pthread_mutex_unlock(&this_conn->event_mutex);

		if (our_disconnect_state == 0x7) {
			/* We're the last ones out. */
			TRACE("%s %s: doing final cleanup\n",
					_time_prefix(), message);
			our_free_qp(this_conn);
			our_undo_exs_socket(sockno, this_conn, 4, message,
						&qptr, &signo);
			this_conn = NULL;
		}
	}

#if defined(ALLOW_FULL_TRACE)
	TRACE("%s %s: finished, socket %d, returning %d, signo %d\n",
		_time_prefix(), message, sockno, ret, signo);
#endif


	if(signo >= 0)
		_evt_queue_do_notify(qptr, signo);

	return ret;
}	/* our_start_close */


/*
 *	exs_close() - not in Interconnect Software Consortium Issue 1.0
 *			      was added by us
 *
 *	returns 0 if started ok
 *		else -1 with errno set to error value
 *
 * If the EXS_BLOCK bit is set in flags, this function will block until the
 * operation completes (i.e., the operation will be synchronous).  In this case,
 * the qhandle and ahandle arguments are ignored.  No other flags are supported.
 */
EXS_EXPORT int
exs_close(int sockno, int flags, exs_qhandle_t qhandle, exs_ahandle_t ahandle)
{
	if (flags & ~(EXS_BLOCK)) {
		errno = EINVAL;
		return -1;
	}

	if (flags & EXS_BLOCK) {
		return exs_blocking_close(sockno);
	} else {
		return our_start_close(sockno, flags, qhandle, ahandle,
				__func__);
	}
}	/* exs_close */


/*
 *	exs_blocking_close() - not in Interconnect Software Consortium Issue 1.0
 *			      was added by us
 *
 *	returns 0 if close finished ok
 *		else -1 with errno set to error value
 */
EXS_EXPORT int
exs_blocking_close(int sockno)
{
	ssize_t			ret;
	our_qhandle_t		local_qhandle;
	exs_ahandle_t		local_ahandle;
	exs_event_t		local_vector;

	/* dynamically create struct evt_queue (can't use connection's
	 * queue_cache because that connection is going away!
	 */
	local_qhandle = _evt_queue_init(1);
	if(local_qhandle == NULL)
	{	/* couldn't create it, errno already set */
		return -1;
	}

	local_ahandle = (void *)local_qhandle;
	ret = our_start_close(sockno, 0, _cnv_ptr_to_qnumb(local_qhandle),
						local_ahandle, __func__);

	TRACE("%s %s: back from our_start_close, qhandle %p, ahandle %p,"
			" ret %ld\n", _time_prefix(), __func__, local_qhandle,
			local_ahandle, ret);

	if(ret == 0)
	{	/* caller wants blocking close, wait for completion event */

		ret = _evt_queue_dequeue_vector(local_qhandle, &local_vector,
								1, NULL);

		TRACE("%s %s: back from dequeue, qhandle %p, "
			"ahandle %p, ret %ld, exs_evt_type %d, "
			"exs_evt_errno %d, exs_evt_ahandle %p\n",
			_time_prefix(), __func__, local_qhandle, local_ahandle,
			ret, local_vector.exs_evt_type,
			local_vector.exs_evt_errno,
			local_vector.exs_evt_ahandle);

		if(ret != -1)
		{	/* no error on dequeue itself, check the completion */
			if((errno = local_vector.exs_evt_errno) != 0)
				ret = -1;
			else
			{	/* no error on the close */
				ret = 0;
			}
		}
	}

	/* always delete the locally created queue (can't return it to
	 * connection's cache because connection is gone!  Anyway, we
	 * didn't get it from that cache for exactly that reason.
	 */
	_evt_queue_deinit(local_qhandle);

	return ret;
}	/* exs_blocking_close */


/*
 *	exs_getsockname() - not in Interconnect Software Consortium Issue 1.0
 *		      was added by us
 *
 *	returns 0 if name obtained ok
 *		else -1 with errno set to error value
 */
EXS_EXPORT int
exs_getsockname(int sockno, struct sockaddr *name, socklen_t *namelen)
{
	struct exs_conn		*this_conn;

#if defined(ALLOW_PID_CHECKING)
	pid_t			current;

	current = getpid();
	TRACE("%s pid %d start %s, socket %d\n",
		_time_prefix(), current, __func__, sockno);

	if(global_owner != current)
#else
	if(global_owner == 0)
#endif
	{
		errno = EPERM;		/* Operation not permitted */
		return -1;
	}

	/* verify that this socket has been created */
	if((this_conn = our_lookup_socket(sockno)) == NULL)
	{
		errno = EBADF;		/* Bad file number */
		return -1;
	}

	/* verify that this socket has already been connected */
	if(this_conn->conn_state <= EXS_CONN_UNBOUND)
	{
		errno = ENOTCONN;	/* Transport endpoint is not connected*/
		return -1;
	}

	TRACE("%s %s: socket %d, this_conn %p, state %s(%d)\n",
		_time_prefix(), __func__, sockno, this_conn,
		state_labels[this_conn->conn_state], this_conn->conn_state);

	*namelen = our_copy_src_addr(this_conn->cm_id, name, *namelen);

	return 0;
}	/* exs_getsockname */


/*
 *	exs_getpeername() - not in Interconnect Software Consortium Issue 1.0
 *		      was added by us
 *
 *	returns 0 if name obtained ok
 *		else -1 with errno set to error value
 */
EXS_EXPORT int
exs_getpeername(int sockno, struct sockaddr *name, socklen_t *namelen)
{
	struct exs_conn		*this_conn;

#if defined(ALLOW_PID_CHECKING)
	pid_t			current;

	current = getpid();
	TRACE("%s pid %d start %s, socket %d\n",
		_time_prefix(), current, __func__, sockno);

	if(global_owner != current)
#else
	if(global_owner == 0)
#endif
	{
		errno = EPERM;		/* Operation not permitted */
		return -1;
	}

	/* verify that this socket has been created */
	if((this_conn = our_lookup_socket(sockno)) == NULL)
	{
		errno = EBADF;		/* Bad file number */
		return -1;
	}

	/* verify that this socket has already been connected */
	if(this_conn->conn_state != EXS_CONN_RUNNING)
	{
		errno = ENOTCONN;	/* Transport endpoint is not connected*/
		return -1;
	}

	TRACE("%s %s: socket %d, this_conn %p, state %s(%d)\n",
		_time_prefix(), __func__, sockno, this_conn,
		state_labels[this_conn->conn_state], this_conn->conn_state);

	*namelen = our_copy_dst_addr(this_conn->cm_id, name, *namelen);

	return 0;
}	/* exs_getpeername */


/* set to non-zero to enable debug printing */
int _our_debug_flag = 0;

/*
 *	exs_set_debug_flag() - not in Interconnect Software Consortium Issue 1.0
 *		      was added by us
 *
 *	returns previous value of global variable _our_debug_flag
 */
EXS_EXPORT int
exs_set_debug_flag(int value)
{
	int	ret;

	ret = _our_debug_flag;
	_our_debug_flag = value;
	return ret;
}	/* exs_set_debug_flag */


/*
 *	exs_fcntl() - not in Interconnect Software Consortium Issue 1.0
 *			      was added by us
 *
 *	returns 0 if attribute's value was set ok
 *		else -1 with errno set to error value
 */
EXS_EXPORT long
exs_fcntl(int fd, int command_name, long command_value)
{
	long			ret = 0;
	int			old_value; /*used only in GET/SETCOMPTHREADCPU*/
	struct exs_conn		*this_conn;

#if defined(ALLOW_PID_CHECKING)
	pid_t			current;

	current = getpid();
	TRACE("%s pid %d start %s, socket %d, "
			"command_name %d, command_value %0lx\n",
		_time_prefix(), current, __func__, fd,
		command_name, command_value);

	if(global_owner != current)
#else
	if(global_owner == 0)
#endif
	{
		errno = EPERM;		/* Operation not permitted */
		return -1;
	}

	/* verify that this socket has been created */
	if((this_conn = our_lookup_socket(fd)) == NULL)
	{
		errno = EBADF;		/* Bad file number */
		return -1;
	}

	TRACE("%s %s: socket %d, this_conn %p, state %s(%d)\n",
		_time_prefix(), __func__, fd, this_conn,
		state_labels[this_conn->conn_state], this_conn->conn_state);

	switch(command_name)
	{
	case EXS_F_GETFD:		/* get file descriptor flags */
		ret = this_conn->fd_flags;
		break;

	case EXS_F_SETFD:		/* set file descriptor flags */
		if((command_value & ~(EXS_FD_ALL_FLAGS_MASK)) != 0)
		{	/* Illegal flag value */
			errno = EINVAL;
			return -1;
		}
		/* verify that this socket has not already been connected */
		if(this_conn->conn_state > EXS_CONN_BOUND)
		{	/* Transport endpoint is already connected */
			errno = EISCONN;
			return -1;
		}

		/* ok to set these flags, return old flag settings */
		ret = this_conn->fd_flags;

		if (this_conn->is_seqpacket) {
			/* SOCK_SEQPACKET does not support Nagle */
			command_value |= EXS_FD_NODELAY;
		} else {
			/* SOCK_STREAM does not support small packet copy
			 * unreg */
			command_value &= ~EXS_FD_SP_COPY_UNREG;
		}

		if (!this_conn->is_seqpacket) {
			if (command_value & EXS_FD_NODELAY) {
				this_conn->xmit_if_match
					= &our_xmit_if_match_stream;
				this_conn->enqueue_send
					= &our_enqueue_send_stream;
				this_conn->handle_recv_advert
					= &our_handle_recv_advert_stream;
				this_conn->handle_send_write
					= &our_handle_send_write_stream;
				this_conn->handle_recv_write_notify
					= &our_handle_recv_write_notify_stream;
			} else {
				this_conn->xmit_if_match
					= &our_xmit_if_match_streamcopy;
				this_conn->enqueue_send
					= &our_enqueue_send_streamcopy;
				this_conn->handle_recv_advert
					= &our_handle_recv_advert_streamcopy;
				this_conn->handle_send_write
					= &our_handle_send_write_streamcopy;
				this_conn->handle_recv_write_notify
					= &our_handle_recv_write_notify_streamcopy;
			}
		}

		this_conn->fd_flags = command_value;
		break;

	case EXS_F_GETSPMAXSIZE:	/* get small packet max size */
		ret = this_conn->small_packet_max_size;
		break;

	case EXS_F_SETSPMAXSIZE:	/* set small packet max size */
		/* verify that this socket has not already been connected */
		if(this_conn->conn_state > EXS_CONN_BOUND)
		{	/* Transport endpoint is already connected */
			errno = EISCONN;
			return -1;
		}

		if(command_value < 0
				|| command_value > MAX_SMALL_PACKET_MAX_SIZE)
		{	/* small packet max size would become too big */
			errno = EINVAL;
			return -1;
		}
		/* new small_packet_max_size is ok, use it, return old value */
		ret = this_conn->small_packet_max_size;
		this_conn->small_packet_max_size = command_value;
		break;

	case EXS_F_GETINLINEMAXSIZE:	/* get inline max size */
		ret = this_conn->inline_limit;
		break;

	case EXS_F_SETINLINEMAXSIZE:	/* set inline max size */
		/* verify that this socket has not already been connected */
		if(this_conn->conn_state > EXS_CONN_BOUND)
		{	/* Transport endpoint is already connected */
			errno = EISCONN;
			return -1;
		}

		/* accept large values since they will be negotiated down
		 * if the card doesn't support it */
		if (command_value < 0) {
			errno = EINVAL;
			return -1;
		}
		/* new small_packet_max_size is ok, use it, return old value */
		ret = this_conn->inline_limit;
		this_conn->inline_limit = command_value;
		break;

	case EXS_F_GETFLOWCONTROLCREDITS:	/* get send credits */
		ret = this_conn->posted_credits;
		break;

	case EXS_F_SETFLOWCONTROLCREDITS:	/* set send credits */
		/* verify that this socket has not already been connected */
		if(this_conn->conn_state > EXS_CONN_BOUND)
		{	/* Transport endpoint is already connected */
			errno = EISCONN;
			return -1;
		}

		if(command_value <= 0)
		{	/* send credit max size would become too big */
			errno = EINVAL;
			return -1;
		}
		/* new send credit max size is ok, use it, return old value */
		ret = this_conn->posted_credits;
		our_set_send_credit_limits(this_conn, command_value);
		break;

	case EXS_F_GETSTREAMBUFSIZE:
		if (this_conn->is_seqpacket
				|| (this_conn->fd_flags & EXS_FD_NODELAY)) {
			errno = EINVAL;
			return -1;
		}

		ret = this_conn->stream_buf_size;
		break;

	case EXS_F_SETSTREAMBUFSIZE:
		/* verify that this socket has not already been connected */
		if(this_conn->conn_state > EXS_CONN_BOUND) {
			/* Transport endpoint is already connected */
			errno = EISCONN;
			return -1;
		}

		if (this_conn->is_seqpacket
				|| (this_conn->fd_flags & EXS_FD_NODELAY)) {
			errno = EINVAL;
			return -1;
		}

		if(command_value <= 0
				|| command_value > EXS_MAX_STREAM_BUF_SIZE) {
			errno = EINVAL;
			return -1;
		}

		/* new send credit max size is ok, use it, return old value */
		ret = this_conn->stream_buf_size;
		this_conn->stream_buf_size = command_value;
		break;

	case EXS_F_GETCOMPTHREADCPU:
		if(this_conn->conn_state >= EXS_CONN_SETUP
				&& this_conn->conn_state != EXS_CONN_LISTENING){
			/* Completion thread already started; get/set its
			 * affinity normally */
			if ((ret = _exs_set_thread_affinity(
					this_conn->comp_thread,
					INT_MAX, &old_value)) >= 0) {
				ret = old_value;
			}
		} else {
			ret = this_conn->comp_thread_cpu;
		}
		break;

	case EXS_F_SETCOMPTHREADCPU: /* set cpu for completion thread */
		if(this_conn->conn_state >= EXS_CONN_SETUP
				&& this_conn->conn_state != EXS_CONN_LISTENING){
			/* Completion thread already started; get/set its
			 * affinity normally */
			if ((ret = _exs_set_thread_affinity(
					this_conn->comp_thread,
					command_value, &old_value)) >= 0) {
				ret = old_value;
			}
		} else {
			ret = this_conn->comp_thread_cpu;
			if (command_value != INT_MAX) {
				this_conn->comp_thread_cpu = command_value;
			}
		}

		break;

	default:
		/* caller gave us an invalid command name */
		errno = EINVAL;		/* Invalid argument */
		ret = -1;
		break;
	}	/* switch */

	return ret;
}	/* exs_fcntl */


EXS_EXPORT int
exs_get_version(char *buf, size_t size)
{
	size_t len = strlen(our_version);
	if (len >= size) {
		errno = ENOSPC;
		return -1;
	}
	strncpy(buf, our_version, size);
	return 0;
}


/*
 *	exs_print_connection_attributes() - not in Interconnect Software Consortium Issue 1.0
 *		      was added by us
 *
 *	prints out attributes of a connection
 */
EXS_EXPORT int
exs_print_connection_attributes(FILE *fptr, int sockno, char *message)
{
	struct exs_conn		*this_conn;

#if defined(ALLOW_PID_CHECKING)
	pid_t			current;

	current = getpid();
	TRACE("%s pid %d start %s, socket %d\n",
		_time_prefix(), current, __func__, sockno);

	if(global_owner != current)
#else
	if(global_owner == 0)
#endif
	{
		errno = EPERM;		/* Operation not permitted */
		return -1;
	}

	/* verify that this socket has been created */
	if((this_conn = our_lookup_socket(sockno)) == NULL)
	{
		errno = EBADF;		/* Bad file number */
		return -1;
	}

	fprintf(fptr, "%s fd %d: %36s %u.%u.%u\n", message,
		sockno, "exs version", EXS_VERSION, EXS_SUBVERSION,
		EXS_SUBSUBVERSION);

	fprintf(fptr, "%s fd %d: %36s %u\n", message,
			sockno, "posted credits",
			this_conn->posted_credits);

	fprintf(fptr, "%s fd %d: %36s %u\n", message,
			sockno, "small packet max size",
			this_conn->small_packet_max_size);

	fprintf(fptr, "%s fd %d: %36s %u\n", message,
			sockno, "small packet max size limit",
			MAX_SMALL_PACKET_MAX_SIZE);

	fprintf(fptr, "%s fd %d: %36s %3u (requested %3u)\n", message,
			sockno, "qp max_inline_data",
			this_conn->obtained_cap.max_inline_data,
			this_conn->requested_cap.max_inline_data);

	fprintf(fptr, "%s fd %d: %36s %3u (requested %3u)\n", message,
			sockno, "qp max_send_wr",
			this_conn->obtained_cap.max_send_wr,
			this_conn->requested_cap.max_send_wr);

	fprintf(fptr, "%s fd %d: %36s %3u (requested %3u)\n", message,
			sockno, "qp max_recv_wr",
			this_conn->obtained_cap.max_recv_wr,
			this_conn->requested_cap.max_recv_wr);

	fprintf(fptr, "%s fd %d: %36s %3u (requested %3u)\n", message,
			sockno, "number of completion queue elements",
			this_conn->obtained_cq_nelements,
			this_conn->requested_cq_nelements);

	fprintf(fptr, "%s fd %d: %36s %d\n", message,
			sockno, "path mtu size",
			this_conn->obtained_path_mtu_size);

	fflush(fptr);
	return 0;
}	/* exs_print_connection_attributes */

static void
print_send_wr(FILE *fptr, struct ibv_send_wr *wr)
{
	char opcode[40];
	int i;
	int has_imm_data;
	int has_rdma;

	if (! TRACING_IS_ON) {
		return;
	}

	while (wr != NULL) {
		switch (wr->opcode) {
		case IBV_WR_RDMA_WRITE:
			strcpy(opcode, "IBV_WR_RDMA_WRITE");
			has_rdma = 1;
			has_imm_data = 0;
			break;
		case IBV_WR_RDMA_WRITE_WITH_IMM:
			strcpy(opcode, "IBV_WR_RDMA_WRITE_WITH_IMM");
			has_rdma = has_imm_data = 1;
			break;
		case IBV_WR_SEND:
			strcpy(opcode, "IBV_WR_SEND");
			has_rdma = has_imm_data = 0;
			break;
		case IBV_WR_SEND_WITH_IMM:
			strcpy(opcode, "IBV_WR_SEND_WITH_IMM");
			has_rdma = 0;
			has_imm_data = 1;
			break;
		case IBV_WR_RDMA_READ:
			strcpy(opcode, "IBV_WR_RDMA_READ");
			has_rdma = 1;
			has_imm_data = 0;
			break;
		default:
			has_rdma = 0;
			has_imm_data = 0;
			snprintf(opcode, sizeof(opcode) - 1,
						"(unknown opcode %d)", wr->opcode);
		}

		fprintf(fptr, "ibv_send_wr wr_id=%016" PRIx64 " opcode=%s\nFlags: ",
				wr->wr_id, opcode);

		if (wr->send_flags & IBV_SEND_SIGNALED) {
			fprintf(fptr, "IBV_SEND_SIGNALED ");
		}
		if (wr->send_flags & IBV_SEND_INLINE) {
			fprintf(fptr, "IBV_SEND_INLINE ");
		}

		if (has_imm_data) {
			fprintf(fptr, "\nimm_data=%" PRIu32, ntohl(wr->imm_data));
		}

		if (has_rdma) {
			fprintf(fptr, "\nrdma: rkey=%08" PRIx32 " remote_addr=%016" PRIx64,
				wr->wr.rdma.rkey, wr->wr.rdma.remote_addr);
		}

		fprintf(fptr, "\nnum_sge=%d\n", wr->num_sge);

		for (i = 0; i < wr->num_sge; ++i) {
			fprintf(fptr, "sg_list[%d]: addr=%016" PRIx64 " length=%" PRIu32 " lkey=%08" PRIx32 "\n",
					i, wr->sg_list[i].addr, wr->sg_list[i].length,
					wr->sg_list[i].lkey);
		}

		fprintf(fptr, "\n");
		wr = wr->next;
	}
}

/* vi: set autoindent tabstop=8 shiftwidth=8 : */
